////////////////////////////////////////////////////////////////////////////
// ar_parm.cc
//      Parameters that affect mapping of video / scenery to headset
//      screens.  May need to be changed for different headests,
//      different mounting of the camera, etc.
//
// Author: Jon W. Wallace, Lafayette College
// Date:   May 2020
////////////////////////////////////////////////////////////////////////////

#include <string.h>

#include "ar_parm.h"
#include "log.h"

// DEBUG: Try just center location where calibration was done
//const float s_cap_crop_w_frac = 0.5;
//const float s_cap_crop_h_frac = 0.5;
// Full capture frame
const float s_cap_crop_w_frac = 1.0;
const float s_cap_crop_h_frac = 1.0;
float s_cap_crop_lx = 0.5;
float s_cap_crop_ly = 0.25;
float s_cap_crop_rx = 0.5;
float s_cap_crop_ry = 0.75;
float s_cap_crop_w = s_cap_crop_w_frac*1.0;
float s_cap_crop_h = s_cap_crop_h_frac*0.5;

// Output frame mapping.  These values give the corners of a
// quadralateral that the captured video frames will be mapped
// to.  Ordering is 1. lower-left, 2. lower-right, 3. upper-right, 4. upper-left.
// These were found using the "calibrate" code that was processed with
// matlab scripts to find the mapping.

float M_cap_l[6] = 
{
    1.04208742909605e-03,  -5.78652539205006e-05,  -3.99556985999964e-01,
    -1.08540699875441e-05,   9.51593905122146e-04,  -6.35441560164299e-01
};

float M_cap_r[6] =
{
    1.01539822373828e-03,  -2.21578071760672e-05,  -5.39723481489585e-01,
    -1.18277978919478e-05,   9.40428898326115e-04,  -6.37940278011912e-01
};

// Globals to store corners of the quadralaterals
// that the capture video gets placed into for
// the L and R frames.

float s_rend_vid_lx1;
float s_rend_vid_ly1;
float s_rend_vid_lx2;
float s_rend_vid_ly2;
float s_rend_vid_lx3;
float s_rend_vid_ly3;
float s_rend_vid_lx4;
float s_rend_vid_ly4;

float s_rend_vid_rx1;
float s_rend_vid_ry1;
float s_rend_vid_rx2;
float s_rend_vid_ry2;
float s_rend_vid_rx3;
float s_rend_vid_ry3;
float s_rend_vid_rx4;
float s_rend_vid_ry4;
    
////////////////////////////////////////////////////////////////////////////
// ar_parm_map1
//      Maps a single point from capture image coordinates to output
//      texture logical coordinates.  This is also used for mapping
//      gauges that are detected using Aruco tags.
////////////////////////////////////////////////////////////////////////////
void ar_parm_map1(float x, float y, float *out_x, float *out_y, float *M)
{
    *out_x = M[0]*x + M[1]*y + M[2];
    *out_y = M[3]*x + M[4]*y + M[5];
}

////////////////////////////////////////////////////////////////////////////
// ar_print_s_rend
//      Print the render mapping values (debug)
////////////////////////////////////////////////////////////////////////////
void ar_print_s_rend()
{
    log_printf(log_quiet, "s_rend_vid_lx1= %.4f\n", s_rend_vid_lx1);
    log_printf(log_quiet, "s_rend_vid_ly1= %.4f\n", s_rend_vid_ly1);
    log_printf(log_quiet, "s_rend_vid_lx2= %.4f\n", s_rend_vid_lx2);
    log_printf(log_quiet, "s_rend_vid_ly2= %.4f\n", s_rend_vid_ly2);
    log_printf(log_quiet, "s_rend_vid_lx3= %.4f\n", s_rend_vid_lx3);
    log_printf(log_quiet, "s_rend_vid_ly3= %.4f\n", s_rend_vid_ly3);
    log_printf(log_quiet, "s_rend_vid_lx4= %.4f\n", s_rend_vid_lx4);
    log_printf(log_quiet, "s_rend_vid_ly4= %.4f\n", s_rend_vid_ly4);
                                                                  
    log_printf(log_quiet, "s_rend_vid_rx1= %.4f\n", s_rend_vid_rx1);
    log_printf(log_quiet, "s_rend_vid_ry1= %.4f\n", s_rend_vid_ry1);
    log_printf(log_quiet, "s_rend_vid_rx2= %.4f\n", s_rend_vid_rx2);
    log_printf(log_quiet, "s_rend_vid_ry2= %.4f\n", s_rend_vid_ry2);
    log_printf(log_quiet, "s_rend_vid_rx3= %.4f\n", s_rend_vid_rx3);
    log_printf(log_quiet, "s_rend_vid_ry3= %.4f\n", s_rend_vid_ry3);
    log_printf(log_quiet, "s_rend_vid_rx4= %.4f\n", s_rend_vid_rx4);
    log_printf(log_quiet, "s_rend_vid_ry4= %.4f\n", s_rend_vid_ry4);
}

////////////////////////////////////////////////////////////////////////////
// ar_parm_get_s_rend
//      Computes the quadralateral corners for rendering the
//      capture frames onto the VR video based on the mapping matrices.
////////////////////////////////////////////////////////////////////////////
void ar_parm_get_s_rend(int Nx, int Ny)
{
    // Corners of cropped input area (in pixels) Lower-left, upper-right
    float s_cap_x1 = (0.5 - s_cap_crop_w_frac/2)*(float)Nx;
    float s_cap_y1 = (0.5 - s_cap_crop_h_frac/2)*(float)Ny;
    float s_cap_x2 = (0.5 + s_cap_crop_w_frac/2)*(float)Nx;
    float s_cap_y2 = (0.5 + s_cap_crop_h_frac/2)*(float)Ny;

    // Compute mapping of captured image corners (left frame)
    ar_parm_map1(s_cap_x1, s_cap_y1, &s_rend_vid_lx1, &s_rend_vid_ly1, M_cap_l);
    ar_parm_map1(s_cap_x2, s_cap_y1, &s_rend_vid_lx2, &s_rend_vid_ly2, M_cap_l);
    ar_parm_map1(s_cap_x2, s_cap_y2, &s_rend_vid_lx3, &s_rend_vid_ly3, M_cap_l);
    ar_parm_map1(s_cap_x1, s_cap_y2, &s_rend_vid_lx4, &s_rend_vid_ly4, M_cap_l);
    // Right frame
    ar_parm_map1(s_cap_x1, s_cap_y1, &s_rend_vid_rx1, &s_rend_vid_ry1, M_cap_r);
    ar_parm_map1(s_cap_x2, s_cap_y1, &s_rend_vid_rx2, &s_rend_vid_ry2, M_cap_r);
    ar_parm_map1(s_cap_x2, s_cap_y2, &s_rend_vid_rx3, &s_rend_vid_ry3, M_cap_r);
    ar_parm_map1(s_cap_x1, s_cap_y2, &s_rend_vid_rx4, &s_rend_vid_ry4, M_cap_r);

    ar_print_s_rend();
}

////////////////////////////////////////////////////////////////////////////
// ar_parm_make_ideal_map
//      Mainly for debugging.  This function sets the mapping matrix to
//      map the captured frames to the full display window (texture) size.
////////////////////////////////////////////////////////////////////////////
void ar_parm_make_ideal_map(int w, int h)
{
    memset(M_cap_l, 0, sizeof(M_cap_l));
    memset(M_cap_r, 0, sizeof(M_cap_r));

    M_cap_l[0] = 2.0/(float)w;
    M_cap_l[4] = 2.0/(float)h;
    M_cap_l[2] = -1;
    M_cap_l[5] = -1;
    
    M_cap_r[0] = 2.0/(float)w;
    M_cap_r[4] = 2.0/(float)h;
    M_cap_r[2] = -1;
    M_cap_r[5] = -1;
}




