/*============================================================================
 * aruco.cc
 *      Front end to OpenCV to detect ARUCO tags.
 *
 * Author: Jon W. Wallace, Lafayette College
 * Date:   May 2020
 *==========================================================================*/

#include <opencv2/opencv.hpp>
#include <opencv2/highgui.hpp>
#include <opencv2/aruco.hpp>
#include <math.h>
#include "main.h"
#include "log.h"
#include "capture.h"
#include "myProfile.h"
#include "process.h"

using namespace cv;

#include "aruco.h"

static const char *arucoDicts[] = {
  "DICT_4X4_50",
  "DICT_4X4_100",
  "DICT_4X4_250",
  "DICT_4X4_1000",
  "DICT_5X5_50",
  "DICT_5X5_100",
  "DICT_5X5_250",
  "DICT_5X5_1000",
  "DICT_6X6_50",
  "DICT_6X6_100",
  "DICT_6X6_250",
  "DICT_6X6_1000",
  "DICT_7X7_50",
  "DICT_7X7_100",
  "DICT_7X7_250",
  "DICT_7X7_1000",
  "DICT_ARUCO_ORIGINAL"};
static const int N_arucoDicts = (sizeof(arucoDicts)/sizeof(char *));

static Ptr<aruco::Dictionary> dictionary;

typedef void (*arucoUndistortFuncPtr)(Mat& , Mat&);

// Profile variables for aruco thread
typedef struct
{
    cpuProfDef fifoWait;
    cpuProfDef deinterlace;
    cpuProfDef convert;
    cpuProfDef detect;
    cpuProfDef updateTable;
    cpuProfDef profAll;
    wallProfDef profWAll;
    uint32_t numFrames;
} profArucoDef;

typedef struct
{
    uint32_t bufferIdx;
    pthread_t thread;
    arucoPosDef tags[ARUCO_BUFFER_DEPTH][ARUCO_MAX_IDS];
    pthread_mutex_t mutex;
    volatile int threadStop;
    volatile int threadDone;
    fifoDef *fifo;
    profArucoDef p;
    arucoUndistortFuncPtr uFunc;

} arucoThreadDataDef;

static arucoThreadDataDef arucoThreadDataL, arucoThreadDataR;

int arucoCapWidth = 0;
int arucoCapHeight = 0;

fifoDef *arucoCapFifoL, *arucoCapFifoR;
volatile int arucoFifosReady = 0;

void arucoThreadFunc(void *parm);

void arucoProfilePrintOne(profArucoDef *p)
{
    profPrintOne("FIFO Wait", &(p->fifoWait), &(p->profAll), p->numFrames);
    profPrintOne("Deinterlace", &(p->deinterlace), &(p->profAll), p->numFrames);
    profPrintOne("Convert", &(p->convert), &(p->profAll), p->numFrames);
    profPrintOne("Detect", &(p->detect), &(p->profAll), p->numFrames);
    profPrintOne("Update Table", &(p->updateTable), &(p->profAll), p->numFrames);
    log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
    profPrintOne("Total", &(p->profAll), &(p->profAll), p->numFrames);
    log_printf(log_quiet, "%-20s %8.2f\n", "CPU FPS", (double)p->numFrames/profTimeSec(&p->profAll));
    log_printf(log_quiet, "%-20s %8.2f\n", "FPS (Wall)", (double)p->numFrames/profTimeSec(&p->profWAll));
    log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
}

////////////////////////////////////////////////////////////////////////////
// arucoProfilePrint()
////////////////////////////////////////////////////////////////////////////
void arucoProfilePrint()
{
    log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
    log_printf(log_quiet, "Aruco Detect Thread L                                      | Output Rate\n");
    log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
    arucoProfilePrintOne(&(arucoThreadDataL.p));
    log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
    log_printf(log_quiet, "Aruco Detect Thread R                                      | Output Rate\n");
    log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
    arucoProfilePrintOne(&(arucoThreadDataR.p));
}

////////////////////////////////////////////////////////////////////////////
// arucoListDicts
//      Lists current dictionaries.
////////////////////////////////////////////////////////////////////////////
void arucoListDicts()
{
    fprintf(stderr, "Current Dictionaries:\n");
    for (int i=0; i<N_arucoDicts; i++)
    {
        fprintf(stderr, "%d: %s\n", i, arucoDicts[i]);
    }
}

////////////////////////////////////////////////////////////////////////////
// arucoInitOneSide
//      Initializes data for a single side (L or R)
////////////////////////////////////////////////////////////////////////////
int arucoInitOneSide(arucoThreadDataDef *data, fifoDef *fifo, arucoUndistortFuncPtr uFunc)
{
    int ccode;
    uint32_t data_size;

    // Clear tag circular buffer
    data->bufferIdx = 0;
    memset(data->tags, 0, sizeof(data->tags));
    
    // Initialize a Mutex to protect the tags structures, in case they are written in one
    // thread (capture) and read in another thread (render)
    if (pthread_mutex_init(&(data->mutex), NULL) != 0)
    {
        log_printf(log_error, "Unable to create tags mutex in aruco module.\n");
        ccode = -1;
        goto backout_base;
    }

    // Store reference to FIFO that should be used for input
    data->fifo = fifo;

    // Store reference to undistort function
    if ((!mainOptions.no_undistort) && (uFunc))
    {
        data->uFunc = uFunc;
    }
    
    // Start the processing threads
    data->threadStop=0;
    data->threadDone=0;
    if (pthread_create(&(data->thread),
                       NULL,
                       (void * (*)(void *)) arucoThreadFunc,
                       (void *) data) != 0)
    {
        log_printf(log_error, "Unable to start aruco thread.");
        ccode = -2;
        goto backout_mutex;
    }
    
    // Create required OpenCV images
    // frame_sub = Mat(h, w, CV_8UC4, Scalar(0,0,0,0));

    return(0);
    
 backout_mutex:
    pthread_mutex_destroy(&data->mutex);
    
 backout_base:
    return(ccode);
}

void arucoUninitOneSide(arucoThreadDataDef *data)
{
    // Wait for thread to stop
    data->threadStop = 1;
    while (!data->threadDone)
    {
        WAIT_SHORT;
    }
    pthread_mutex_destroy(&data->mutex);
}

////////////////////////////////////////////////////////////////////////////
// arucoInit
//      Initialize data used for detection of aruco tags.
////////////////////////////////////////////////////////////////////////////
int arucoInit(int tag_dict, int w, int h)
{
    int ccode;
    uint32_t data_size;

    // Get Aruco tag dictionary
    dictionary = aruco::getPredefinedDictionary(aruco::PREDEFINED_DICTIONARY_NAME(tag_dict));

    arucoCapWidth = w;
    arucoCapHeight = h;

    // Create the needed FIFOs
    data_size = sizeof(struct timespec) + arucoCapWidth*arucoCapHeight;
    if (fifoCreate(data_size, NUM_ARUCO_FRAMES, &arucoCapFifoL) != 0)
    {
        log_printf(log_error, "Unable to create arcuo capture L FIFO.\n");
        ccode = -3;
        goto backout_base;
    }
    if (fifoCreate(data_size, NUM_ARUCO_FRAMES, &arucoCapFifoR) != 0)
    {
        log_printf(log_error, "Unable to create arcuo capture R FIFO.\n");
        ccode = -3;
        goto backout_L_fifo;
    }

    arucoFifosReady = 1;
    
    if (arucoInitOneSide(&arucoThreadDataL, arucoCapFifoL, &processUndistortL) != 0)
    {
        log_printf(log_error, "Unable to allocate L side in aruco module.\n");
        ccode = -1;
        goto backout_R_fifo;
    }
    
    if (arucoInitOneSide(&arucoThreadDataR, arucoCapFifoR, &processUndistortR) != 0)
    {
        log_printf(log_error, "Unable to allocate R side in aruco module.\n");
        ccode = -1;
        goto backout_left;
    }

    return(0);

 backout_left:
    arucoUninitOneSide(&arucoThreadDataL);

 backout_R_fifo:
    arucoFifosReady = 0;
    fifoDestroy(arucoCapFifoR);
    
 backout_L_fifo:
    arucoFifosReady = 0;
    fifoDestroy(arucoCapFifoL);
    
 backout_base:
    return(ccode);
}

void arucoUninit()
{
    arucoFifosReady = 0;
    arucoUninitOneSide(&arucoThreadDataL);
    arucoUninitOneSide(&arucoThreadDataR);
    fifoDestroy(arucoCapFifoR);
    fifoDestroy(arucoCapFifoL);
}

char buffer[1024*1024*4];
void colorMapFlip(void *data, uint32_t width, uint32_t height);

//#define ARUCO_DETECT_CONVERT_MANUALLY

////////////////////////////////////////////////////////////////////////////
// updateTable
////////////////////////////////////////////////////////////////////////////
void updateTable(std::vector<std::vector<Point2f> > &corners,
                 std::vector<int> &ids,
                 arucoPosDef *tags)
{
    int N, i, id, j;

    // Mark all tags at new insertion point in table as invalid.
    memset(tags, 0, sizeof(arucoPosDef)*ARUCO_MAX_IDS);
        
    // Extract detected corners
    N = ids.size();
    for (i=0; i<N; i++)
    {
        id = ids[i];
        if (id >= ARUCO_MAX_IDS) continue;
        
        tags[id].valid = 1;
        
        for (j=0; j<4; j++)
        {
            tags[id].x[j] = (float)corners[i][j].x;
            tags[id].y[j] = (float)(arucoCapHeight-1)-((float)corners[i][j].y);
        }
    }
}

////////////////////////////////////////////////////////////////////////////
// arucoThreadFunc
//      Thread that detects aruco tags in the background.
////////////////////////////////////////////////////////////////////////////
void arucoThreadFunc(void *parm)
{
    arucoThreadDataDef *data = (arucoThreadDataDef *) parm;
    arucoFrameDef *cap_data;
    Mat cap_frame_bayer(arucoCapHeight, arucoCapWidth, CV_8UC1, Scalar(0));
    Mat gray_frame(arucoCapHeight, arucoCapWidth, CV_8UC1, Scalar(0));
    Mat gray1_frame(arucoCapHeight, arucoCapWidth, CV_8UC1, Scalar(0));
    struct timespec t;

#if defined(CAPTURE_SEND_ARUCO)
    // When taken from capture frame, images are raw bayer / distorted images
    uint8_t *cam_data = cap_frame_bayer.ptr<uchar>(0);
#elif defined(PROCESS_SEND_ARUCO)
    // When taken from process thread, camera data is already bayer mapped and undistorted
    uint8_t *cam_data = gray_frame.ptr<uchar>(0);
#endif
    
    std::vector<int> ids;
    std::vector<std::vector<Point2f> > corners;

    profCpuStart(&(data->p.profAll));
    profWallStart(&(data->p.profWAll));
    data->p.numFrames = 0;

    while ((!data->threadStop) && (!stopping))
    {
        profCpuStart(&(data->p.fifoWait));
        if ((cap_data = (arucoFrameDef *) fifoReadPtr(data->fifo)) == 0)
        {
            profCpuEnd(&(data->p.fifoWait), 0);
            WAIT_SHORT;
            continue;
        }
        profCpuEnd(&(data->p.fifoWait), 0);

        profCpuStart(&(data->p.deinterlace));
        memcpy(&t, &(cap_data->t), sizeof(t));
        memcpy(cam_data, cap_data->data, arucoCapHeight*arucoCapWidth);
        
        fifoReadDone(data->fifo);
        profCpuEnd(&(data->p.deinterlace), 0);
        
#if defined(CAPTURE_SEND_ARUCO)
        // Convert Bayer to GRAY and possibly undistort
        profCpuStart(&(data->p.convert));
        if (data->uFunc)
        {
            cvtColor(cap_frame_bayer, gray1_frame, COLOR_BayerGB2GRAY);
            (*(data->uFunc))(gray1_frame, gray_frame);
        }
        else
        {
            cvtColor(cap_frame_bayer, gray_frame, COLOR_BayerGB2GRAY);
        }
        profCpuEnd(&(data->p.convert), 0);
#endif
        profCpuStart(&(data->p.detect));
        
        // Detect tags
        aruco::detectMarkers(gray_frame, dictionary, corners, ids);
        profCpuEnd(&(data->p.detect), 0);
        profCpuStart(&(data->p.updateTable));
        
        // Only allow single-thread access to tags structure
        pthread_mutex_lock(&(data->mutex));
        updateTable(corners, ids, data->tags[data->bufferIdx]);
        data->bufferIdx = (data->bufferIdx + 1) & ARUCO_BUFFER_DEPTH_MASK;
        pthread_mutex_unlock(&(data->mutex));

        profCpuEnd(&(data->p.updateTable), 0);

        data->p.numFrames++;
    }

    profCpuEnd(&(data->p.profAll), 0);
    profWallEnd(&(data->p.profWAll));
    
    data->threadDone = 1;
}

////////////////////////////////////////////////////////////////////////////
// arucoGetTagCorners
//      Predict tag corner positions based on detected values in
//      the buffer.
////////////////////////////////////////////////////////////////////////////
void arucoGetTagCorners(int id,
                        arucoPosDef *pos_L, arucoPosDef *pos_R,
                        struct timespec *t)
{
    arucoThreadDataDef *data;
    arucoPosDef *pos;
    uint32_t idx;
    
    // Returns the estimated corners of an Aruco tag based on
    // samples in the circular buffer.

    // For now, just return are most recent sample.  Later we can do
    // a more advanced tracking thing if lag is a problem.

    // Left
    data = &arucoThreadDataL;
    pthread_mutex_lock(&(data->mutex));
    idx = (data->bufferIdx - 1) & ARUCO_BUFFER_DEPTH_MASK;
    pos = &(data->tags[idx][id]);
    memcpy(pos_L, pos, sizeof(arucoPosDef));
    pthread_mutex_unlock(&(data->mutex));

    // Right
    data = &arucoThreadDataR;
    pthread_mutex_lock(&(data->mutex));
    idx = (data->bufferIdx - 1) & ARUCO_BUFFER_DEPTH_MASK;
    pos = &(data->tags[idx][id]);
    memcpy(pos_R, pos, sizeof(arucoPosDef));
    pthread_mutex_unlock(&(data->mutex));
}


