/////////////////////////////////////////////////////////////////////////////
// capture.cc
//      Code to support synchronous capture from X11 and stereo camera.
//
// Author: Jon W. Wallace, Lafayette College
// Date:   May 2020
////////////////////////////////////////////////////////////////////////////

//#include <ovrvision_pro.h>
#include <inttypes.h>
#include <pthread.h>
#include <math.h>
#include <string.h>

#include "log.h"
#include "grab_x11.h"
#include "myProfile.h"
#include "fifo.h"
#include "main.h"
#include "ar_parm.h"
#include "gauges.h"
#include "v4l_cap.h"
#include "aruco.h"

// Uncomment this to generate a file of input sample times to
// analyze jitter.
//#define DUMP_INPUT_JITTER
static const char *fname_jitter = "in_jitter.txt";
static FILE *f_jitter;

//#define DUMP_FRAMES_FILE
static const char *fname_frames = "cap_frames.bin";
static FILE *f_frames;

// Supported modes of the OVR Camera
#define MAX_OVR_MODES 7
struct
{
    int width;
    int height;
    int fps;
} ovr_modes[MAX_OVR_MODES] =
{
  { 2560, 1920, 15 }, // OV_CAM5MP_FULL
  { 1920, 1080, 30 }, // OV_CAM5MP_FHD,  
  { 1280, 960,  45 }, // OV_CAMHD_FULL,  
  { 960,  950,  60 }, // OV_CAMVR_FULL,  
  { 1280, 800,  60 }, // OV_CAMVR_WIDE,  
  { 640,  480,  90 }, // OV_CAMVR_VGA,   
  { 320,  240,  120 } // OV_CAMVR_QVGA,  
};

#define NUM_GRABX11_FRAMES     4
#define NUM_CAMCAP_FRAMES      4

// Profile variables for X11 grab thread
typedef struct
{
    cpuProfDef fifoWait;
    cpuProfDef grabX11;
    cpuProfDef copyFrame;
    cpuProfDef profAll;
    wallProfDef profWAll;
    uint32_t numFrames;
} profX11GrabDef;
profX11GrabDef profX11Grab;

// Profile variables for OVR capture thread
typedef struct
{
    cpuProfDef fifoWait;
    cpuProfDef captureOvr;
    cpuProfDef sendAruco;
    cpuProfDef detectTags;
    cpuProfDef skipFrame;
    cpuProfDef profAll;
    wallProfDef profWAll;
    uint32_t numFrames;

    // Capture jitter measurement
    double dt_mean, dt_mom2;
    uint32_t N_jitter;
    
} profCamCaptureDef;
profCamCaptureDef profCamCapture;

// OVR Processing mode
// OVR::Camqt g_processMode = OVR::OV_CAMQT_DMSRMP;
//OVR::Camqt g_processMode = OVR::OV_CAMQT_DMS;

// Capture data
typedef struct
{
    // OVR Capture parameters
    int camWidth;
    int camHeight;
    int camSize;
    int mode;
    pthread_t camThread;
    
    // X11 Grab Parameters
    int x11Width;
    int x11Height;
    int x11Size;
    int x11Wid;
    pthread_t x11Thread;

    // Capture control
    volatile int stop;
    volatile int x11Status;
    volatile int camStatus;
    pthread_barrier_t barrier;
    
} captureDef;
captureDef cap;

// FIFOs for X11 and camera frames
fifoDef *captureX11Fifo;
fifoDef *captureCamFifo;

////////////////////////////////////////////////////////////////////////////
// captureProfilePrint
//      Prints the results of profiling for the capture thread.
////////////////////////////////////////////////////////////////////////////
void captureProfilePrint()
{
    double dt_mean, dt_std;

    // Print the X11 grab statistics
    if (cap.x11Wid)
    {
        profX11GrabDef *p = &profX11Grab;

        log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
        log_printf(log_quiet, "X11 Capture Thread Time                                    | Output Rate\n");
        log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
        profPrintOne("FIFO Wait", &(p->fifoWait), &(p->profAll), p->numFrames);
        profPrintOne("grab X11", &(p->grabX11), &(p->profAll), p->numFrames);
        profPrintOne("Copy Frame", &(p->copyFrame), &(p->profAll), p->numFrames);
        log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
        profPrintOne("Total", &(p->profAll), &(p->profAll), p->numFrames);
        log_printf(log_quiet, "%-20s %8.2f\n", "CPU FPS", (double)p->numFrames/profTimeSec(&p->profAll));
        log_printf(log_quiet, "%-20s %8.2f\n", "FPS (Wall)", (double)p->numFrames/profTimeSec(&p->profWAll));
        log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
    }
    
    // Print the OVR capture statistics
    {
        profCamCaptureDef *p = &profCamCapture;
    
        // Get mean and standard deviation of capture samples to see jitter.
        dt_mean = p->dt_mean/(double)p->N_jitter;
        dt_std = sqrtf(p->dt_mom2/(double)p->N_jitter - dt_mean*dt_mean);
        
        log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
        log_printf(log_quiet, "OVR Capture Thread Time                                    | Output Rate\n");
        log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
        profPrintOne("FIFO Wait", &(p->fifoWait), &(p->profAll), p->numFrames);
        profPrintOne("Capture OVR", &(p->captureOvr), &(p->profAll), p->numFrames);
        profPrintOne("Send Aruco", &(p->sendAruco), &(p->profAll), p->numFrames);
        profPrintOne("Detect Tags", &(p->detectTags), &(p->profAll), p->numFrames);
        log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
        profPrintOne("Total", &(p->profAll), &(p->profAll), p->numFrames);
        log_printf(log_quiet, "%-20s %8.2f\n", "CPU FPS", (double)p->numFrames/profTimeSec(&p->profAll));
        log_printf(log_quiet, "%-20s %8.2f\n", "FPS (Wall)", (double)p->numFrames/profTimeSec(&p->profWAll));
        log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
        log_printf(log_quiet, "Input Jitter (mean / std): %.3f ms / %.3f ms\n", dt_mean*1000.0, dt_std*1000.0);
        log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
    }
}

////////////////////////////////////////////////////////////////////////////
// captureInit
//      Initialize the capture thread.
////////////////////////////////////////////////////////////////////////////
int captureInit()
{
    int ccode;
    int i;
    int locationID = 0;
    uint32_t data_size;
    
    // Initialize profile variables (setting to 0 works)
    memset(&profX11Grab, 0, sizeof(profX11GrabDef));
    memset(&profCamCapture, 0, sizeof(profCamCaptureDef));

    // Zero out the capture structure to start
    memset(&cap, 0, sizeof(captureDef));

    // DEBUG: Set the mapping so whole capture frame is mapped to the
    // X window.
    //ar_parm_make_ideal_map(mainOptions.cap_width, mainOptions.cap_height);

    // Set up the capture -> VR texture points
    ar_parm_get_s_rend(mainOptions.cap_width, mainOptions.cap_height);
    
    // Initialize X11 grab variables
    if (mainOptions.x_wid)
    {
        cap.x11Wid = mainOptions.x_wid;
        if (grabX11Init(cap.x11Wid) != 0)
        {
            ccode = -1;
            goto backout_base;
        }
        cap.x11Width = grabX11Width();
        cap.x11Height = grabX11Height();
        //fprintf(stdout, "ID 0x%x: %d x %d\n", cap.x11Wid, cap.x11Width, cap.x11Height);
    }
    
    // Figure out which OVR capture mode matches.
    cap.camWidth = mainOptions.cap_width;
    cap.camHeight = mainOptions.cap_height;
    cap.mode = -1;
    
    // Create FIFO for outputting camera capture frames (assumes RGBA, which is the
    // largest format we could have).
    cap.camSize = cap.camWidth*cap.camHeight*2*4;
    if (fifoCreate(cap.camSize, NUM_CAMCAP_FRAMES, &captureCamFifo) != 0)
    {
        log_printf(log_error, "Unable to create camera capture FIFO.\n");
        ccode = -3;
        goto backout_x11;
    }

    // Create FIFO for outputting X11 capture frames
    cap.x11Size = cap.x11Width*cap.x11Height*4;
    if (fifoCreate(cap.x11Size, NUM_GRABX11_FRAMES, &captureX11Fifo) != 0)
    {
        log_printf(log_error, "Unable to create grab_x11 FIFO.\n");
        ccode = -3;
        goto backout_cam_fifo;
    }

    // Initialize Camera
    if (v4l_cap_init(mainOptions.cap_dev, cap.camWidth, cap.camHeight, 30, 0) != 0)
    {
        ccode = -4;
        log_printf(log_error, "Unable to open OVR camera.\n");
        goto backout_x11_fifo;
    }

    // Create a barrier function for X11 Grab / Capture sync
    pthread_barrier_init(&cap.barrier, NULL, 2);

    // Success
    return(0);

 backout_cap:
    v4l_cap_uninit();

 backout_x11_fifo:
     if (cap.x11Wid) fifoDestroy(captureX11Fifo);
     
 backout_cam_fifo:
     fifoDestroy(captureCamFifo);
     
 backout_x11:
     if (cap.x11Wid) grabX11Uninit();
     
 backout_base:
     return(ccode);
}

////////////////////////////////////////////////////////////////////////////
// captureUninit()
//      Frees up structures associated with capturing.
////////////////////////////////////////////////////////////////////////////
void captureUninit()
{
    v4l_cap_uninit();
    if (cap.x11Wid) fifoDestroy(captureX11Fifo);
    fifoDestroy(captureCamFifo);
    if (cap.x11Wid) grabX11Uninit();
}

////////////////////////////////////////////////////////////////////////////
// captureDeinterlaceFrame
//      Converts the interlaced OVR format to L-top, R-bottom frames.
//      Input is Bayer format, so one byte per pixel.
////////////////////////////////////////////////////////////////////////////
void captureDeinterlaceFrame(uint8_t *data, uint8_t *data_out_l, uint8_t *data_out_r, uint32_t width, uint32_t height)
{
    uint32_t i, j;

    for (j=0; j<height; j++)
    {
        for (i=0; i<width; i++)
        {
            data_out_l[i + j*width] = data[2*(i + j*width)];
            data_out_r[i + j*width] = data[2*(i + j*width)+1];
        }
    }
}

////////////////////////////////////////////////////////////////////////////
// camThreadFunc
//      Thread to get frames from camera and put them into a FIFO.
////////////////////////////////////////////////////////////////////////////
void camThreadFunc()
{
    void *fifo_rec;
    unsigned char *pl, *pr;

    double t0, t1, dt;
    double t_start;
    int N;
    int j;
    int bytes_used;

    struct timespec t;
    arucoFrameDef *aruco_rec_l, *aruco_rec_r;

    t0 = 0.0;
    t1 = 0.0;
    N = 0;

    t_start = profWallTime();
    
#ifdef DUMP_INPUT_JITTER
    f_jitter = fopen(fname_jitter, "wb");
#endif
#ifdef DUMP_FRAMES_FILE
    f_frames = fopen(fname_frames, "wb");
#endif    
    profCpuStart(&(profCamCapture.profAll));
    profWallStart(&(profCamCapture.profWAll));
    profCamCapture.numFrames = 0;

    for ( ; ; )
    {
        profCpuStart(&(profCamCapture.fifoWait));

        for ( ; ; )
        {
            /* Check if we should abort. */
            if (cap.stop)
            {
                profCpuEnd(&(profCamCapture.fifoWait), 0);
                goto backout_cap;
            }
            
            /* Try to get write buffer from FIFO. */
            if ((fifo_rec = fifoWritePtr(captureCamFifo)) == 0)
            {
                /* Full, so wait. */
                WAIT_SHORT;
            }
            else
            {
                break;
            }
        }
        profCpuEnd(&(profCamCapture.fifoWait), 0);

        profCpuStart(&(profCamCapture.captureOvr));

        // Get a frame from the camera
        v4l_cap_get_frame(fifo_rec, cap.camSize, &bytes_used);
        if (profCamCapture.numFrames == 4)
        {
            log_printf(log_quiet, "Bytes per frame: %d\n", bytes_used);
        }

        // Get timestamp for frame
        clock_gettime(CLOCK_REALTIME, &t);

#ifdef DUMP_FRAMES_FILE
        if ((profCamCapture.numFrames % 30) == 0)
        {
            fwrite(fifo_rec, bytes_used, 1, f_frames);
        }
#endif        

        t0 = t1;
        t1 = profWallTime();
#ifdef DUMP_INPUT_JITTER
        fprintf(f_jitter, "%.9e\n", t1-t_start);
#endif
        // Measure jitter in the input stream.  Note that we wait for
        // some frames to pass before we start measuring.
        if (profCamCapture.numFrames > 10)
        {
            dt = t1-t0;
            profCamCapture.dt_mean += dt;
            profCamCapture.dt_mom2 += dt*dt;
            profCamCapture.N_jitter++;
        }
        
        profCpuEnd(&(profCamCapture.captureOvr), cap.camSize);

        // Send data to Aruco detect threads as well
#ifdef CAPTURE_SEND_ARUCO
        if ((mainOptions.inst_panel) && (arucoFifosReady))
        {
            if ((!fifoIsFull(arucoCapFifoL)) && (!fifoIsFull(arucoCapFifoR)))
            {
                profCpuStart(&(profCamCapture.sendAruco));
                aruco_rec_l = (arucoFrameDef *) fifoWritePtr(arucoCapFifoL);
                aruco_rec_r = (arucoFrameDef *) fifoWritePtr(arucoCapFifoR);
                
                memcpy(&aruco_rec_l->t, &t, sizeof(t));
                memcpy(&aruco_rec_r->t, &t, sizeof(t));
                
                captureDeinterlaceFrame((uint8_t *) fifo_rec, aruco_rec_l->data, aruco_rec_r->data,
                                        cap.camWidth, cap.camHeight);
                fifoWriteDone(arucoCapFifoL);
                fifoWriteDone(arucoCapFifoR);
                profCpuEnd(&(profCamCapture.sendAruco), 0);
            }
        }
#endif
        
        // Indicate to FIFO that we are done with record
        FIFO_PAYLOAD_SIZE(fifo_rec) = cap.camSize;
        fifoWriteDone(captureCamFifo);
        
        // Increment profile frame counter
        profCamCapture.numFrames++;

        // Detect Gauges 
        profCpuStart(&(profCamCapture.detectTags));
        profCpuEnd(&(profCamCapture.detectTags), 0);

        // Skip specified number of frames
        profCpuStart(&(profCamCapture.skipFrame));
        for (j=0; j<mainOptions.skip_time; j++)
        {
            v4l_cap_get_frame(0, cap.camSize, 0);
        }
        profCpuEnd(&(profCamCapture.skipFrame), 0);
        
        // Synchronize with the X11 capture thread to stay in sync
        // if there is a capture window.
        if (cap.x11Wid)
        {
            pthread_barrier_wait(&cap.barrier);
        }
    }

#ifdef DUMP_INPUT_JITTER
    fclose(f_jitter);
#endif
#ifdef DUMP_FRAMES_FILE
    fclose(f_frames);
#endif
 backout_cap:
    
    profCpuEnd(&(profCamCapture.profAll), 0);
    profWallEnd(&(profCamCapture.profWAll));
    
    cap.camStatus = 1;
}

////////////////////////////////////////////////////////////////////////////
// x11ThreadFunc
//      Thread to grab frames from X11 window and put in FIFO.
////////////////////////////////////////////////////////////////////////////
void x11ThreadFunc()
{
    void *fifo_rec;
    unsigned char *pixels;
    
    profCpuStart(&(profX11Grab.profAll));
    profWallStart(&(profX11Grab.profWAll));
    profX11Grab.numFrames = 0;

    for ( ; ; )
    {
        profCpuStart(&(profX11Grab.fifoWait));

        for ( ; ; )
        {
            /* Check if we should abort. */
            if (cap.stop)
            {
                profCpuEnd(&(profX11Grab.fifoWait), 0);
                goto backout_x11;
            }
            
            /* Try to get write buffer from FIFO. */
            if ((fifo_rec = fifoWritePtr(captureX11Fifo)) == 0)
            {
                /* Full, so wait. */
                WAIT_SHORT;
            }
            else
            {
                break;
            }
        }
        profCpuEnd(&(profX11Grab.fifoWait), 0);

        profCpuStart(&(profX11Grab.grabX11));
        pixels = grabX11Image(cap.x11Wid, 1); // 1=rawMode
        profCpuEnd(&(profX11Grab.grabX11), 0);
 
        // Get a frame from the camera
        profCpuStart(&(profX11Grab.copyFrame));
        if (pixels == 0)
        {
            memset(fifo_rec, 0, cap.x11Size);
        }
        else
        {
            memcpy(fifo_rec, pixels, cap.x11Size);
        }
        
        // Indicate to FIFO that we are done with record
        FIFO_PAYLOAD_SIZE(fifo_rec) = cap.x11Size;
        fifoWriteDone(captureX11Fifo);

        profCpuEnd(&(profX11Grab.copyFrame), cap.x11Size);

        // Increment profile frame counter
        profX11Grab.numFrames++;

        // Synchronize with the X11 capture thread to stay in sync
        pthread_barrier_wait(&cap.barrier);
    }

 backout_x11:    

    profCpuEnd(&(profX11Grab.profAll), cap.x11Size*profX11Grab.numFrames);
    profWallEnd(&(profX11Grab.profWAll));
    
    cap.x11Status = 1;
}

////////////////////////////////////////////////////////////////////////////
// captureStart
//      Starts the capture thread.
////////////////////////////////////////////////////////////////////////////
int captureStart()
{
    int ccode;
    
    if (pthread_create(&(cap.camThread),
                       NULL,
                       (void * (*)(void *)) camThreadFunc,
                       (void *) 0) != 0)
    {
        log_printf(log_error, "Unable to start camera capture thread.");
        ccode = -1;
        goto backout_base;
    }

    if (cap.x11Wid)
    {
        if (pthread_create(&(cap.x11Thread),
                           NULL,
                           (void * (*)(void *)) x11ThreadFunc,
                           (void *) 0) != 0)
        {
            log_printf(log_error, "Unable to start X11 grab thread.");
            ccode = -1;
            goto backout_cam_thread;
        }
    }

    // Success
    return(0);
    
 backout_cam_thread:
    pthread_cancel(cap.camThread);
    
 backout_base:
    return(ccode);
}

////////////////////////////////////////////////////////////////////////////
// captureStop
//      Stop the capturing threads and waits for them to exit.
////////////////////////////////////////////////////////////////////////////
void captureStop()
{
    cap.stop = 1;
    while (((cap.x11Wid != 0) && (cap.x11Status == 0)) || (cap.camStatus == 0))
    {
        WAIT_SHORT;
    }
}




