////////////////////////////////////////////////////////////////////////////
// cudaRender.cc
//      Code to support processing of data on CUDA hardware to
//      provide better performance and lower latency.
//
// Author: Jon W. Wallace, Lafayette College
// Date:   May 2020
////////////////////////////////////////////////////////////////////////////

#include <GL/glew.h>
#include <GLFW/glfw3.h>

#include <inttypes.h>
#include <cuda_gl_interop.h>
#include <stdio.h>

#include "main.h"
#include "log.h"
#include "grab_x11.h"
#include "renderState.h"

#include  <thrust/complex.h>

// Threads in a block for CUDA processing
#define CUDA_BLOCKSIZE          128
#define CUDA_BLOCKSIZE_2D       32

static uint32_t cam_width, cam_height, cam_size;
static uint32_t x11_width, x11_height, x11_size;

typedef struct
{
    float h, s, v, c;
} hsv_def;

static void *hostX11Frame, *gpuX11Frame;
static hsv_def *gpuHsvFrame;

void *gpuCamFrame;
void *gpuCamRawFrameL, *gpuCamRawFrameR;

static cudaGraphicsResource *gpuCamFrameOutResource;
static cudaGraphicsResource *gpuX11FrameOutResource;

GLuint buffer_cam, buffer_x11;
GLuint tex_cam, tex_x11;

#define COS(x)          (cosf(x))
#define SIN(x)          (sinf(x))
#define EXPJ(x)         (Cplx(COS(x), SIN(x)))
typedef thrust::complex<float> Cplx;

extern renderStateDef renderState;

void *crHostBuffer;

int crInit(GLuint tex_cam_in, GLuint tex_x11_in)
{
    uint32_t num_devices;
    int devices[16];
    int ccode;
    
    hostX11Frame = 0;
    gpuCamFrame = 0;
    gpuX11Frame = 0;
    gpuCamRawFrameL = 0;
    gpuCamRawFrameR = 0;
    crHostBuffer = 0;
    
    if (cudaGLGetDevices(&num_devices, devices, sizeof(devices)/sizeof(int), cudaGLDeviceListAll) !=
        cudaSuccess)
    {
        log_printf(log_error, "Unable to get CUDA OpenGL devices.\n");
        ccode = -1;
        goto backout_base;
    }
    if (num_devices < 1)
    {
        log_printf(log_error, "No CUDA OpenGL device found.\n");
        ccode = -1;
        goto backout_base;
    }
    cudaSetDevice(devices[0]);

    cam_width = mainOptions.cap_width;
    cam_height = mainOptions.cap_height*2;
    cam_size = cam_width*cam_height*4;

    if (mainOptions.x_wid)
    {
        x11_width = grabX11Width();
        x11_height = grabX11Height();
        x11_size = x11_width*x11_height*4;
    }
    else
    {
        x11_size = 0;
    }

    // Allocate buffers needed for transfer of data from CUDA => OpenGL
    glCreateBuffers(1, &buffer_cam);
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, buffer_cam);
    glBufferData(GL_PIXEL_UNPACK_BUFFER, cam_size, NULL, GL_DYNAMIC_COPY);
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0); // !!!
    
    if (mainOptions.x_wid)
    {
        glCreateBuffers(1, &buffer_x11);
        glBindBuffer(GL_PIXEL_UNPACK_BUFFER, buffer_x11);
        glBufferData(GL_PIXEL_UNPACK_BUFFER, x11_size, NULL, GL_DYNAMIC_COPY);
        glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0); // !!!
    }
    
    if (cudaMalloc(&gpuCamFrame, cam_size) != cudaSuccess)
    {
        log_printf(log_error, "Unable to allocate % bytes for cuda gpu frame.\n", cam_size);
        ccode = -2;
        goto backout_cuda_mem;
    }

    // Allocate a buffer to use for conversion of RGB to HSV.  The
    // latter is used for chroma key detection.
    if (cudaMalloc(&gpuHsvFrame, cam_width*cam_height*sizeof(hsv_def)) != cudaSuccess)
    {
        log_printf(log_error, "Unable to allocate gpu HSV frame.\n");
        ccode = -2;
        goto backout_cuda_mem;
    }

    // Allocate raw (Bayer format) L and R frames deinterlaced frames
    if ((cudaMalloc(&gpuCamRawFrameL, cam_width*cam_height) != cudaSuccess) ||
        (cudaMalloc(&gpuCamRawFrameR, cam_width*cam_height) != cudaSuccess))
    {
        log_printf(log_error, "Unable to allocate deinterlaced L/R raw GPU frames.\n");
        ccode = -2;
        goto backout_cuda_mem;
    }

    // Allocate space for host buffer transfers.
    if (cudaMallocHost(&crHostBuffer, 2*4*cam_width*cam_height) != cudaSuccess)
    {
        log_printf(log_error, "Unable to allocate cuda host buffer frame.\n");
        ccode = -2;
        goto backout_cuda_mem;
    }

    if (x11_size)
    {
        if (cudaMalloc(&gpuX11Frame, x11_size) != cudaSuccess)
        {
            log_printf(log_error, "Unable to allocate cuda GPU X11 frame.\n");
            ccode = -2;
            goto backout_cuda_mem;
        }
    }

    // Get resource objects needed to write into the OpenGL textures from CUDA
    tex_cam = tex_cam_in;
    tex_x11 = tex_x11_in;

    //if (cudaGraphicsGLRegisterImage(&gpuCamFrameOutResource, tex_cam, GL_TEXTURE_2D, cudaGraphicsRegisterFlagsWriteDiscard) !=
    //cudaSuccess)
    if (cudaGraphicsGLRegisterBuffer(&gpuCamFrameOutResource, buffer_cam, cudaGraphicsRegisterFlagsWriteDiscard) != cudaSuccess)
    {
        log_printf(log_error, "Unable to register cuda cam_buffer.\n");
        ccode = -2;
        goto backout_cuda_mem;
    }
    
    // Get resource objects needed to write into the OpenGL textures from CUDA
    if ((x11_size) && (tex_x11))
    {
        //if (cudaGraphicsGLRegisterImage(&gpuX11FrameOutResource, tex_x11, GL_TEXTURE_2D, cudaGraphicsRegisterFlagsWriteDiscard) !=
        //cudaSuccess)
        if (cudaGraphicsGLRegisterBuffer(&gpuX11FrameOutResource, buffer_x11, cudaGraphicsRegisterFlagsWriteDiscard) != cudaSuccess)
        {
            log_printf(log_error, "Unable to register cuda image for tex_x11.\n");
            ccode = -2;
            goto backout_cuda_mem;
        }
    }
    return(0);

 backout_cuda_mem:
    if (gpuCamFrame) cudaFree(gpuCamFrame);
    if (hostX11Frame) cudaFreeHost(hostX11Frame);
    if (gpuX11Frame) cudaFree(gpuX11Frame);
    if (gpuCamRawFrameL) cudaFree(gpuCamRawFrameL);
    if (gpuCamRawFrameR) cudaFree(gpuCamRawFrameR);
    if (crHostBuffer) cudaFreeHost(crHostBuffer);
    
 backout_base:
    return(ccode);
}

////////////////////////////////////////////////////////////////////////////
// crLoadCamRaw
//      Loads the cam_frame data in Bayer format down to the GPU.
////////////////////////////////////////////////////////////////////////////
void crLoadCamRaw(void *cam_frame)
{
    cudaMemcpy(gpuCamFrame, cam_frame, 2*cam_width*cam_height, cudaMemcpyHostToDevice);
}

////////////////////////////////////////////////////////////////////////////
// crLoadX11Frame
//      Copies an X11 captured frame down to the GPU.
////////////////////////////////////////////////////////////////////////////
void crLoadX11Frame(void *x11_frame)
{
    if ((gpuX11Frame) && (x11_frame))
    {
        cudaMemcpy(gpuX11Frame, x11_frame, x11_size, cudaMemcpyHostToDevice);
    }
}

////////////////////////////////////////////////////////////////////////////
// deinterlace_image_k
//      Simple kernel to deinterlace the stereo L and R Bayer images.
////////////////////////////////////////////////////////////////////////////
__global__ void deinterlace_image_k(uint8_t *raw_data, uint8_t *raw_data_l, uint8_t *raw_data_r, uint32_t width, uint32_t height)
{
    int i_idx = blockIdx.x * blockDim.x + threadIdx.x;
    int i_stride = blockDim.x * gridDim.x;
    
    for (int i=i_idx; i<width*height; i+=i_stride)
    {
        raw_data_l[i] = raw_data[i*2+0];
        raw_data_r[i] = raw_data[i*2+1];
    }
}

////////////////////////////////////////////////////////////////////////////
// crDeinterlace
//      Deinterlaces the raw image which was previously loaded using
//      crLoadCamRaw().
////////////////////////////////////////////////////////////////////////////
void crDeinterlace()
{
    uint32_t N_blocks;

    N_blocks = (cam_width*cam_height + CUDA_BLOCKSIZE - 1)/CUDA_BLOCKSIZE;
    deinterlace_image_k<<<N_blocks, CUDA_BLOCKSIZE>>>((uint8_t *) gpuCamFrame, (uint8_t *) gpuCamRawFrameL, (uint8_t *) gpuCamRawFrameR,
                                                      cam_width, cam_height);
    cudaDeviceSynchronize();
}

__global__ void remap_image_k(void *frame_data, void *frame_data_out, uint32_t width, uint32_t height)
{
    uint32_t *bgra_data = (uint32_t *) frame_data;
    uint32_t *bgra_data_out = (uint32_t *) frame_data_out;
    uint32_t p1, p2, i, j;
    
    // Get pixel ID
    int i_idx = blockIdx.x * blockDim.x + threadIdx.x;
    int i_stride = blockDim.x * gridDim.x;
    int j_idx = blockIdx.y * blockDim.y + threadIdx.y;
    int j_stride = blockDim.y * gridDim.y;

    for (j=j_idx; j<height; j+=j_stride)
    {
        for (i=i_idx; i<width; i+=i_stride)
        {
            // Pixel from top
            p1 = bgra_data[i + j*width];
            p2 =
                (p1 & 0xff00ff00) |
                ((p1 & 0x00ff0000) >> 16) |
                ((p1 & 0x000000ff) << 16);
            bgra_data_out[i + (height-1-j)*width] = p2;
        }
    }
}

////////////////////////////////////////////////////////////////////////////
// get_image_hsv_k
//      Converts image pixels from RGB[A] to HSV
////////////////////////////////////////////////////////////////////////////
__global__ void get_image_hsv_k(uint32_t *rgba_data, hsv_def *frame_hsv_data, uint32_t width, uint32_t height)
{
    // Get pixel ID
    int i_idx = blockIdx.x * blockDim.x + threadIdx.x;
    int i_stride = blockDim.x * gridDim.x;
    int j_idx = blockIdx.y * blockDim.y + threadIdx.y;
    int j_stride = blockDim.y * gridDim.y;
    int i, j;
    float R, G, B, H, S, V, max_color, min_color;
    
    for (j=j_idx; j<height; j+=j_stride)
    {
        for (i=i_idx; i<width; i+=i_stride)
        {
            R = (float)(rgba_data[i + j*width] & 0x000000ff)/255.0;
            G = (float)((rgba_data[i + j*width] & 0x0000ff00) >> 8)/255.0;
            B = (float)((rgba_data[i + j*width] & 0x00ff0000) >> 16)/255.0;

            max_color = fmaxf(fmaxf(R, G), B);
            min_color = fminf(fminf(R, G), B);

            if (max_color == 0.0)
            {
                H = 0;
                S = 0;
                V = 0;
            }
            else
            {
                H = arg(Cplx(2.0*R - G - B, sqrtf(3.0)*(G-B)))*180.0/M_PI;
                //H = arg(Cplx(2.0*R - G - B, sqrt(3.0)*(G-B)*EXPJ(-2.0*M_PI*h_center/180.0)))*180.0/(2.0*M_PI);
                V = (R+G+B)*0.33333333333333333333;
                S =  (max_color - min_color) / (max_color + min_color);
            }
            frame_hsv_data[i + j*width].h = H;
            frame_hsv_data[i + j*width].s = S;
            frame_hsv_data[i + j*width].v = V;
            // Chroma key flag set to 0.  Filled in later            
            frame_hsv_data[i + j*width].c = 0.0;  
        }
    }
}

////////////////////////////////////////////////////////////////////////////
// find_green1_k
//      Detects green (chroma key) pixels.
////////////////////////////////////////////////////////////////////////////
__global__ void find_green1_k(hsv_def *hsv_data, uint32_t width, uint32_t height, float hue_min, float hue_max, float sat_min)
{
    float hue_avg = 0.5*(hue_min + hue_max);
    float hue_dev = 0.5*(hue_max - hue_min);
    float H, S, H_shift;
    // float V;
    
    // Get pixel ID
    int i_idx = blockIdx.x * blockDim.x + threadIdx.x;
    int i_stride = blockDim.x * gridDim.x;
    int j_idx = blockIdx.y * blockDim.y + threadIdx.y;
    int j_stride = blockDim.y * gridDim.y;
    int i, j;
    
    for (j=j_idx; j<height; j+=j_stride)
    {
        for (i=i_idx; i<width; i+=i_stride)
        {
            H = hsv_data[i + j*width].h;
            S = hsv_data[i + j*width].s;
            //V = hsv_data[i + j*width].v;

            H_shift = arg(EXPJ(H*M_PI/180.0)*EXPJ(-hue_avg*M_PI/180.0))*180.0/M_PI;

            if ((fabs(H_shift) < hue_dev) && (S > sat_min))
            {
                // Green detected.  
                hsv_data[i + j*width].c = 1.0;
            }
            else
            {
                hsv_data[i + j*width].c = 0.0;
            }
        }
    }
}

////////////////////////////////////////////////////////////////////////////
// find_green2_k
//      Do a block sum and threshold to spatially filter the
//      green_screen detection values.
////////////////////////////////////////////////////////////////////////////
__global__ void find_green2_k(hsv_def *hsv_data, uint32_t *rgba_data, uint32_t width, uint32_t height, uint32_t rad)
{
    // Get pixel ID
    int i_idx = blockIdx.x * blockDim.x + threadIdx.x;
    int i_stride = blockDim.x * gridDim.x;
    int j_idx = blockIdx.y * blockDim.y + threadIdx.y;
    int j_stride = blockDim.y * gridDim.y;
    int i, j, ii, jj;
    float avg;
    int N_pix;
    
    for (j=j_idx; j<height; j+=j_stride)
    {
        for (i=i_idx; i<width; i+=i_stride)
        {
            // Do central block average
            avg = 0.0;
            N_pix = 0;
            for (jj=max(j-rad, 0); jj<min(j+rad,height-1); jj++)
            {
                for (ii=max(i-rad,0); ii<min(i+rad,width-1); ii++) 
                {
                    N_pix++;
                    avg += hsv_data[ii + jj*width].c;
                }
            }
            if ((i <= rad) || (avg > (0.4*(float)N_pix))) //(float)(rad*rad)))
            {
                // Green detected.  Make see through (ALPHA=0) 
                rgba_data[i + j*width] &= 0x00ffffff;
                // DEBUG: Make it white.
                // rgba_data[i + j*width] = 0xffffffff;
            }
            else
            {
                // Not green.  Leave opaque/transparent according to input.
                // Can uncomment this to force opaque.
                // rgba_data[i + j*width] |= 0xff000000;
            }
            // DEBUG
            //rgba_data[i + j*width] = (uint8_t)((H_shift+180.0)/360.0*255.0) | 0xff000000;
            //rgba_data[i + j*width] = (uint8_t)(S*255.0) | 0xff000000;
        }
    }
}

////////////////////////////////////////////////////////////////////////////
// crRemapX11Frame
//      Processes the grabbed X11 frame:
//      Does the GBRA to RGBA conversion, flip, and copies into an
//      openGL texture.
////////////////////////////////////////////////////////////////////////////
void crRemapX11Frame()
{
    void *gpuX11FrameOut;
    size_t size_obj;

    dim3 dimBlock((x11_width + CUDA_BLOCKSIZE_2D - 1) / CUDA_BLOCKSIZE_2D,
                  (x11_height + CUDA_BLOCKSIZE_2D - 1) / CUDA_BLOCKSIZE_2D);
    dim3 dimGrid(CUDA_BLOCKSIZE_2D, CUDA_BLOCKSIZE_2D);

    if (gpuX11Frame)
    {
        if (cudaGraphicsMapResources(1, &gpuX11FrameOutResource, 0) != cudaSuccess)
        {
            log_printf(log_error, "Failed to map CUDA resource for tex_x11.\n");
            return;
        }
    
        if (cudaGraphicsResourceGetMappedPointer(&gpuX11FrameOut, &size_obj, gpuX11FrameOutResource) != cudaSuccess)
        {
            cudaGraphicsUnmapResources(1, &gpuX11FrameOutResource, 0);
            log_printf(log_error, "Failed to get CUDA array pointer for tex_x11.\n");
            return;
        }
        remap_image_k<<<dimBlock, dimGrid>>>(gpuX11Frame, gpuX11FrameOut, x11_width, x11_height);
        cudaDeviceSynchronize();
        cudaGraphicsUnmapResources(1, &gpuX11FrameOutResource, 0);
        
        // Copy the buffer data into the texture
        glBindBuffer(GL_PIXEL_UNPACK_BUFFER, buffer_x11);
        glBindTexture(GL_TEXTURE_2D, tex_x11);
        glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, x11_width, x11_height, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    }
}

////////////////////////////////////////////////////////////////////////////
// crRemapFrames
//      Remaps GBRA to RGBA and flips y.
////////////////////////////////////////////////////////////////////////////
void crRemapFrames(uint8_t *gpuFrameLR)
{
    void *gpuCamFrameOut;
    size_t size_obj;

    dim3 dimBlock((cam_width + CUDA_BLOCKSIZE_2D - 1) / CUDA_BLOCKSIZE_2D,
                  (cam_height + CUDA_BLOCKSIZE_2D - 1) / CUDA_BLOCKSIZE_2D);
    dim3 dimGrid(CUDA_BLOCKSIZE_2D, CUDA_BLOCKSIZE_2D);

    if (cudaGraphicsMapResources(1, &gpuCamFrameOutResource, 0) != cudaSuccess)
    {
        log_printf(log_error, "Failed to map CUDA resource for tex_cam.\n");
        return;
    }
    
    if (cudaGraphicsResourceGetMappedPointer(&gpuCamFrameOut, &size_obj, gpuCamFrameOutResource) != cudaSuccess)
    {
        cudaGraphicsUnmapResources(1, &gpuCamFrameOutResource, 0);
        log_printf(log_error, "Failed to get CUDA array pointer for tex_cam.\n");
        return;
    }

    if (mainOptions.green_screen)
    {
        // Doing green-screen detection.

        // Get HSV version of camera image.
        get_image_hsv_k<<<dimBlock, dimGrid>>>((uint32_t *)gpuFrameLR, gpuHsvFrame, cam_width, cam_height);
        cudaDeviceSynchronize();

        // Now detect green pixels
        find_green1_k<<<dimBlock, dimGrid>>>(gpuHsvFrame, cam_width, cam_height,
                                            renderState.green_h_min, renderState.green_h_max, renderState.green_s_min);
        cudaDeviceSynchronize();

        // Copy original camera frame in
        cudaMemcpy(gpuCamFrameOut, gpuFrameLR, cam_size, cudaMemcpyDeviceToDevice);
        
        // Do block average
        find_green2_k<<<dimBlock, dimGrid>>>(gpuHsvFrame, (uint32_t *)gpuCamFrameOut, cam_width, cam_height, renderState.green_avg_rad);
        cudaDeviceSynchronize();
    }
    else
    {
        // Just copy frame into output texture.  Probably there is a way to avoid this
        // copy...
        cudaMemcpy(gpuCamFrameOut, gpuFrameLR, cam_size, cudaMemcpyDeviceToDevice);
    }
    cudaGraphicsUnmapResources(1, &gpuCamFrameOutResource, 0);
    
    // Now, copy the buffer data into the texture
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, buffer_cam);
    glBindTexture(GL_TEXTURE_2D, tex_cam);
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, cam_width, cam_height, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
}

void crUninit()
{
    if (gpuCamFrame) cudaFree(gpuCamFrame);
    if (hostX11Frame) cudaFreeHost(hostX11Frame);
    if (gpuX11Frame) cudaFree(gpuX11Frame);
    if (gpuCamRawFrameL) cudaFree(gpuCamRawFrameL);
    if (gpuCamRawFrameR) cudaFree(gpuCamRawFrameR);
    if (crHostBuffer) cudaFreeHost(crHostBuffer);
}


////////////////////////////////////////////////////////////////////////////
// Dot Finding Algorithm 
////////////////////////////////////////////////////////////////////////////

// Generic Maximization Algorithms

struct
{
    uint32_t *idx[2];
    float *val[2];
    uint32_t max_N;
} reduce_max_data_gpu;

////////////////////////////////////////////////////////////////////////////
// reduce_max_k()
//      Simple CUDA maximization reduction algorithm.  This kernel reduces
//      each block to a sigle element.  Needs to be executed multiple
//      times to reduce complete problem.
////////////////////////////////////////////////////////////////////////////
__global__ void reduce_max_k(float *val_in, uint32_t *idx_in, float *val_out, uint32_t *idx_out, int first, int N_elem)
{
    __shared__ float val_work[CUDA_BLOCKSIZE];
    __shared__ uint32_t idx_work[CUDA_BLOCKSIZE];

    // Copy from global memory to local block elements
    int idx_g = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx_g < N_elem)
    {
        val_work[threadIdx.x] = val_in[idx_g];
        // If first iteration is indicated, fill in the indices with 0,1,2 ...
        // otherwise, copy paired indices.
        if (first)
        {
            idx_work[threadIdx.x] = idx_g;
        }
        else
        {
            idx_work[threadIdx.x] = idx_in[idx_g];
        }
    }
    else
    {
        val_work[threadIdx.x] = -1.0e-38;
        idx_work[threadIdx.x] = 0xffffffff;
    }
    
    __syncthreads();

    // s=size of half group.  Start with half the elements and
    // reduce pairs, divide in two, repeat, etc.  Note that
    // half of all threads are idle in this algorithm.  Kind of wasteful...
    for (int s=CUDA_BLOCKSIZE/2; s>0; s>>=1)
    {
        // Are we within the active working group?
        if (threadIdx.x < s)
        {
            // Are we less than our paired value?  If so, take max.
            if (val_work[threadIdx.x] < val_work[threadIdx.x + s])
            {
                val_work[threadIdx.x] = val_work[threadIdx.x + s];
                idx_work[threadIdx.x] = idx_work[threadIdx.x + s];
            }
        }

        __syncthreads();
    }

    // Lowest element now has the max index within the block.  Put in global memory.
    if (threadIdx.x == 0)
    {
        val_out[blockIdx.x] = val_work[0];
        idx_out[blockIdx.x] = idx_work[0];
    }
}

////////////////////////////////////////////////////////////////////////////
// reduce_max
//      Find the maximum value in a floating point array on CUDA.
//      Assumes the array is already loaded into reduce_max_data_gpu.val[0]
//      GPU array.
////////////////////////////////////////////////////////////////////////////
void reduce_max(int N_elem, uint32_t *idx_out)
{
    uint32_t N_blocks;
    uint32_t n, idx_this, idx_next;
    uint32_t N_elem1;
    
    N_elem1 = N_elem;
    N_blocks = (N_elem1 + CUDA_BLOCKSIZE - 1)/CUDA_BLOCKSIZE;

    n = 0;
    idx_this = 0;
    idx_next = 1;
    while (N_elem1 > 1)
    {
        // Do iteration
        reduce_max_k<<<N_blocks, CUDA_BLOCKSIZE>>>(
                                                   reduce_max_data_gpu.val[idx_this], reduce_max_data_gpu.idx[idx_this],
                                                   reduce_max_data_gpu.val[idx_next], reduce_max_data_gpu.idx[idx_next],
                                                   (n==0), N_elem1);
        cudaDeviceSynchronize();
        
        N_elem1 = N_blocks;
        N_blocks = (N_elem1 + CUDA_BLOCKSIZE - 1)/CUDA_BLOCKSIZE;

        idx_this = !idx_this;
        idx_next = !idx_next;
        n++;
    }

    cudaMemcpy(idx_out, reduce_max_data_gpu.idx[idx_this], sizeof(float), cudaMemcpyDeviceToHost);
}

////////////////////////////////////////////////////////////////////////////
// reduce_max_init
////////////////////////////////////////////////////////////////////////////
int reduce_max_init(uint32_t max_size)
{
    int ii;
    
    // Zero pointers
    reduce_max_data_gpu.val[0] = 0;
    reduce_max_data_gpu.val[1] = 0;
    reduce_max_data_gpu.idx[0] = 0;
    reduce_max_data_gpu.idx[1] = 0;
    
    // Allocate ping pong buffers
    for (ii=0; ii<2; ii++)
    {
        if (cudaMalloc(&reduce_max_data_gpu.val[ii], sizeof(float)*max_size) != cudaSuccess)
        {
            log_printf(log_error, "Unable to allocate space for reduce max values.\n");
            goto backout_base;
        }

        if (cudaMalloc(&reduce_max_data_gpu.idx[ii], sizeof(uint32_t)*max_size) != cudaSuccess)
        {
            log_printf(log_error, "Unable to allocate space for reduce max indices.\n");
            goto backout_base;
        }
    }

    reduce_max_data_gpu.max_N = max_size;

    return(0);

 backout_base:
    if (reduce_max_data_gpu.val[0]) cudaFree(reduce_max_data_gpu.val[0]);
    if (reduce_max_data_gpu.val[1]) cudaFree(reduce_max_data_gpu.val[1]);
    if (reduce_max_data_gpu.idx[0]) cudaFree(reduce_max_data_gpu.idx[0]);
    if (reduce_max_data_gpu.idx[1]) cudaFree(reduce_max_data_gpu.idx[1]);
    return(-1);
}


void reduce_max_uninit()
{
    if (reduce_max_data_gpu.val[0]) cudaFree(reduce_max_data_gpu.val[0]);
    if (reduce_max_data_gpu.val[1]) cudaFree(reduce_max_data_gpu.val[1]);
    if (reduce_max_data_gpu.idx[0]) cudaFree(reduce_max_data_gpu.idx[0]);
    if (reduce_max_data_gpu.idx[1]) cudaFree(reduce_max_data_gpu.idx[1]);
}

// Functions specific to green dot finding

////////////////////////////////////////////////////////////////////////////
// crFindGreenDotInit
//      Initializes the CUDA version of the green dot finding algorithm.
////////////////////////////////////////////////////////////////////////////
int crFindGreenDotInit()
{
    // Initialize the generic maximizer
    if (reduce_max_init(mainOptions.cap_width*mainOptions.cap_height) != 0)
    {
        return(-1);
    }

    return(0);
}

////////////////////////////////////////////////////////////////////////////
// crFindGreenDotUninit
////////////////////////////////////////////////////////////////////////////
void crFindGreenDotUninit()
{
    reduce_max_uninit();
}

////////////////////////////////////////////////////////////////////////////
// get_green_dot_G_value_k
//      CUDA kernel to get green value of each pixel and store as
//      a float.
////////////////////////////////////////////////////////////////////////////
__global__ void get_green_dot_G_value_k(uint32_t *rgba_data, float *G_data, uint32_t frame_size)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    int stride = blockDim.x * gridDim.x;
    int i;

    float R, G, B, val1;
    
    for (i=idx; i < frame_size; i+=stride)
    {
        R = (float)(rgba_data[i] & 0x000000ff)/255.0;
        G = (float)((rgba_data[i] & 0x0000ff00) >> 8)/255.0;
        B = (float)((rgba_data[i] & 0x00ff0000) >> 16)/255.0;

        val1 = fmaxf((G - fmaxf(R,B)), 0.0);
        G_data[i] = val1;
    }
}

////////////////////////////////////////////////////////////////////////////
// crFindDot()
//      Finds dot in camera frame using CUDA.  Assumes that the
//      gpuCamFrameRgba already holds RGBA data for remapped frame.
////////////////////////////////////////////////////////////////////////////
void crFindDot(int *xl, int *yl, int *xr, int *yr, void *gpuCamFrameRgba)
{
    uint32_t N_blocks;
    uint32_t frame_size;
    uint32_t idx;

    frame_size = mainOptions.cap_width*mainOptions.cap_height;
    
    // Left Frame
    
    // Convert RGBA data to floating point green data.  Copies from
    // gpuCamFrameOut to gpu.val[0].
    N_blocks = (frame_size + CUDA_BLOCKSIZE - 1)/CUDA_BLOCKSIZE;    
    get_green_dot_G_value_k<<<N_blocks, CUDA_BLOCKSIZE>>>((uint32_t *) gpuCamFrameRgba,
                                                          reduce_max_data_gpu.val[0],
                                                          frame_size);
    // Find maximum index
    reduce_max(frame_size, &idx);

    *yl = idx/mainOptions.cap_width;
    *xl = idx % mainOptions.cap_width;

    // Right Frame
    N_blocks = (frame_size + CUDA_BLOCKSIZE - 1)/CUDA_BLOCKSIZE;    
    get_green_dot_G_value_k<<<N_blocks, CUDA_BLOCKSIZE>>>((uint32_t *) gpuCamFrameRgba + frame_size,
                                                          reduce_max_data_gpu.val[0],
                                                          frame_size);

    // Find maximum index
    reduce_max(frame_size, &idx);

    *yr = idx/mainOptions.cap_width;
    *xr = idx % mainOptions.cap_width;
}


////////////////////////////////////////////////////////////////////////////
// cudaMemcpy wrapper functions.
////////////////////////////////////////////////////////////////////////////
void crMemcpyD2H(void *dest, void *src, size_t size)
{
    cudaMemcpy(dest, src, size, cudaMemcpyDeviceToHost);
}
void crMemcpyH2D(void *dest, void *src, size_t size)
{
    cudaMemcpy(dest, src, size, cudaMemcpyHostToDevice);
}
int crMalloc(void **ptr, size_t size)
{
    return(cudaMalloc(ptr, size));
}
void crFree(void *ptr)
{
    cudaFree(ptr);
}
