/*============================================================================
 * fifo.cc
 *      Implements a simple software FIFO.
 *
 * Author: Jon W. Wallace, Lafayette College
 * Date:   May 2020
 *==========================================================================*/

#include <stdio.h>
#include <malloc.h>
#ifdef __WINDOWS__
#include <Windows.h>
#else
#include <pthread.h>
#endif
#include <string.h>

#include "fifo.h"

/*----------------------------------------------------------------------------
 * fifoCreate
 *      Creates a new FIFO.
 * Inputs:
 *      frameSize       Max size of a single FIFO record.
 *      N_frames        Total number of frames in FIFO.
 *--------------------------------------------------------------------------*/
int fifoCreate(fifoIntType frameSize, fifoIntType N_frames, fifoDef **fifo)
{
    int ccode;
    fifoDef *f;

    if ((f = (fifoDef *) malloc(sizeof(fifoDef) + (frameSize+FIFO_NODE_SIZE)*N_frames)) == 0)
    {
        fprintf(stderr, "Unable to allocate space for FIFO.\n");
        ccode = -1;
        goto backout_base;
    }
    
    memset(f, 0, sizeof(fifoDef));

    /* FIFO data records stored right after header. */
    f->data = (void *) (f + 1);

    f->N_frames = N_frames;
    f->frameSize = frameSize;
    
    /* Initialize the mutex */
#ifdef __WINDOWS__
    if ((f->write_mutex = CreateMutex(NULL, FALSE, NULL)) == 0)
#else
    if (pthread_mutex_init(&(f->write_mutex), NULL) != 0)
#endif    
    {
        fprintf(stderr, "Unable to initialize FIFO write mutex.\n");
        ccode = -2;
        goto backout_data;
    }

#ifdef __WINDOWS__
    if ((f->read_mutex = CreateMutex(NULL, FALSE, NULL)) == 0)
#else
    if (pthread_mutex_init(&(f->read_mutex), NULL) != 0)
#endif        
    {
        fprintf(stderr, "Unable to initialize FIFO read mutex.\n");
        ccode = -2;
        goto backout_write_mutex;
    }

    *fifo = f;

    /* Success. */
    ccode = 0;
    return(ccode);

 backout_write_mutex:
#ifdef __WINDOWS__
    CloseHandle(f->write_mutex);
#else    
    pthread_mutex_destroy(&(f->write_mutex));
#endif    
    
 backout_data:
    free(f);

 backout_base:
    return(ccode);
}

/*----------------------------------------------------------------------------
 * fifoDestroy
 *      Frees up a FIFO.
 *--------------------------------------------------------------------------*/
void fifoDestroy(fifoDef *fifo)
{
#ifdef __WINDOWS__
    CloseHandle(fifo->read_mutex);
    CloseHandle(fifo->write_mutex);
#else    
    pthread_mutex_destroy(&(fifo->read_mutex));
    pthread_mutex_destroy(&(fifo->write_mutex));
#endif    
    free(fifo);
}

/*----------------------------------------------------------------------------
 * fifoIsEmpty
 *      Indicates whether FIFO is empty (1) or not (0).
 *--------------------------------------------------------------------------*/
int fifoIsEmpty(fifoDef *fifo)
{
    int empty;

    /* Lock the FIFO. */
#ifdef __WINDOWS__
    WaitForSingleObject(fifo->read_mutex, INFINITE);
#else    
    pthread_mutex_lock(&(fifo->read_mutex));
#endif
    
    /* Check empty status. */
    empty = FIFO_IS_EMPTY(fifo);
    
#ifdef __WINDOWS__
    ReleaseMutex(fifo->read_mutex);
#else
    pthread_mutex_unlock(&(fifo->read_mutex));
#endif        
    return(empty);
}

/*----------------------------------------------------------------------------
 * fifoIFull
 *      Indicates whether FIFO is empty (1) or not (0).
 *--------------------------------------------------------------------------*/
int fifoIsFull(fifoDef *fifo)
{
    int full;

    /* Lock the FIFO. */
#ifdef __WINDOWS__
    WaitForSingleObject(fifo->read_mutex, INFINITE);
#else    
    pthread_mutex_lock(&(fifo->read_mutex));
#endif
    
    /* Check empty status. */
    full = FIFO_IS_FULL(fifo);
    
#ifdef __WINDOWS__
    ReleaseMutex(fifo->read_mutex);
#else
    pthread_mutex_unlock(&(fifo->read_mutex));
#endif        
    return(full);
}

/*----------------------------------------------------------------------------
 * fifoReadPtr
 *      Returns a pointer to data on top of the FIFO.  If the FIFO is
 *      empty, returns NULL.  Does not actually remove the data from the
 *      the FIFO.  FIFO read interface stays locked (to make thread safe)
 *      until you call fifoReadDone.
 *--------------------------------------------------------------------------*/
void *fifoReadPtr(fifoDef *fifo)
{
    /* Lock the FIFO. */
#ifdef __WINDOWS__
    WaitForSingleObject(fifo->read_mutex, INFINITE);
#else    
    pthread_mutex_lock(&(fifo->read_mutex));
#endif
    
    /* Check if read is possible. */
    if (FIFO_IS_EMPTY(fifo))
    {
        /* Empty. */
#ifdef __WINDOWS__
        ReleaseMutex(fifo->read_mutex);
#else
        pthread_mutex_unlock(&(fifo->read_mutex));
#endif        
        return(0);
    }
    else
    {
        /* Return pointer to data.  Keep FIFO locked. */
        return((void *) ((unsigned char *) fifo->data + (fifo->frameSize+FIFO_NODE_SIZE)*fifo->headIdx + FIFO_NODE_SIZE));
    }
}

/*----------------------------------------------------------------------------
 * fifoReadDone
 *      Removes data from head of FIFO.
 *--------------------------------------------------------------------------*/
void fifoReadDone(fifoDef *fifo)
{
    fifo->headIdx = FIFO_HEAD_NEXT(fifo);
#ifdef __WINDOWS__
    ReleaseMutex(fifo->read_mutex);
#else    
    pthread_mutex_unlock(&(fifo->read_mutex));
#endif    
}

/*----------------------------------------------------------------------------
 * fifoReadDoneAbort
 *      Stops a read.  Doesn't removed data from FIFO.
 *--------------------------------------------------------------------------*/
void fifoReadDoneAbort(fifoDef *fifo)
{
#ifdef __WINDOWS__
    ReleaseMutex(fifo->read_mutex);
#else
    pthread_mutex_unlock(&(fifo->read_mutex));
#endif    
}

/*----------------------------------------------------------------------------
 * fifoRead
 *      Copies data from head of FIFO into a buffer.  Returns != 0 if
 *      data is not available.
 *--------------------------------------------------------------------------*/
int fifoRead(fifoDef *fifo, void *data)
{
    /* Lock the FIFO. */
#ifdef __WINDOWS__
    WaitForSingleObject(fifo->read_mutex, INFINITE);
#else
    pthread_mutex_lock(&(fifo->read_mutex));
#endif
    
    /* Check if read is possible. */
    if (FIFO_IS_EMPTY(fifo))
    {
        /* Empty. */
#ifdef __WINDOWS__
        ReleaseMutex(fifo->read_mutex);
#else        
        pthread_mutex_unlock(&(fifo->read_mutex));
#endif        
        return(-1);
    }

    /* Copy data from FIFO. */
    memcpy(data, (unsigned char *) fifo->data + (fifo->frameSize+FIFO_NODE_SIZE)*fifo->headIdx + FIFO_NODE_SIZE,
           FIFO_PAYLOAD_SIZE(fifo->data));

    /* Advance pointer. */
    fifo->headIdx = FIFO_HEAD_NEXT(fifo);
    
    /* Unlock it. */
#ifdef __WINDOWS__
    ReleaseMutex(fifo->read_mutex);
#else    
    pthread_mutex_unlock(&(fifo->read_mutex));
#endif    

    return(0);
}

/*----------------------------------------------------------------------------
 * fifoWritePtr
 *      Gets a pointer to the record at tail of FIFO.  Returns NULL if the
 *      FIFO is full.
 *--------------------------------------------------------------------------*/
void *fifoWritePtr(fifoDef *fifo)
{
    void *data;

    /* Lock FIFO write interface. */
#ifdef __WINDOWS__    
    WaitForSingleObject(fifo->write_mutex, INFINITE);
#else
    pthread_mutex_lock(&(fifo->write_mutex));
#endif
    
    /* Check if it is not full */
    if (FIFO_IS_FULL(fifo))
    {
        /* Full.  Can't do write. */
#ifdef __WINDOWS__
        ReleaseMutex(fifo->write_mutex);
#else        
        pthread_mutex_unlock(&(fifo->write_mutex));
#endif        
        return(0);
    }
    else
    {
        /* Get pointer to data. */
        data = (void *) ((unsigned char *) fifo->data + (fifo->frameSize + FIFO_NODE_SIZE)*fifo->tailIdx + FIFO_NODE_SIZE);
        /* Initially set size of payload to say full size. */
        FIFO_PAYLOAD_SIZE(data) = fifo->frameSize;
        
        /* Return pointer.  FIFO write interface locked. */
        return(data);
    }
}

/*----------------------------------------------------------------------------
 * fifoWriteDone
 *      Finishes access to write FIFO interface.
 *--------------------------------------------------------------------------*/
void fifoWriteDone(fifoDef *fifo)
{
    /* Advance pointer past element to indicate element used. */
    fifo->tailIdx = FIFO_TAIL_NEXT(fifo);
#ifdef __WINDOWS__
    ReleaseMutex(fifo->write_mutex);
#else    
    pthread_mutex_unlock(&(fifo->write_mutex));
#endif
}

/*----------------------------------------------------------------------------
 * fifoWriteDoneAbort
 *      Finishes access to the FIFO, but does not keep write.
 *--------------------------------------------------------------------------*/
void fifoWriteDoneAbort(fifoDef *fifo)
{
    /* Just unlock.  Don't advance tail pointer. */
#ifdef __WINDOWS__
    ReleaseMutex(fifo->write_mutex);
#else    
    pthread_mutex_unlock(&(fifo->write_mutex));
#endif    
}

/*----------------------------------------------------------------------------
 * fifoElemUsed
 *      Get number of used elements in the FIFO.
 *--------------------------------------------------------------------------*/
fifoIntType fifoElemUsed(fifoDef *fifo)
{
    fifoIntType head1, tail1;

    head1 = fifo->headIdx;
    tail1 = fifo->tailIdx;
    
    if (fifo->tailIdx < fifo->headIdx)
    {
        return(fifo->N_frames - (head1 - tail1));
    }
    else
    {
        return(tail1 - head1);
    }
    
}

/*----------------------------------------------------------------------------
 * fifoWrite
 *      Writes data into FIFO using a copy.
 *--------------------------------------------------------------------------*/
int fifoWrite(fifoDef *fifo, void *data)
{
    unsigned char *fifo_data;

    /* Lock FIFO write interface. */
#ifdef __WINDOWS__
    WaitForSingleObject(fifo->write_mutex, INFINITE);
#else    
    pthread_mutex_lock(&(fifo->write_mutex));
#endif    

    /* Check if it is not full */
    if (FIFO_IS_FULL(fifo))
    {
        /* Full.  Can't do write. */
#ifdef __WINDOWS__
        ReleaseMutex(fifo->write_mutex);
#else
        pthread_mutex_unlock(&(fifo->write_mutex));
#endif        
        return(-1);
    }

    /* Get pointer to data. */
    fifo_data = (unsigned char *) fifo->data + (fifo->frameSize + FIFO_NODE_SIZE)*fifo->tailIdx + FIFO_NODE_SIZE;
    memcpy((unsigned char *) fifo_data, data, fifo->frameSize);
    FIFO_PAYLOAD_SIZE(fifo_data) = fifo->frameSize;

    /* Increment FIFO pointer. */
    fifo->tailIdx = FIFO_TAIL_NEXT(fifo);
    
    /* Unlock FIFO. */
#ifdef __WINDOWS__
    ReleaseMutex(fifo->write_mutex);
#else    
    pthread_mutex_unlock(&(fifo->write_mutex));
#endif    
    return(0);
}


