////////////////////////////////////////////////////////////////////////////
// gaugeCards.cc
//      Code to simplify drawing "steam gauge" type instruments using
//      openGL.  Each gauge is represented by a stack of PNG image
//      cards that can be rotated.  A configuration file specifies
//      various properties of the gauges.
//
// Note: Uses the lodepng.c library (by Lode Vandevenne) to load
//       texture images.
//
// Author: Jon W. Wallace, Lafayette College
// Date:   May 2020
////////////////////////////////////////////////////////////////////////////

//#include <windows.h>  // for MS Windows

#include <stdio.h>

#include <GL/glew.h>
#include <GLFW/glfw3.h>
#include "glh.h"

#include "math.h"
#include "lodepng.h"
#include "gaugeCards.h"

int gcGaugeCount;
gaugeDef gcGauges[MAX_GAUGES];

static char line[256];

////////////////////////////////////////////////////////////////////////////
// readGauges
//      Reads in the gauges configuration file.
////////////////////////////////////////////////////////////////////////////
int readGauges(const char *fname)
{
    int ccode;
    FILE *f;
    gaugeDef *g;
    int i, j;
    unsigned int w, h;
    
    memset(gcGauges, 0, sizeof(gcGauges));
    gcGaugeCount = 0;

    if ((f = fopen(fname, "rb")) == 0)
    {
        fprintf(stderr, "Unable to open gauge configuration file.\n");
        ccode = -1;
        goto backout_base;
    }

    for ( ; ; )
    {
        // Go until we find a line with a * indicating a new gauge
        if (fgets(line, sizeof(line), f) == 0)
        {
            break;
        }

        if (line[0] == '*')
        {
            /* Process a single gauge. */
            if (gcGaugeCount >= MAX_GAUGES)
            {
                fprintf(stderr, "Too many gauges.\n");
                ccode = -2;
                goto backout_base;
            }
            
            g = &gcGauges[gcGaugeCount];
            if (sscanf(line, "*%d %le %le %le %le %le", &g->N, &g->x, &g->y, &g->scale_x, &g->scale_y, &g->rot) != 6)
            {
                fprintf(stderr, "Error parsing gauge line: %s\n", line);
                ccode = -3;
                goto backout_base;
            }

            // Parse card lines
            for (i=0; i<g->N; i++)
            {
                for ( ; ; )
                {
                    if (fgets(line, sizeof(line), f) == 0)
                    {
                        fprintf(stderr, "Unexpected end of file.\n");
                        ccode = -4;
                        goto backout_base;
                    }

                    if (line[0] != '#') break;
                }
                
                // Allocate space for filename
                if ((g->cards[i].img_name = (char *) malloc(strlen(line)+1)) == 0)
                {
                    fprintf(stderr, "Unable to allocate space for string.\n");
                    ccode = -5;
                    goto backout_base;
                }

                // Find the end of the filename (end at a space, CR, NL, or null)
                for (j=0; j<strlen(line); j++)
                {
                    if ((line[j] == ' ') || (line[j] == '\n') || (line[j] == '\r')) break;
                }
                memcpy(g->cards[i].img_name, line, j);
                g->cards[i].img_name[j] = 0;
                line[j] = 0;

                for ( ; ; )
                {
                    if (fgets(line, sizeof(line), f) == 0)
                    {
                        fprintf(stderr, "Unexpected end of file.\n");
                        ccode = -6;
                        goto backout_base;
                    }

                    if (line[0] != '#') break;
                }
                    
                if (sscanf(line, "%le %le %le %le", &g->cards[i].xl, &g->cards[i].yl, &g->cards[i].xg, &g->cards[i].yg) != 4)
                {
                    fprintf(stderr, "Error parsing card line: %s\n", line);
                    ccode = -7;
                    goto backout_base;
                }

                // Read in the card image data
                if (lodepng_decode32_file(&g->cards[i].image, &w, &h, g->cards[i].img_name))
                {
                    fprintf(stderr, "Unable to open image: %s.\n", g->cards[i].img_name);
                    ccode = -8;
                    goto backout_base;
                }
                g->cards[i].w = (double)w;
                g->cards[i].h = (double)h;

            } // For cards

            gcGaugeCount++;
        }
        
    }   // For gagues
    
    fclose(f);
    return(0);
    
 backout_base:
    return(ccode);
}

void printGauges()
{
    int i, j;

    for (j=0; j<gcGaugeCount; j++)
    {
        printf("Gauge %d: N=%d x=%.1lf y=%.1lf scale_x=%.1lf scale_y=%.1lf rot=%.1lf\n",
               j, gcGauges[j].N, gcGauges[j].x, gcGauges[j].y, gcGauges[j].scale_x, gcGauges[j].scale_y, gcGauges[j].rot);
        for (i=0; i<gcGauges[j].N; i++)
        {
            printf("    \"%s\" xg=%.1lf yg=%.1lf xl=%.1lf yl=%.1lf\n",
                   gcGauges[j].cards[i].img_name,
                   gcGauges[j].cards[i].xg,
                   gcGauges[j].cards[i].yg,
                   gcGauges[j].cards[i].xl,
                   gcGauges[j].cards[i].yl);
        }
    }
}

void allocTextures()
{
    int i, j;
    GLuint id;
    cardDef *c;
    
    for (j=0; j<gcGaugeCount; j++)
    {
        for (i=0; i<gcGauges[j].N; i++)
        {
            c = &(gcGauges[j].cards[i]);
            glGenTextures(1, &c->textureID);
            glBindTexture(GL_TEXTURE_2D, c->textureID);
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, c->w, c->h, 0, GL_RGBA, GL_UNSIGNED_BYTE, c->image);
            glGenerateTextureMipmap(c->textureID);
            //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
            //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
            glSamplerParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
            //glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, 1.0f);            
        }
    }
}

////////////////////////////////////////////////////////////////////////////
// gcRenderGauge
//      Renders the gauge at the given index into the specified
//      glh framebuffer.
////////////////////////////////////////////////////////////////////////////
void gcRenderGauge(int gauge_idx, glhFbDef *fb)
{
    int i, j, k;
    
    double ws_x, ws_y;
    double x0, y0;

    double x1, y1, x2, y2, x3, y3, x4, y4;
    
    gaugeDef *g;
    cardDef *c;

    glBindFramebuffer(GL_FRAMEBUFFER, fb->fbo_id);
    //glBindFramebuffer(GL_FRAMEBUFFER, 0);
    glViewport(0, 0, fb->width, fb->height);

    glPushMatrix();
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-1.0,   // left
            1.0,   // right
            -1.0,   // bottom
            1.0,   // top
            1.0,  // near
            -1.0);  // far
    
    glClearColor(1.0f, 1.0f, 0.0f, 1.0f); // Set background color to black and opaque
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);  // Clear the color buffer (background)

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   
    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

    
    g = &gcGauges[gauge_idx];
        
    // Get scale factors for gauge
    //ws_x = 2.0/(double)fb->width*g->scale_x;
    //ws_y = 2.0/(double)fb->height*g->scale_y;
    ws_x = 2.0/(double)g->cards[0].w*g->scale_x;
    ws_y = 2.0/(double)g->cards[0].h*g->scale_y;
    
    // Get corner of the gauge
    x0 = -1.0 + ws_x*g->x; y0 = -1.0 + ws_y*g->y;

    // Draw the cards
    for (j=0; j<g->N; j++)
    {
        c = &g->cards[j];
        
        // Map corners of card
        x1 =  (0.0-c->xl)*cos(c->rot) + (c->h-c->yl)*sin(c->rot) + c->xg;
        y1 = -(0.0-c->xl)*sin(c->rot) + (c->h-c->yl)*cos(c->rot) + c->yg;
            
        x2 =  (c->w-c->xl)*cos(c->rot) + (c->h-c->yl)*sin(c->rot) + c->xg;
        y2 = -(c->w-c->xl)*sin(c->rot) + (c->h-c->yl)*cos(c->rot) + c->yg;
            
        x3 =  (c->w-c->xl)*cos(c->rot) + (0.0-c->yl)*sin(c->rot) + c->xg;
        y3 = -(c->w-c->xl)*sin(c->rot) + (0.0-c->yl)*cos(c->rot) + c->yg;

        x4 =  (0.0-c->xl)*cos(c->rot) + (0.0-c->yl)*sin(c->rot) + c->xg;
        y4 = -(0.0-c->xl)*sin(c->rot) + (0.0-c->yl)*cos(c->rot) + c->yg;

        // Draw the mapped rectangle
        glBindTexture(GL_TEXTURE_2D, c->textureID);
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

        glBegin(GL_QUADS);

        glTexCoord2f(0.0, 0.0); 
        glVertex2f(x0 + x1*ws_x, y0 + y1*ws_y);
        glTexCoord2f(1.0, 0.0); 
        glVertex2f(x0 + x2*ws_x, y0 + y2*ws_y);
        glTexCoord2f(1.0, 1.0); 
        glVertex2f(x0 + x3*ws_x, y0 + y3*ws_y);
        glTexCoord2f(0.0, 1.0); 
        glVertex2f(x0 + x4*ws_x, y0 + y4*ws_y);

        glEnd();

    }

    if (fb->mipmap)
    {
        glGenerateTextureMipmap(fb->tex_id);
    }
    
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    glPopMatrix();
}

////////////////////////////////////////////////////////////////////////////
// gcInit
//      Initialize the gauge card code.
////////////////////////////////////////////////////////////////////////////
int gcInit()
{
    if (readGauges("gauges.cfg") != 0)
    {
        return(-1);
    }
    //printGauges();  // Debug
    allocTextures();

    return(0);
}

////////////////////////////////////////////////////////////////////////////
// gcUninit
//      Uninitialize the gauge card code.
////////////////////////////////////////////////////////////////////////////
void gcUninit()
{
    int i, j;
    
    for (i=0; i<gcGaugeCount; i++)
    {
        for (j=0; j<gcGauges[i].N; j++)
        {
            if (gcGauges[i].cards[j].img_name) free(gcGauges[i].cards[j].img_name);
            if (gcGauges[i].cards[j].image) free(gcGauges[i].cards[j].image);
        }
    }
    memset(gcGauges, 0, sizeof(gcGauges));
    gcGaugeCount = 0;
}
