////////////////////////////////////////////////////////////////////////////
// gauges.cc
//      Code to support computing position and drawing gauges on
//      instrument panel.
//
// Author: Jon W. Wallace, Lafayette College
// Date:   May 2020
////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <math.h>

#include "gauges.h"
#include "aruco.h"
#include "log.h"
#include "ar_parm.h"
#include "gaugeCards.h"
#include "udp_recv_stats.h"
#include "xplaneXfer.h"
#include "main.h"

int gaugesDetectInit(int tag_dict, int cap_w, int cap_h)
{
    int ccode;
    int i;
    
    return(0);

 backout_base:
    return(ccode);
}

void gaugesDetectUninit()
{
}

void gaugesDetectTags(unsigned char *imgL, unsigned char *imgR)
{
    //arucoDetect(imgL, tags_L);
    //arucoDetect(imgR, tags_R);
}

//============================================================================
// GaugesDraw interface.  Functions below this line are to be called from
// the render thread that has OpenGL context.
//============================================================================

// Framebuffer to render gauge to
glhFbDef fb_gauge;
const int gauge_tex_width       = 512;
const int gauge_tex_height      = 512;
const int gauge_mipmap          = 1;

typedef void (*gaugeDrawFunc)(glhFbDef *fb);

#define GAUGE_ID_RPM    0
#define GAUGE_ID_ALT    1
#define GAUGE_ID_ASI    2

void gaugeDrawRpm(glhFbDef *fb);
void gaugeDrawAlt(glhFbDef *fb);
void gaugeDrawAsi(glhFbDef *fb);

gaugeDrawFunc gaugeDrawFuncs[ARUCO_MAX_IDS] =
    {gaugeDrawRpm,
     gaugeDrawAlt,
     gaugeDrawAsi,
     0,
     0,
     0,
     0,
     0};

////////////////////////////////////////////////////////////////////////////
// gaugesDrawInit
//      Load all the textures associated with the gauges.  Must be called
//      from thread having openGL context.
////////////////////////////////////////////////////////////////////////////
int gaugesDrawInit()
{
    int ccode;
    
    if (gcInit() != 0)
    {
        log_printf(log_error, "Error initializing gauge cards module.\n");
        ccode = -1;
        goto backout_base;
    }

    // Texture to hold single rendered gauge
    if (glhFbNew(&fb_gauge, gauge_tex_width, gauge_tex_height, 1, 0, 0, gauge_mipmap) != 0)
    {
        log_printf(log_error, "Unable to allocate gauge render texture.\n");
        ccode = -1;
        goto backout_gc;
    }
    
    return(0);

 backout_gc:
    gcUninit();
    
 backout_base:
    return(ccode);
}

////////////////////////////////////////////////////////////////////////////
// tag_add_border
//      Places a border around the detected tag.
////////////////////////////////////////////////////////////////////////////
void tag_add_border(float x_in[], float y_in[], float x_out[], float y_out[],
                    float b_left, float b_right, float b_bot, float b_top)
{
    float u0[2], u1[2], u2[2], u3[2];

    // Get edge vectors
    u0[0] = x_in[1] - x_in[0];
    u0[1] = y_in[1] - y_in[0];

    u1[0] = x_in[2] - x_in[1];
    u1[1] = y_in[2] - y_in[1];

    u2[0] = x_in[3] - x_in[2];
    u2[1] = y_in[3] - y_in[2];

    u3[0] = x_in[0] - x_in[3];
    u3[1] = y_in[0] - y_in[3];

    x_out[0] = x_in[0] + u3[0]*b_top - u0[0]*b_left;
    y_out[0] = y_in[0] + u3[1]*b_top - u0[1]*b_left;

    x_out[1] = x_in[1] - u1[0]*b_top + u0[0]*b_right;
    y_out[1] = y_in[1] - u1[1]*b_top + u0[1]*b_right;

    x_out[2] = x_in[2] + u1[0]*b_bot - u2[0]*b_right;
    y_out[2] = y_in[2] + u1[1]*b_bot - u2[1]*b_right;

    x_out[3] = x_in[3] - u3[0]*b_bot + u2[0]*b_left;
    y_out[3] = y_in[3] - u3[1]*b_bot + u2[1]*b_left;
}

////////////////////////////////////////////////////////////////////////////
// gaugesDraw
//      Draw all the gauges to the L/R 
////////////////////////////////////////////////////////////////////////////
void gaugesDraw(glhFbDef *fb_l, glhFbDef *fb_r)
{
    int i, j;
    float x_map_l[4], y_map_l[4];
    float x_map_r[4], y_map_r[4];
    float tag_new_x[4], tag_new_y[4];

    struct timespec t;
    arucoPosDef pos_L, pos_R;
    
    if (mainOptions.xplane)
    {
        // Get gauge information from shared memory
        xplaneXferGet();
    }
    
    for (i=0; i<ARUCO_MAX_IDS; i++)
    {
        arucoGetTagCorners(i, &pos_L, &pos_R, &t);

        if ((pos_L.valid) || (pos_R.valid))
        {
            // Render the gauge into the frame buffer
            if (gaugeDrawFuncs[i])
            {
                (*gaugeDrawFuncs[i])(&fb_gauge);
            }
        }
        
        if (pos_L.valid)
        {
            // Shift border of tag
            tag_add_border(pos_L.x, pos_L.y, tag_new_x, tag_new_y, 0.2, 0.2, 0.2, 0.2);
            
            // Map corners of detected gauge to logical framebuffer coordinates.
            for (j=0; j<4; j++)
            {
                ar_parm_map1(tag_new_x[j], tag_new_y[j], &(x_map_l[j]), &(y_map_l[j]), M_cap_l);
            }
            //fprintf(stdout, "%02d: (%.0f, %.0f) => (%.2f, %.2f)\n",
            //        i, pos_L.x[0], pos_L.y[0], x_map_l[0], y_map_l[0]);

            if (gaugeDrawFuncs[i])
            {
                glhFbCopyQuad(&fb_gauge, fb_l,
                              0.5, 0.5, 1.0, 1.0,
                              x_map_l[3], y_map_l[3],
                              x_map_l[2], y_map_l[2],
                              x_map_l[1], y_map_l[1],
                              x_map_l[0], y_map_l[0],
                              0, 0);
            }
            else
            {
                glhDrawQuad(fb_l, x_map_l, y_map_l);
            }
        }

        if (pos_R.valid)
        {
            // Shift border of tag
            tag_add_border(pos_R.x, pos_R.y, tag_new_x, tag_new_y, 0.2, 0.2, 0.2, 0.2);
            
            // Map corners of detected gauge to logical framebuffer coordinates.
            for (j=0; j<4; j++)
            {
                ar_parm_map1(tag_new_x[j], tag_new_y[j], &(x_map_r[j]), &(y_map_r[j]), M_cap_r);
            }
            
            if (gaugeDrawFuncs[i])
            {
                glhFbCopyQuad(&fb_gauge, fb_r,
                               0.5, 0.5, 1.0, 1.0,
                              x_map_r[3], y_map_r[3],
                              x_map_r[2], y_map_r[2],
                              x_map_r[1], y_map_r[1],
                              x_map_r[0], y_map_r[0],
                              0, 0);
            }
            else
            {
                glhDrawQuad(fb_r, x_map_r, y_map_r);
            }
        }
    }
}

////////////////////////////////////////////////////////////////////////////
// gaugesDrawUninit
//      Free resources associated with the gaugesDraw interface.
////////////////////////////////////////////////////////////////////////////
void gaugesDrawUninit()
{
    gcUninit();
}

////////////////////////////////////////////////////////////////////////////
// Code to render specific gauges

// Tachomoeter

const float rpm_min = 0.0;
const float rpm_max = 3500.0;
const float rpm_ang_min = -145.0*M_PI/180.0;
const float rpm_ang_max = 122.0*M_PI/180.0;
void gaugeDrawRpm(glhFbDef *fb)
{
    volatile float rpm_val;

    if (mainOptions.xplane)
    {
        rpm_val = xplane_shm_local.r.engine_speed_rpm;
    }
    else
    {
        rpm_val = udp_stat_data.tach;
    }
    
    if (rpm_val < rpm_min) rpm_val = rpm_min;
    if (rpm_val > rpm_max) rpm_val = rpm_max;
    
    gcGauges[GAUGE_ID_RPM].cards[1].rot =
        (rpm_val - rpm_min)/(rpm_max - rpm_min)*(rpm_ang_max - rpm_ang_min) + rpm_ang_min;

    gcRenderGauge(GAUGE_ID_RPM, fb);
}


// Altimeter
void gaugeDrawAlt(glhFbDef *fb)
{
    volatile float alt_val;
    float alt_100, alt_1k, alt_10k;
    float ang_100, ang_1k, ang_10k;

    if (mainOptions.xplane)
    {
        alt_val = xplane_shm_local.r.altitude_ft_pilot;
    }
    else
    {
        alt_val = udp_stat_data.alt;
    }
    if (alt_val < 0.0) alt_val = 0.0;
    
    alt_100 = alt_val/1000.0f - floor(alt_val/1000.0f);
    alt_1k  = alt_val/10000.0f - floor(alt_val/10000.0f);
    alt_10k = alt_val/100000.0f - floor(alt_val/100000.0f);
    
    gcGauges[GAUGE_ID_ALT].cards[1].rot = alt_100*2.0*M_PI;
    gcGauges[GAUGE_ID_ALT].cards[2].rot = alt_1k*2.0*M_PI;
    gcGauges[GAUGE_ID_ALT].cards[3].rot = alt_10k*2.0*M_PI;
    
    gcRenderGauge(GAUGE_ID_ALT, fb);
}


// Air Speed

const float asi_clamp_min = 27.0;
const float asi_clamp_max = 180.0;
const float asi_coeff[3] = {-8.9940e+01, 3.4095e+00, -5.5357e-03};

void gaugeDrawAsi(glhFbDef *fb)
{
    volatile float asi_val;
    float th;

    if (mainOptions.xplane)
    {
        asi_val = xplane_shm_local.r.airspeed_kts_pilot;
    }
    else
    {
        asi_val = udp_stat_data.speed;
    }
    if (asi_val < asi_clamp_min) asi_val = asi_clamp_min;
    if (asi_val > asi_clamp_max) asi_val = asi_clamp_max;

    th = asi_coeff[0] + (asi_coeff[1]*asi_val) + (asi_coeff[2]*asi_val*asi_val);
    gcGauges[GAUGE_ID_ASI].cards[1].rot = th*M_PI/180.0;

    gcRenderGauge(GAUGE_ID_ASI, fb);
}


