////////////////////////////////////////////////////////////////////////////
// glh.cc
//      Some functions to hide the complexity of some OpenGL things.
//
// Author: Jon W. Wallace, Lafayette College
// Date:   May 2020
////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <GL/glew.h>
#include "glh.h"

#include "Matrices.h"
#include "Vectors.h"

////////////////////////////////////////////////////////////////////////////
// glhFbNew
//      Creates a new "framebuffer" object.  It is actually a texture with
//      an optional framebuffer backing that allows it to be rendered
//      into by OpenGL.
// Inputs:
//      fb              Framebuffer structre to be initialized
//      width, height   Size of the framebuffer
//      rend_flag       Flag indicating if it needs to support OpenGL
//                      rendering into it.
//      bgra            Flag to use BGRA encoding instead of RGBA
// Returns:
//      0               Success
//      nz              Error
////////////////////////////////////////////////////////////////////////////
int glhFbNew(glhFbDef *fb, uint32_t width, uint32_t height, int rend_flag, int bgra, unsigned char *data, uint32_t mipmap)
{
    uint32_t ii;
    int ccode;

    memset(fb, 0, sizeof(glhFbDef));
    fb->width = width;
    fb->height = height;
    fb->mipmap = mipmap;
    
    // Generate the associated texture object
    glGenTextures(1, &(fb->tex_id));
    if (fb->tex_id == GL_INVALID_VALUE)
    {
        fprintf(stderr, "Unable to create OpenGL texture.\n");
        ccode = -1;
        goto backout_base;
    }

    // Put initial data in texture
    glBindTexture(GL_TEXTURE_2D, fb->tex_id);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

    // Set some default parameters for the texture.
    if (!mipmap)
    {
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    }
    else
    {
        glGenerateTextureMipmap(fb->tex_id);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    }
    glBindTexture(GL_TEXTURE_2D, 0);

    // Fill out a reference for headset copy
    fb->tex.eColorSpace = ColorSpace_Gamma;
    fb->tex.eType = TextureType_OpenGL;
    fb->tex.handle = (void*)(uintptr_t)(fb->tex_id);

    if (rend_flag)
    {
        // Create framebuffer / render structures that we need in order
        // to render scenes to this texture.

        // create a framebuffer object
        glGenFramebuffers(1, &(fb->fbo_id));
        glBindFramebuffer(GL_FRAMEBUFFER, fb->fbo_id);

        // attach the texture to FBO color attachment point
        glFramebufferTexture2D(GL_FRAMEBUFFER,        // 1. fbo target: GL_FRAMEBUFFER
                               GL_COLOR_ATTACHMENT0,  // 2. attachment point
                               GL_TEXTURE_2D,         // 3. tex target: GL_TEXTURE_2D
                               fb->tex_id,            // 4. tex ID
                               0);                    // 5. mipmap level: 0(base)

        // check FBO status
        if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
        {
            fprintf(stderr, "Creation of framebuffer failed.\n");
            ccode = -2;
            goto backout_base;
        }

        // switch back to window-system-provided framebuffer
        glBindFramebuffer(GL_FRAMEBUFFER, 0);
    }
    
    // Success
    ccode = 0;
    
 backout_base:
    return(ccode);
}


int glhFbFree(glhFbDef *fb)
{
    glBindTexture(GL_TEXTURE_2D, 0);
    glDeleteTextures(1, &(fb->tex_id));
    memset(fb, 0, sizeof(glhFbDef));
}

////////////////////////////////////////////////////////////////////////////
// glhFbCpuWrite
//      Copies raw pixel data from CPU to a framebuffer (texture) object.
////////////////////////////////////////////////////////////////////////////
int glhFbCpuWrite(glhFbDef *fb, void *data)
{
    glBindTexture(GL_TEXTURE_2D, fb->tex_id);
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, fb->width, fb->height, GL_RGBA, GL_UNSIGNED_BYTE, (unsigned char *) data);
    return(0);
}

////////////////////////////////////////////////////////////////////////////
// glhFbCpuRead
//      Copies raw pixel data from framebuffer texture to CPU memory.
////////////////////////////////////////////////////////////////////////////
int glhFbCpuRead(glhFbDef *fb, void *data, int bufsize)
{
    glBindTexture(GL_TEXTURE_2D, fb->tex_id);
    glGetnTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, bufsize, (unsigned char *) data);
    return(0);
}

////////////////////////////////////////////////////////////////////////////
// glhFbCopyRect
//   Uses openGL to render one framebuffer (texture) into another.
//   Copies a rectangular region and supports cropping and scaling.
//   Coordinates are logical coordinates (0,0) lower left to (1,1) upper right.
// Inputs:
//      fb_src, fb_dest         Source and destination framebuffers
//      src_x, src_y            Center of source frame to copy
//      src_w, src_h            Width and height of source frame
//      dst_x, dst_y            Center of target frame
//      dst_w, dst_h            Width and height of target frame
////////////////////////////////////////////////////////////////////////////
int glhFbCopyRect(glhFbDef *fb_src, glhFbDef *fb_dest,
                  float src_x, float src_y, float src_w, float src_h,
                  float dst_x, float dst_y, float dst_w, float dst_h,
                  int do_clear, int do_flip)
{
    float src_x1, src_x2, src_y1, src_y2;
    float dst_x1, dst_x2, dst_y1, dst_y2;
    float tmp;
    
    src_x1 = src_x - 0.5*src_w;
    src_x2 = src_x + 0.5*src_w;
    if (do_flip)
    {
        src_y1 = src_y + 0.5*src_h;
        src_y2 = src_y - 0.5*src_h;
    }
    else
    {
        src_y1 = src_y - 0.5*src_h;
        src_y2 = src_y + 0.5*src_h;
    }
    dst_x1 = dst_x - 0.5*dst_w;
    dst_x2 = dst_x + 0.5*dst_w;
    dst_y1 = dst_y - 0.5*dst_h;
    dst_y2 = dst_y + 0.5*dst_h;
    
    glBindFramebuffer(GL_FRAMEBUFFER, fb_dest->fbo_id);
    glViewport(0, 0, fb_dest->width, fb_dest->height);

    glPushMatrix();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0,   // left
            1.0,   // right
            0.0,   // bottom
            1.0,   // top
            1.0,  // near
            -1.0);  // far

    if (do_clear)
    {
        glClearColor(0,0,0,1);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    }

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, fb_src->tex_id);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glBegin(GL_QUADS);
    // Lower left
    glTexCoord2f(src_x1, src_y1);
    glVertex2d(dst_x1, dst_y1);
    // Lower Right
    glTexCoord2f(src_x2, src_y1);
    glVertex2d(dst_x2, dst_y1);
    // Upper Right
    glTexCoord2f(src_x2, src_y2);
    glVertex2d(dst_x2, dst_y2);
    // Upper Left
    glTexCoord2f(src_x1, src_y2);
    glVertex2d(dst_x1, dst_y2);
    glEnd();

    glPopMatrix();

    if (fb_dest->mipmap)
    {
        glGenerateTextureMipmap(fb_dest->tex_id);
    }
    
    glDisable(GL_TEXTURE_2D);
}

////////////////////////////////////////////////////////////////////////////
// glhFbCopyQuad
// fbCopy3
//   Renders one framebuffer onto another one, where the destination is
//   an arbitrary quadralateral.
//   Coordinates are logical coordinates (-1,-1) lower left to (1,1) upper right.
// Inputs:
//      fb_src, fb_dest         Source and destination framebuffers
//      src_x, src_y            Center of source frame to copy
//      src_w, src_h            Width and height of source frame
//      dst_x1, dst_y1          Lower left target
//      dst_x2, dst_y2          Lower right
//      dst_x3, dst_y3          Upper right
//      dst_x4, dst_y4          Upper left
////////////////////////////////////////////////////////////////////////////
int glhFbCopyQuad(glhFbDef *fb_src, glhFbDef *fb_dest,
                  float src_x, float src_y, float src_w, float src_h,
                  float dst_x1, float dst_y1,
                  float dst_x2, float dst_y2,
                  float dst_x3, float dst_y3,
                  float dst_x4, float dst_y4,
                  int do_clear, int do_flip)
{
    float src_x1, src_x2, src_y1, src_y2;
    float tmp;
    
    src_x1 = src_x - 0.5*src_w;
    src_x2 = src_x + 0.5*src_w;
    if (do_flip)
    {
        src_y1 = src_y + 0.5*src_h;
        src_y2 = src_y - 0.5*src_h;
    }
    else
    {
        src_y1 = src_y - 0.5*src_h;
        src_y2 = src_y + 0.5*src_h;
    }

    glBindFramebuffer(GL_FRAMEBUFFER, fb_dest->fbo_id);
    glViewport(0, 0, fb_dest->width, fb_dest->height);

    glPushMatrix();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-1.0,   // left
            1.0,   // right
            -1.0,   // bottom
            1.0,   // top
            1.0,  // near
            -1.0);  // far

    if (do_clear)
    {
        glClearColor(0,0,0,1);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    }

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, fb_src->tex_id);

    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glBegin(GL_QUADS);
    // Lower left
    glTexCoord2f(src_x1, src_y1);
    glVertex2d(dst_x1, dst_y1);
    // Lower Right
    glTexCoord2f(src_x2, src_y1);
    glVertex2d(dst_x2, dst_y2);
    // Upper Right
    glTexCoord2f(src_x2, src_y2);
    glVertex2d(dst_x3, dst_y3);
    // Upper Left
    glTexCoord2f(src_x1, src_y2);
    glVertex2d(dst_x4, dst_y4);
    glEnd();

    glPopMatrix();

    if (fb_dest->mipmap)
    {
        glGenerateTextureMipmap(fb_dest->tex_id);
    }
    
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
}

////////////////////////////////////////////////////////////////////////////
// glhDrawQuad
//      Draws a quadralateral onto an output texture.  Logical coordinates
//      are (-1,-1) lower-left to (1,1) upper-right.
////////////////////////////////////////////////////////////////////////////
void glhDrawQuad(glhFbDef *fb, float x[], float y[])
{
    int i;
    
    glBindFramebuffer(GL_FRAMEBUFFER, fb->fbo_id);
    glViewport(0, 0, fb->width, fb->height);

    // Draw on the render texture
    glPushMatrix();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-1.0,   // left
            1.0,   // right
            -1.0,   // bottom
            1.0,   // top
            1.0,  // near
            -1.0);  // far

    glDisable(GL_TEXTURE_2D);

    glColor3f(0.1, 0.1, 1.0);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

    glBegin(GL_QUADS);
    for (i=0; i<4; i++)
    {
        glVertex2d(x[i], y[i]);
    }
    glEnd();
    glPopMatrix();

    if (fb->mipmap)
    {
        glGenerateTextureMipmap(fb->tex_id);
    }
    
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

