////////////////////////////////////////////////////////////////////////////
// grab_x11.cc
//      Code to grab raw RGB data from a given X11 window (flight simulator).
//      This is very CPU intensive code.  Eventually it should be replaced
//      with a GPU-only capture method.
//
// Author: Jon W. Wallace, Lafayette College
// Date:   May 2020
////////////////////////////////////////////////////////////////////////////

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <stdio.h>
#include <malloc.h>
#include <string.h>

#include "log.h"
#include "grab_x11.h"

typedef enum
{
    depth8 = 0,
    depth16,
    depth24,
    depth32,
    depth_max
} depthEnum;

typedef struct
{
    unsigned int Nx, Ny;

    unsigned char *rgbData;
    unsigned char *pixelData;
    
    Display *display;
    int screen;
    Visual *visual;
    Window window;
    XImage *image;
    depthEnum depth;
    GC gc;

    /* Number of bits for each color. */
    unsigned r_bits;
    unsigned b_bits;
    unsigned g_bits;

    /* Masks for the various colors. */
    unsigned r_mask;
    unsigned g_mask;
    unsigned b_mask;

    /* Number of bits to shift for r, g, b. */
    unsigned r_shift;
    unsigned g_shift;
    unsigned b_shift;

} xInfoDef;

xInfoDef t;
int tInit = 0;

int pixelSize[depth_max] = {1, 2, 4, 4};
int intDepth[depth_max] = {8, 16, 24, 32};

/*----------------------------------------------------------------------------
 * getImageColorExtents
 *      Get number of RGB bits in this image.
 *--------------------------------------------------------------------------*/
void getImageColorExtents(xInfoDef *t)
{
    unsigned r1, g1, b1;
    
    t->r_mask = r1 = t->image->red_mask;
    t->g_mask = g1 = t->image->green_mask;
    t->b_mask = b1 = t->image->blue_mask;

    t->r_shift = 0;
    t->g_shift = 0;
    t->b_shift = 0;

    t->r_bits = 0;
    t->g_bits = 0;
    t->b_bits = 0;

    /* Compute the shift values. */
    while ((r1 & 1) == 0)
    {
        t->r_shift++;
        r1 >>= 1;
    }

    while ((g1 & 1) == 0)
    {
        t->g_shift++;
        g1 >>= 1;
    }

    while ((b1 & 1) == 0)
    {
        t->b_shift++;
        b1 >>= 1;
    }

    /* Compute the number of bits. */
    r1++;
    g1++;
    b1++;

    while ((r1 & 1) == 0)
    {
        t->r_bits++;
        r1 >>= 1;
    }

    while ((g1 & 1) == 0)
    {
        t->g_bits++;
        g1 >>= 1;
    }

    while ((b1 & 1) == 0)
    {
        t->b_bits++;
        b1 >>= 1;
    }

}

////////////////////////////////////////////////////////////////////////////
// grabX11Init
//      This function should be called once when the grab function is
//      first called.  It sets up the X11 variables needed to grab the
//      window contents.
////////////////////////////////////////////////////////////////////////////
int grabX11Init(int w_id)
{
    unsigned int size;
    
    /* Get X windows stuff. */
    XWindowAttributes attr;
    
    // Open the X windows display.
    if ((t.display = XOpenDisplay((char *)NULL)) == NULL)
    {
	// Unable to open display
	log_printf(log_error, "Unable to open XWindows display\n");
        goto backout_base;
    }

    // Get a handle to the screen
    t.screen = DefaultScreen(t.display);

    // Get the default visual
    t.visual = DefaultVisual(t.display, t.screen);

    // Get reference to the window
    t.window = (Window) w_id;
        
    // Get information on the window.
    if (XGetWindowAttributes(t.display, t.window, &attr) != True)
    {
        log_printf(log_error, "Unable to get window attribues\n");
        goto backout_display;
    }

    t.Nx = attr.width;
    t.Ny = attr.height;

    // Get the depth of the visual
    switch(DefaultDepth(t.display, t.screen))
    {
        default:
            log_printf(log_error, "Depth of %d is not supported.\n", DefaultDepth(t.display, t.screen));
            return(-1);
        case 24:
            t.depth = depth24;
            break;
        case 32:
            t.depth = depth32;
            break;
    }

    // Allocate space for the image
    size = t.Nx * t.Ny * pixelSize[t.depth];
    if ((t.pixelData = (unsigned char *) malloc(size*sizeof(char))) == 0)
    {
        log_printf(log_error, "Unable to allocate image.\n");
        goto backout_display;
    }
    memset(t.pixelData, 0, size*sizeof(char));

    if ((t.rgbData = (unsigned char *) malloc(t.Nx * t.Ny * 3)) == 0)
    {
        log_printf(log_error, "Unable to allocate image.\n");
        goto backout_image_data;
    }
    
    // Create image for rendering slides
    t.image = XCreateImage(
                         t.display,
                         t.visual,
                         intDepth[t.depth],
                         ZPixmap,
                         0,  /* offset. */
                         (char *) t.pixelData,
                         t.Nx,
                         t.Ny,
                         8,   /* bitmap pad. */
                         0);  /* bytes per line */

    if (t.image == None)
    {
        log_printf(log_error, "Unable to allocate image.\n");
        goto backout_rgb_data;
    }

    getImageColorExtents(&t);
    XInitImage(t.image);

    // Successful initialization
    tInit = 1;
    return(0);

 backout_image_data:

    free(t.pixelData);

 backout_rgb_data:

    free(t.rgbData);
    
 backout_display:
    
    XCloseDisplay(t.display);
    
 backout_base:

    return(-1);
}

////////////////////////////////////////////////////////////////////////////
// grabX11Uninit()
//      This function should be called when a reinitialization is
//      necessary, such as when the window size changes.
////////////////////////////////////////////////////////////////////////////
int grabX11Uninit()
{
    free(t.pixelData);
    free(t.rgbData);
    XCloseDisplay(t.display);
    tInit = 0;
    return(0);
}

////////////////////////////////////////////////////////////////////////////
// grabX11Width
//      Returns width of X11 frames.
////////////////////////////////////////////////////////////////////////////
int grabX11Width()
{
    return(t.Nx);
}

////////////////////////////////////////////////////////////////////////////
// grabX11Height
//      Returns height of X11 frames.
////////////////////////////////////////////////////////////////////////////
int grabX11Height()
{
    return(t.Ny);
}

////////////////////////////////////////////////////////////////////////////
// grabX11Image
//      Grabs the image from the winow.  Makes sure X11 has been
//      initialized properly.
////////////////////////////////////////////////////////////////////////////
unsigned char *grabX11Image(int win_id, int rawMode)
{
    unsigned short *imgPtr16;
    unsigned int *imgPtr24;
    unsigned char *rgb;
    unsigned int x, y;

    XWindowAttributes attr;
    
    if (tInit == 0)
    {
        if (grabX11Init(win_id) != 0)
        {
            return(0);
        }
    }
    else
    {
        // Make sure the window size has not changed.
        if (XGetWindowAttributes(t.display, t.window, &attr) != True)
        {
            return(0);
        }
        if ((attr.width != (signed) t.Nx) || (attr.height != (signed) t.Ny))
        {
            grabX11Uninit();
            if (grabX11Init(win_id) != 0)
            {
                return(0);
            }
        }
    }

    if (XGetSubImage(t.display, t.window, 0, 0, t.Nx, t.Ny, 0xffffffff, ZPixmap, t.image, 0, 0) == 0)
    {
        log_printf(log_error, "unable to get image.\n");
        return(0);
    }

    if (!rawMode)
    {
        /* Translate RGBA data. */
        
        /* Copy the frame to our RGB data store. */
        switch (t.depth)
        {
            case depth16:
                rgb = ((unsigned char *) t.rgbData);
                for (y=0; y<t.Ny; y++)
                {
                    imgPtr16 = ((unsigned short *) t.pixelData) + (t.Ny-1-y)*t.Nx;
                    for (x=0; x<t.Nx; x++)
                    {
                        *(rgb++) = (*imgPtr16 & t.r_mask) >> t.r_shift;  // Red
                        *(rgb++) = (*imgPtr16 & t.g_mask) >> t.g_shift;  // Green
                        *(rgb++) = (*imgPtr16 & t.b_mask) >> t.b_shift;  // Blue
                        imgPtr16++;
                    }
                }
                break;

            case depth24:
            case depth32:
                rgb = ((unsigned char *) t.rgbData);
                for (y=0; y<t.Ny; y++)
                {
                    imgPtr24 = ((unsigned int *) t.pixelData) + (t.Ny-1-y)*t.Nx;
                    for (x=0; x<t.Nx; x++)
                    {
                        *(rgb++) = (*imgPtr24 & t.r_mask) >> t.r_shift;  // Red
                        *(rgb++) = (*imgPtr24 & t.g_mask) >> t.g_shift;  // Green
                        *(rgb++) = (*imgPtr24 & t.b_mask) >> t.b_shift;  // Blue
                        imgPtr24++;
                    }
                }
                break;
                
            default:
                log_printf(log_error, "Invalid depth?\n");
                return(0);
        }

        return(t.rgbData);
    }
    else
    {
        return(t.pixelData);
    }
}

