////////////////////////////////////////////////////////////////////////////
// main.cc
//      Main program entry.
//
// Author: Jon W. Wallace, Lafayette College
// Date:   May 2020
////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <signal.h>
#include <string.h>

#include "main.h"
#include "log.h"
#include "capture.h"
#include "process.h"
#include "udp_send_hmd_pos.h"
#include "udp_recv_stats.h"
#include "xplaneXfer.h"
#include "aruco.h"
#include "nve.h"

//#define MIN(x,y) (((x) < (y)) ? (x) : (y))

mainOptionsDef mainOptions;
volatile int stopping;

////////////////////////////////////////////////////////////////////////////
// printUsage
//      Prints out some help.
////////////////////////////////////////////////////////////////////////////
void printUsage(const char *name, int longHelp)
{
    fprintf(stderr, "usage: %s [options]\n", name);
    if (longHelp) 
    {
        fprintf(stderr, "\n");
        fprintf(stderr, "Options:\n");
        fprintf(stderr, "   -h                  This help\n");
        fprintf(stderr, "   -a [skip]           Aruco frames to skip.\n");
        fprintf(stderr, "   -d                  Do not undistort cameras.\n");
        fprintf(stderr, "   -g                  Enable GPU accelerations.\n");
        fprintf(stderr, "   -c [dev]{:[w]x[h]}  Set capture device, width/height.\n");
        fprintf(stderr, "   -r [w]x[h]          Set render width/height.\n");
        fprintf(stderr, "   -v [level]          Set level of output messages.\n");
        fprintf(stderr, "   -w [wid]            Set capture X11 window ID.\n");
        fprintf(stderr, "   -o [spec]           Output rendering (additive)\n");
        fprintf(stderr, "                       v => VR\n");
        fprintf(stderr, "                       x => X11\n");
        fprintf(stderr, "                       f:[name]:[type] => file \n");
        fprintf(stderr, "   -C [clip_amt]       Clip fraction for X or file render.\n");
        fprintf(stderr, "   -f [encoder_ops]    Set encoder ops.  Use -e ? for help.\n");
        fprintf(stderr, "   -s [skip]           Capture frames to skip\n");
        fprintf(stderr, "   -G                  Do green (chroma key) det\n");
        fprintf(stderr, "   -H [port]           Send HMD pose to FG UDP port\n");
        fprintf(stderr, "   -S [port]           Receive aircraft status port\n");
        fprintf(stderr, "   -X                  Connect to X-plane\n");
        fprintf(stderr, "   -I                  Turn on instrument panel\n");
    }
}

////////////////////////////////////////////////////////////////////////////
// mainWaitAndCleanup
//      Kill threads and cleanup.  Put in a function to avoid duplicating
//      code in signal handler and main().
////////////////////////////////////////////////////////////////////////////
void mainWaitAndCleanup(const char *exit_message)
{
    int allDone;
    int i;

    // Stop the capture thread. 
    captureStop();

    // Stop processing thread
    processUninit();
    
    // Stop aruco processing thread
    if (mainOptions.inst_panel)
    {
        arucoUninit();
    }

    if (mainOptions.xplane)
    {
        xplaneXferUninit();
    }
    
    if (mainOptions.fg_stat_port)
    {
        udpRecvStatUninit();
    }
    
    if (mainOptions.fg_hmd_port)
    {
        udpSendHmdUninit();
    }
    
    // Print profiling statistics.
    if (mainOptions.do_profile) 
    {
        captureProfilePrint();
        if (mainOptions.inst_panel)
        {
            arucoProfilePrint();
        }
        processProfilePrint();
    }

    // Done. 
    log_printf(log_info, "%s\n", exit_message);
}

////////////////////////////////////////////////////////////////////////////
// signalHandler
//     Try to gracefully exit on termination.  Kill threads and cleanup.
////////////////////////////////////////////////////////////////////////////
void signalHandler(int signum)
{
    static int sig_handler_called = 0;

    // If it is TERM (^C) then clear the signal, and exit gracefully.
    if (signum == SIGINT)
    {
        //fprintf(stderr, "Got SIGINT.\n");
        signal(signum, signalHandler);
        stopping = 1;
    }
    else
    {
        //fprintf(stderr, "Signal %d.\n", signum);
        
        // Only allow signal handler to be called once. 
        if (sig_handler_called) return;
        sig_handler_called = 1;

        // Do cleanup.
        log_printf(log_quiet, "\n");
        mainWaitAndCleanup("Aborted");

        // Exit application.
        exit(0);
    }
}

////////////////////////////////////////////////////////////////////////////
// signalInit
//      Initialize signal handler.  Allows for a graceful exit when we
//      are terminated.
////////////////////////////////////////////////////////////////////////////
void signalInit()
{
    // Put in our signal handler.
    signal(SIGBUS, signalHandler);
    signal(SIGTRAP, signalHandler);
    signal(SIGSEGV, signalHandler);
    signal(SIGABRT, signalHandler);
    signal(SIGFPE, signalHandler);
    signal(SIGTERM, signalHandler);
    signal(SIGINT, signalHandler);
}

////////////////////////////////////////////////////////////////////////////
// main()
//      Main entry of program.
////////////////////////////////////////////////////////////////////////////
int main(int argc, char *argv[])
{
    const char *pos, *pos2;
    int ccode;
    int i, len;
    int opt;
    
    // Set default options.
    memset(&mainOptions, 0, sizeof(mainOptions));
    mainOptions.cap_width = 960;
    mainOptions.cap_height = 950;
    mainOptions.render_width = 640;
    mainOptions.render_height = 480;
    mainOptions.render_x11 = 0;
    mainOptions.render_vr = 0;
    
    // Get command-line options.
    while ((opt = getopt(argc, argv, "c:df:ghpqr:s:v:w:o:C:GH:IS:X")) != -1)
    {
        switch (opt)
        {
            case 'd':
                mainOptions.no_undistort = 1;
                break;
            
            case 'c':
                pos = strstr(optarg, ":");
                if (pos == 0)
                {
                    pos = optarg + strlen(optarg);
                }

                len = MIN(pos-optarg, (sizeof(mainOptions.cap_dev)-1));
                memcpy(mainOptions.cap_dev, optarg, len);
                mainOptions.cap_dev[len] = 0;

                if (*pos != 0)
                {
                    pos++;
                    i = sscanf(pos, "%dx%d", &mainOptions.cap_width, &mainOptions.cap_height);
                    if (i < 2)
                    {
                        log_printf(log_error, "Invalid argument for capture size.\n");
                        ccode = -1;
                        goto backout_base;
                    }
                }
                break;
            case 'r':
                i = sscanf(optarg, "%dx%d", &mainOptions.render_width, &mainOptions.render_height);
                if (i < 2)
                {
                    log_printf(log_error, "Invalid argument for render size.\n");
                    ccode = -1;
                    goto backout_base;
                }
                break;
            case 's':
                mainOptions.skip_time = atoi(optarg);
                break;
            case 'g':
                mainOptions.gpu = 1;
                break;
            case 'h':
                printUsage(argv[0], 1);
                ccode = -1;
                goto backout_base;

            case 'v':
                mainOptions.log_level = atoi(optarg);
                break;

            case 'o':
                switch(optarg[0])
                {
                    case 'x':
                        mainOptions.render_x11 = 1;
                        break;

                    case 'v':
                        mainOptions.render_vr = 1;
                        break;

                    case 'f':
                        mainOptions.render_file_type = 'l';
                        pos = strstr(optarg, ":");
                        if (pos == 0)
                        {
                            log_printf(log_error, "No filename supplied for file rendering.\n");
                            ccode = -1;
                            goto backout_base;
                        }
                        pos++;
                        pos2 = strstr(pos, ":");
                        if (pos2 == 0)
                        {
                            pos2 = optarg + strlen(optarg);
                        }
                        len = MIN(pos2-pos, sizeof(mainOptions.render_file_fname)-1);
                        memcpy(mainOptions.render_file_fname, pos, len);
                        mainOptions.render_file_fname[len] = 0;
                        
                        if (*pos2 != 0) 
                        {
                            mainOptions.render_file_type = *(pos2+1);
                            if ((mainOptions.render_file_type != 'l') &&
                                (mainOptions.render_file_type != 'r') &&
                                (mainOptions.render_file_type != 'b'))
                            {
                                log_printf(log_error, "Invalid render file type.  Should be 'l', 'r', or 'b'.\n");
                                goto backout_base;
                            }
                        }

                        mainOptions.render_file = 1;
                        break;
                        
                    default:
                        log_printf(log_error, "Invalid render specification: %s.\n", optarg);
                        break;
                }
                break;

            case 'C':
                if ((sscanf(optarg, "%f", &mainOptions.render_clip) != 1) ||
                    (mainOptions.render_clip <= 0.0) ||
                    (mainOptions.render_clip >= 1.0))
                {
                    log_printf(log_error, "Invalid render clip specified.\n");
                    goto backout_base;
                }
                
                break;
            case 'f':
                if (optarg[0] == '?')
                {
                    nveHelp();
                    goto backout_base;
                }
                mainOptions.render_file_options = optarg;
                break;
                
            case 'p':
                mainOptions.do_profile = 1;
                break;

            case 'q':
                mainOptions.quiet = 1;
                break;
                
            case 'x':
                mainOptions.do_x_output = 1;
                break;

            case 'w':
                if ((optarg[0] == '0') && (optarg[1] == 'x'))
                {
                    if (sscanf(optarg, "%x", &mainOptions.x_wid) != 1)
                    {
                        log_printf(log_error, "Invalid window ID.\n");
                        ccode = -2;
                        goto backout_base;
                    }
                }
                else
                {
                    mainOptions.x_wid = atoi(optarg);
                }
                log_printf(log_info, "Grabbing X Window ID 0x%x\n", mainOptions.x_wid);
                break;
            case 'G':
                mainOptions.green_screen = 1;
                break;
            case 'H':
                mainOptions.fg_hmd_port = atoi(optarg);
                break;
            case 'I':
                mainOptions.inst_panel = 1;
                mainOptions.inst_panel_aruco_dict = 0;
                break;
            case 'S':
                mainOptions.fg_stat_port = atoi(optarg);
                break;
            case 'X':
                mainOptions.xplane = 1;
                break;
            default:
                log_printf(log_error, "Invalid argument.\n");
                ccode = -2;
                goto backout_base;
        }
    }

    // Only allow one interface X-plane or FG
    if (mainOptions.xplane)
    {
        mainOptions.fg_stat_port = 0;
        mainOptions.fg_hmd_port = 0;
    }

    // Initialize reporting log
    log_init((log_level_enum) mainOptions.log_level);
    
    // Initialize capture
    if (captureInit() != 0)
    {
        log_printf(log_error, "Unable to initialize capture.\n");
        ccode = -3;
        goto backout_base;
    }

    // Initialize Aruco tag processing
    if (mainOptions.inst_panel)
    {
        if (arucoInit(mainOptions.inst_panel_aruco_dict, mainOptions.cap_width, mainOptions.cap_height) != 0)
        {
            log_printf(log_error, "Unable to initialize aruco module.\n");
            ccode = -3;
            goto backout_capture;
        }
    }
    
    // Initialize processing functions
    if (processInit() != 0)
    {
        log_printf(log_error, "Unable to initialize processing module.\n");
        ccode = -3;
        goto backout_aruco;
    }
    
    if (mainOptions.fg_hmd_port)
    {
        if (udpSendHmdInit(mainOptions.fg_hmd_port) != 0)
        {
            log_printf(log_error, "Unable to initialize UDP port for HMD position.\n");
            goto backout_proc;
        }
    }

    if (mainOptions.fg_stat_port)
    {
        if (udpRecvStatInit(mainOptions.fg_stat_port) != 0)
        {
            log_printf(log_error, "Unable to initialize receive UDP port for aircraft stats.\n");
            goto backout_udp1;
        }
    }

    if (mainOptions.xplane)
    {
        if (xplaneXferInit() != 0)
        {
            log_printf(log_error, "Unable to initialize X-plane connection.\n");
            goto backout_udp2;
        }
    }
    
    // Start capture thread
    if (captureStart() != 0)
    {
        // Shouldn't happen...
        exit(0);
    }

    // Everything started OK, intialize signal handler.
    stopping = 0;
    signalInit();

    // Run our main processing loop
    processRun();

    /* Wait for completion and cleanup. */
    mainWaitAndCleanup("Complete");

    /* Success. */
    ccode = 0;
    return(ccode);
    
 backout_xplane:
    if (mainOptions.xplane)
    {
        xplaneXferUninit();
    }
    
 backout_udp2:
    if (mainOptions.fg_stat_port)
    {
        udpRecvStatUninit();
    }
    
 backout_udp1:
    if (mainOptions.fg_hmd_port)
    {
        udpSendHmdUninit();
    }
    
 backout_proc:
    processUninit();

 backout_aruco:
    if (mainOptions.inst_panel)
    {
        arucoUninit();
    }
    
 backout_capture:
    captureUninit();

 backout_base:
    return(ccode);
}

