/*============================================================================
 * myProfile.cc
 *      Code to allow some simple profiling.
 *
 * Author: Jon W. Wallace, Lafayette College
 * Date:   May 2020
 *==========================================================================*/

#include <time.h>
#include <string.h>
#include <pthread.h>
#include <stdio.h>
#include <inttypes.h>

#include "log.h"
#include "myProfile.h"

#ifdef __WINDOWS__
HANDLE profSampleThread;
#else
pthread_t profSampleThread;
#endif

#define CONST_1E9       1000000000

//#define CLOCK_WALL    CLOCK_MONOTONIC
#define CLOCK_WALL      CLOCK_REALTIME

/*----------------------------------------------------------------------------
 * profCpuInit
 *      Initializes a profiling variable.
 *--------------------------------------------------------------------------*/
void profCpuInit(cpuProfDef *p)
{
    memset(p, 0, sizeof(cpuProfDef));
    p->io_count = 0;
}

/*----------------------------------------------------------------------------
 * profWallInit
 *      Initializes a profiling variable.
 *--------------------------------------------------------------------------*/
void profWallInit(wallProfDef *p)
{
    memset(p, 0, sizeof(wallProfDef));
}

/*----------------------------------------------------------------------------
 * profCpuStart
 *      Start profiling block.
 *--------------------------------------------------------------------------*/
void profCpuStart(cpuProfDef *p)
{
#ifdef __WINDOWS__
    p->start = GetTickCount();
#else
    clock_gettime(CLOCK_THREAD_CPUTIME_ID, &p->start);
#endif    
    //p->start = clock();
}

/*----------------------------------------------------------------------------
 * profWallStart
 *      Start profiling block.
 *----------------------------------------------------------------------------*/
void profWallStart(wallProfDef *p)
{
#ifdef __WINDOWS__
    p->start = GetTickCount();
#else
    clock_gettime(CLOCK_WALL, &p->start);
#endif    
}

/*----------------------------------------------------------------------------
 * profCpuEnd
 *      End profiling block.
 *--------------------------------------------------------------------------*/
void profCpuEnd(cpuProfDef *p, io_int count)
{
#ifdef __WINDOWS__
    p->end = GetTickCount();
    p->total += (p->end - p->start);
#else    
    clock_gettime(CLOCK_THREAD_CPUTIME_ID, &p->end);
    p->total.tv_sec += (p->end.tv_sec - p->start.tv_sec);
    p->total.tv_nsec += (p->end.tv_nsec - p->start.tv_nsec);
    if (p->total.tv_nsec < 0)
    {
        p->total.tv_nsec += CONST_1E9;
        p->total.tv_sec--;
    }
    else if (p->total.tv_nsec >= CONST_1E9)
    {
        p->total.tv_nsec -= CONST_1E9;
        p->total.tv_sec++;
    }
#endif    
    p->io_count += count;
}

/*----------------------------------------------------------------------------
 * profWallEnd
 *      End profiling block.
 *--------------------------------------------------------------------------*/
void profWallEnd(wallProfDef *p)
{
#ifdef __WINDOWS__
    p->end = GetTickCount();
    p->total += (p->end - p->start);
#else    
    clock_gettime(CLOCK_WALL, &p->end);
    p->total.tv_sec += (p->end.tv_sec - p->start.tv_sec);
    p->total.tv_nsec += (p->end.tv_nsec - p->start.tv_nsec);
    if (p->total.tv_nsec < 0)
    {
        p->total.tv_nsec += CONST_1E9;
        p->total.tv_sec--;
    }
    else if (p->total.tv_nsec >= CONST_1E9)
    {
        p->total.tv_nsec -= CONST_1E9;
        p->total.tv_sec++;
    }
#endif    
}

/*----------------------------------------------------------------------------
 * profTimeSec
 *      Returns the amount of time in seconds stored in the profile
 *      record.
 *--------------------------------------------------------------------------*/
double profTimeSec(cpuProfDef *p)
{
    //return(p->total/(double)CLOCKS_PER_SEC);
#ifdef __WINDOWS__
    return((double)p->total/(double)1000.0);
#else    
    return((double)p->total.tv_sec +
           (double)p->total.tv_nsec/(double)CONST_1E9);
#endif    
}

double profTimeSec(wallProfDef *p)
{
#ifdef __WINDOWS__
    return((double)p->total/(double)1000.0);
#else
    return((double)p->total.tv_sec +
           (double)p->total.tv_nsec/(double)CONST_1E9);
#endif    
}

/*----------------------------------------------------------------------------
 * profPrintOne
 *      Prints a single profiling value.
 *--------------------------------------------------------------------------*/
void profPrintOne(const char *name, cpuProfDef *time1, cpuProfDef *timeAll, int numFrames)
{
    double time1_d;
    double timeAll_d;
    double timePerFrame_d;
    double io_mb;
    double io_mb_frame;
    
    time1_d = profTimeSec(time1);
    timeAll_d = profTimeSec(timeAll);
    timePerFrame_d = time1_d/(double)numFrames;
    io_mb = (double)(time1->io_count)/1024.0/1024.0;
    io_mb_frame = io_mb/(double)numFrames;

    log_printf(log_quiet,
            "%-20s %8.2lf s %8.3lf s/frame (%5.1lf %%) |",
            name,
            time1_d,
            timePerFrame_d,
            time1_d/timeAll_d*100.0);

    /* Print total IO. */
    if (io_mb > 1024.0)
    {
        /* Print GB. */
        log_printf(log_quiet, " %6.2lfGB ", io_mb/1024.0);
    }
    else
    {
        /* Print MB. */
        log_printf(log_quiet, " %6.2lfMB ", io_mb);
    }

    /* Print per frame. */
    log_printf(log_quiet, "%6.2lfMB/frame\n", io_mb_frame);
}

/*----------------------------------------------------------------------------
 * profPrintOne
 *      Prints a single profiling value.
 *--------------------------------------------------------------------------*/
void profPrintOne(const char *name, wallProfDef *time1, wallProfDef *timeAll, int numFrames)
{
    double time1_d;
    double timeAll_d;
    double timePerFrame_d;
    
    time1_d = profTimeSec(time1);
    timeAll_d = profTimeSec(timeAll);
    timePerFrame_d = time1_d/(double)numFrames;

    log_printf(log_quiet,
            "%-20s %8.2lf s %8.3lf s/frame (%5.1lf %%)\n",
            name,
            time1_d,
            timePerFrame_d,
            time1_d/timeAll_d*100.0
            );
}

/*----------------------------------------------------------------------------
 * profWallTime
 *      Start profiling block.
 *----------------------------------------------------------------------------*/
double profWallTime()
{
#ifdef __WINDOWS__
    DWORD t;
    t = GetTickCount();
    return((double)t/(double)1000.0);
#else
    struct timespec t;
    clock_gettime(CLOCK_WALL, &t);
    return((double)t.tv_sec +
           (double)t.tv_nsec/(double)CONST_1E9);
#endif
}
