////////////////////////////////////////////////////////////////////////////
// nve.cc
//      Wrapper to allow use of the NvEncodeGL example from the
//      Nvidia SDK, allowing hardware encoding of rendered video.
//
// Author: Jon W. Wallace, Lafayette College
// Date:   May 2020
////////////////////////////////////////////////////////////////////////////

#include "NvEncoderGL.h"
#include "NvEncoderCLIOptions.h"
#include "NvCodecUtils.h"

#include "nve.h"
#include "log.h"

#define WAIT_SHORT      usleep(1000)
#define CONST_1E9       1000000000
#define CONST_1E6       1000000

simplelogger::Logger *logger = simplelogger::LoggerFactory::CreateConsoleLogger();
FILE *nve_f_ts = 0;
FILE *nve_f = 0;
NvEncoderGL *nve = 0;
int nFrameSize = 0;
int nveWidth;
int nveHeight;

int nve_t_started;
struct timespec nve_t_start;

#define ENC_FORMAT NV_ENC_BUFFER_FORMAT_ABGR
//#define ENC_FORMAT NV_ENC_BUFFER_FORMAT_ARGB
//#define ENC_FORMAT NV_ENC_BUFFER_FORMAT_IYUV

NvEncoderInitParam encodeCLIOptions;

char *str_buffer;

////////////////////////////////////////////////////////////////////////////
// nveInit()
//      Initialize the NVidia hardware encoder.
////////////////////////////////////////////////////////////////////////////
int nveInit(const char *fname, int width, int height, const char *encode_options)
{
    int ccode;
    int len;

    if (encode_options)
    {
        encodeCLIOptions = NvEncoderInitParam(encode_options);
    }

    nveWidth = width;
    nveHeight = height;

    NV_ENC_INITIALIZE_PARAMS initializeParams = { NV_ENC_INITIALIZE_PARAMS_VER };
    NV_ENC_CONFIG encodeConfig = { NV_ENC_CONFIG_VER };

    // Open the output video file
    if ((nve_f = fopen(fname, "wb")) == 0)
    {
        log_printf(log_error, "nveInit(): Unable to open output '%s'.\n", fname);
        ccode = -1;
        goto backout_base;
    }

    // Open a file for timecodes
    len = strlen(fname);
    str_buffer = (char *) malloc(len+10);
    memcpy(str_buffer, fname, len);
    memcpy(str_buffer + len, ".ts", 4);
    if ((nve_f_ts = fopen(str_buffer, "wb")) == 0)
    {
        log_printf(log_error, "nveInit(): Unable to open timestamp output '%s'.\n", fname);
        ccode = -1;
        free(str_buffer);
        goto backout_file;
    }
    free(str_buffer);
    
    nve = new NvEncoderGL(nveWidth, nveHeight, ENC_FORMAT);

    initializeParams.encodeConfig = &encodeConfig;
    nve->CreateDefaultEncoderParams(&initializeParams, encodeCLIOptions.GetEncodeGUID(), encodeCLIOptions.GetPresetGUID());

    encodeCLIOptions.SetInitParams(&initializeParams, ENC_FORMAT);

    // Print out initialization parameters
    // std::cout << encodeCLIOptions.FullParamToString(initializeParams);

    nve->CreateEncoder(&initializeParams);

    nFrameSize = nve->GetFrameSize();
    log_printf(log_info, "nveInit(): nFrameSize=%d\n", nFrameSize);

    // Set this flag to 0.  On the first frame, we will output
    // the starting global time as a comment in the .ts file.
    nve_t_started = 0;
    
    return(0);

 backout_encoder:
    delete(nve);
    nve = 0;

 backout_file2:
    fclose(nve_f_ts);
    nve_f_ts = 0;
    
 backout_file:
    fclose(nve_f);
    nve_f = 0;
    
 backout_base:
    return(ccode);
}

////////////////////////////////////////////////////////////////////////////
// nveWriteTimeStamp
////////////////////////////////////////////////////////////////////////////
void nveWriteTimeStamp()
{
    struct timespec t_diff;
    struct timespec t;
    
    // Get current timestamp
    clock_gettime(CLOCK_REALTIME, &t);
    // First frame?
    if (!nve_t_started)
    {
        // Yes.  Store starting real time and output header.
        memcpy(&nve_t_start, &t, sizeof(struct timespec));
        fprintf(nve_f_ts, "# timecode format v2\n");
        fprintf(nve_f_ts, "# start_time_s_ns %ld %ld\n", t.tv_sec, t.tv_nsec);
        fprintf(nve_f_ts, "0\n");
        nve_t_started = 1;
    }
    else
    {
        // Not first frame, compute difference from start time.
        t_diff.tv_sec = (t.tv_sec - nve_t_start.tv_sec);
        t_diff.tv_nsec = (t.tv_nsec - nve_t_start.tv_nsec);
        if (t_diff.tv_nsec < 0)
        {
            t_diff.tv_nsec += CONST_1E9;
            t_diff.tv_sec--;
        }
        else if (t_diff.tv_nsec >= CONST_1E9)
        {
            t_diff.tv_nsec -= CONST_1E9;
            t_diff.tv_sec++;
        }

        // Output in ms
        fprintf(nve_f_ts, "%.2lf\n", ((double)t_diff.tv_sec*1000.0 + t_diff.tv_nsec/(double)(CONST_1E6)));
    }
}

////////////////////////////////////////////////////////////////////////////
// nveUninit
////////////////////////////////////////////////////////////////////////////
int nveUninit()
{
    nveWriteTimeStamp();
    
    // Write the data we have left
    std::vector<std::vector<uint8_t>> vPacket;
    nve->EndEncode(vPacket);

    for (std::vector<uint8_t> &packet : vPacket)
    {
        fwrite(reinterpret_cast<char*>(packet.data()), packet.size(), 1, nve_f);
    }

    nve->DestroyEncoder();
    
    delete(nve);
    fclose(nve_f);
    fclose(nve_f_ts);
    
    nve_f = 0;
    nve = 0;
}

////////////////////////////////////////////////////////////////////////////
// nveHelp()
////////////////////////////////////////////////////////////////////////////
void nveHelp()
{
    std::cout << encodeCLIOptions.GetHelpMessage();
}

////////////////////////////////////////////////////////////////////////////
// nveEncode()
////////////////////////////////////////////////////////////////////////////
int nveEncode(glhFbDef *fb)
{
    int i;
    int bytes_written;
    struct timespec t;
    
    nveWriteTimeStamp();
    
    const NvEncInputFrame* encoderInputFrame = nve->GetNextInputFrame();
    NV_ENC_INPUT_RESOURCE_OPENGL_TEX *pResource = (NV_ENC_INPUT_RESOURCE_OPENGL_TEX *)encoderInputFrame->inputPtr;

    // Fast GPU-only copy
    glCopyImageSubData(fb->tex_id,
                       GL_TEXTURE_2D,
                       0, // Level
                       0, 0, 0, // X,Y,Z
                       pResource->texture,
                       GL_TEXTURE_RECTANGLE,
                       0, // Level
                       0, 0, 0, // X,Y,Z
                       nveWidth, nveHeight, 1);

    // Slow CPU copy.
    /*
    glBindFramebuffer(GL_FRAMEBUFFER, fb->fbo_id);
    glReadPixels(0, 0, x_render_width, x_render_height, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    glBindTexture(GL_TEXTURE_RECTANGLE, pResource->texture);
    glTexSubImage2D(GL_TEXTURE_RECTANGLE, 0, 0, 0, x_render_width*4, x_render_height, GL_RED, GL_UNSIGNED_BYTE, buffer);
    glBindTexture(GL_TEXTURE_RECTANGLE, 0);
    */

    std::vector<std::vector<uint8_t>> vPacket;
    nve->EncodeFrame(vPacket);

    bytes_written = 0;
    for (std::vector<uint8_t> &packet : vPacket)
    {
        fwrite(reinterpret_cast<char*>(packet.data()), packet.size(), 1, nve_f);
        bytes_written += packet.size();
    }

    return(bytes_written);
}
