/*============================================================================
 * openvr_render.
 *      Interface for rendering and localization using OpenVR.
 *
 * Author: Jon W. Wallace, Lafayette College
 * Date:   May 2020
 *==========================================================================*/

#include <openvr/openvr.h>

#include <unistd.h>
#include <stdio.h>

#include "openvr_render.h"

//#include <SDL2/SDL.h>

#include <GL/glew.h>
#include <GL/glu.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <GL/gl.h>
#include <GL/glx.h>
#include <X11/keysym.h>

#include "myProfile.h"
#include "ar_global_data.h"

#include "Vectors.h"
#include "Matrices.h"

#include "kbd.h"

#include "udp_send_hmd_pos.h"

#include "c172_win1.cc"

void render_cubes1();
void render_cubes2();

using namespace vr;

#define SQ(x) ((x)*(x))

typedef struct
{
    uint32_t width;
    uint32_t height;
    GLuint tex_id;
    GLuint fbo_id;
    GLuint rbo_id;
    Texture_t tex;
} fbDef;

// Global Data
static IVRSystem *pVr=0;
static fbDef fb_cap_lr;
static fbDef fb_l, fb_r;
static fbDef fb_s_l, fb_s_r;
static fbDef fb_grab;
static TrackedDevicePose_t trackedDevicePose[k_unMaxTrackedDeviceCount];
static TrackedDevicePose_t trackedDevicePose_prev[k_unMaxTrackedDeviceCount];
static int quitting = 0;

// Orientation matrix and origin of home headset position.  Should be pilot's
// normal head position when looking straight forward.
Vector3 pilot_x(1, 0, 0);
Vector3 pilot_y(0, 1, 0);
Vector3 pilot_z(0, 0, 1);
Vector3 pilot_origin(0, 0, 0);
const float pilot_fov_x_default = 100.0*M_PI/180.0;
const float pilot_fov_y_default = 100.0*M_PI/180.0;
float pilot_fov_x = pilot_fov_x_default;
float pilot_alpha_x = 2.0*tanf(pilot_fov_x/2.0);
float pilot_fov_y = pilot_fov_y_default;
float pilot_alpha_y = 2.0*tanf(pilot_fov_y/2.0);
//float pilot_ipd = (float) 3.5*2.54/100.0;
//float pilot_ipd = (float) 6.0/100.0;
float pilot_fov_z_near = 0.1;
float pilot_fov_z_far = 5.0;

// Position of cockpit window in global
float win_tris_vertex_map[N_poly_win_tris][N_vert_win_tris][3];

static FILE *f_cal;

Matrix4 M_eye_pose_l;
Matrix4 M_eye_pose_r;
Matrix4 M_proj_l;
Matrix4 M_proj_r;
Matrix4 M_pose;

// Positioning of scenery on output texture
float scen_lx = 0.52;
float scen_ly = 0.46;
float scen_lw = 0.63;
float scen_lh = 0.56;

float scen_rx = 0.48;
float scen_ry = 0.46;
float scen_rw = 0.63;
float scen_rh = 0.56;

int fbNew(fbDef *fb, uint32_t width, uint32_t height, uint32_t initial_abgr, int rend_flag, int bgra)
{
    uint32_t *data;
    uint32_t ii;
    int ccode;

    memset(fb, 0, sizeof(fbDef));
    fb->width = width;
    fb->height = height;
    
    if ((data = (uint32_t *) malloc(width*height*sizeof(uint32_t))) == 0)
    {
        return(-1);
    }

    for (ii=0; ii<width*height; ii++)
    {
        data[ii] = initial_abgr;
    }

    // Generate the associated texture object
    glGenTextures(1, &(fb->tex_id));
    if (fb->tex_id == GL_INVALID_VALUE)
    {
        fprintf(stderr, "Unable to create OpenGL texture.\n");
        ccode = -1;
        goto backout_data;
    }

    // Put initial data in texture
    glBindTexture(GL_TEXTURE_2D, fb->tex_id);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
    
    // Set some default parameters for the texture.
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glBindTexture(GL_TEXTURE_2D, 0);

    // Fill out a reference for headset copy
    fb->tex.eColorSpace = ColorSpace_Gamma;
    fb->tex.eType = TextureType_OpenGL;
    fb->tex.handle = (void*)(uintptr_t)(fb->tex_id);

    if (rend_flag)
    {
        // Create framebuffer / render structures that we need in order
        // to render scenes to this texture.

        // create a framebuffer object
        glGenFramebuffers(1, &(fb->fbo_id));
        glBindFramebuffer(GL_FRAMEBUFFER, fb->fbo_id);

        // attach the texture to FBO color attachment point
        glFramebufferTexture2D(GL_FRAMEBUFFER,        // 1. fbo target: GL_FRAMEBUFFER
                               GL_COLOR_ATTACHMENT0,  // 2. attachment point
                               GL_TEXTURE_2D,         // 3. tex target: GL_TEXTURE_2D
                               fb->tex_id,            // 4. tex ID
                               0);                    // 5. mipmap level: 0(base)

        // check FBO status
        if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
        {
            fprintf(stderr, "Creation of framebuffer failed.\n");
            ccode = -2;
            goto backout_data;
        }

        // switch back to window-system-provided framebuffer
        glBindFramebuffer(GL_FRAMEBUFFER, 0);
    }
    
    // Success
    ccode = 0;
    
 backout_data:
    free(data);

 backout_base:
    return(ccode);
}

int fbFree(fbDef *fb)
{
    glBindTexture(GL_TEXTURE_2D, 0);
    glDeleteTextures(1, &(fb->tex_id));
    memset(fb, 0, sizeof(fbDef));
}

int fbWrite(fbDef *fb, void *data)
{
    glBindTexture(GL_TEXTURE_2D, fb->tex_id);
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, fb->width, fb->height, GL_RGBA, GL_UNSIGNED_BYTE, (unsigned char *) data);
}

int fbCopy(fbDef *fb_src, fbDef *fb_dest,
           GLint x_src, GLint y_src, GLint x_dest, GLint y_dest,
           GLint width, GLint height)
{
    int ccode;
    
    glCopyImageSubData(fb_src->tex_id, GL_TEXTURE_2D, //srcName, srcTarget
                       0, // srcLevel
                       x_src, y_src, 0,
                       fb_dest->tex_id, GL_TEXTURE_2D,
                       0,
                       x_dest, y_dest, 0,
                       width, height, 1);

    if ((ccode=glGetError()) != 0)
    {
        fprintf(stderr, "glCopyImageSubData failed code: %d\n", ccode);
    }
    return(ccode);
}

void printMatrix(HmdMatrix34_t &matrix)
{
    int i, j;

    for (i=0; i<3; i++)
    {
        for (j=0; j<4; j++)
        {
            fprintf(stdout, "%+4.2f  ", matrix.m[i][j]);
        }
        fprintf(stdout, "\n");
    }
    fprintf(stdout, "----\n");
}

Matrix4 getEyeProjMatrix( vr::Hmd_Eye nEye )
{
    HmdMatrix44_t mat = pVr->GetProjectionMatrix( nEye, pilot_fov_z_near, pilot_fov_z_far);

    return Matrix4(
                   mat.m[0][0], mat.m[1][0], mat.m[2][0], mat.m[3][0],
                   mat.m[0][1], mat.m[1][1], mat.m[2][1], mat.m[3][1], 
                   mat.m[0][2], mat.m[1][2], mat.m[2][2], mat.m[3][2], 
                   mat.m[0][3], mat.m[1][3], mat.m[2][3], mat.m[3][3]
                   );
}

Matrix4 getEyePoseMatrix( vr::Hmd_Eye nEye )
{
	HmdMatrix34_t matEye = pVr->GetEyeToHeadTransform( nEye );
	Matrix4 matrixObj(
		matEye.m[0][0], matEye.m[1][0], matEye.m[2][0], 0.0, 
		matEye.m[0][1], matEye.m[1][1], matEye.m[2][1], 0.0,
		matEye.m[0][2], matEye.m[1][2], matEye.m[2][2], 0.0,
		matEye.m[0][3], matEye.m[1][3], matEye.m[2][3], 1.0f
		);

	return matrixObj.invert();
}

Matrix4 ConvertSteamVRMatrixToMatrix4( const HmdMatrix34_t &matPose )
{
	Matrix4 matrixObj(
		matPose.m[0][0], matPose.m[1][0], matPose.m[2][0], 0.0,
		matPose.m[0][1], matPose.m[1][1], matPose.m[2][1], 0.0,
		matPose.m[0][2], matPose.m[1][2], matPose.m[2][2], 0.0,
		matPose.m[0][3], matPose.m[1][3], matPose.m[2][3], 1.0f
		);
	return matrixObj;
}

/*----------------------------------------------------------------------------
 * get_openGL_context_X()
 *      Code to open the X11 display and use the root window as our
 *      OpenGL context.
 *--------------------------------------------------------------------------*/
int opengl_context = 0;
Display *x11_dpy = 0;
int x11_scr = 0;
Visual *x11_vis;
int get_openGL_context_X()
{
    Window root;
    XVisualInfo *vi;
    GLint attr[] = { GLX_RGBA, GLX_DEPTH_SIZE, 24, GLX_DOUBLEBUFFER, None };
    GLXContext glc;
 
    /* open display */
    if ( ! (x11_dpy = XOpenDisplay(NULL)) ) {
        fprintf(stderr, "cannot connect to X server\n\n");
        return(-1);
    }
 
    /* get root window */
    root = DefaultRootWindow(x11_dpy);

    /* Get screen */
    x11_scr = DefaultScreen(x11_dpy);

    /* Get default visual */
    x11_vis = DefaultVisual(x11_dpy, x11_scr);

    /* get visual matching attr */
    if( ! (vi = glXChooseVisual(x11_dpy, 0, attr)) ) {
        fprintf(stderr, "no appropriate visual found\n\n");
        return(-1);
    }
 
    /* create a context using the root window */
    if ( ! (glc = glXCreateContext(x11_dpy, vi, NULL, GL_TRUE)) ){
        fprintf(stderr, "failed to create context\n\n");
        exit(1);
    }
    glXMakeCurrent(x11_dpy, root, glc);
 
    /* try it out, remember to *NOT* render to the default framebuffer! */
    printf("vendor: %s\n", (const char*)glGetString(GL_VENDOR));

    glewInit();

    return 0;
}
void free_openGL_context_X()
{
    XCloseDisplay(x11_dpy);
}

XImage *grabImage = 0;
Window grabWindow = 0;
int grabNx;
int grabNy;
unsigned char *grabPixelData;
static int grabInitDone = 0;
int grabInit(int w_id)
{
    int size;
    int depth;
    XWindowAttributes attr;

    grabPixelData = 0;
    grabImage = 0;
    grabInitDone = 0;
    
    // Get reference to the window
    grabWindow = (Window) w_id;
        
    // Get information on the window.
    if (XGetWindowAttributes(x11_dpy, grabWindow, &attr) != True)
    {
        fprintf(stderr, "grabInit(): Unable to get window attribues.");
        goto backout_base;
    }

    grabNx = attr.width;
    grabNy = attr.height;

    if ((depth = DefaultDepth(x11_dpy, x11_scr)) != 24)
    {
        fprintf(stderr, "grabInit(): Unsupported depth (%d != 24).\n", depth);
        goto backout_base;
    }

    size = grabNx * grabNy * 4;
    if ((grabPixelData = (unsigned char *) malloc(size*sizeof(char))) == 0)
    {
        fprintf(stderr, "grabInit(): Unable to allocate image.");
        goto backout_base;
    }
    memset(grabPixelData, 0, size*sizeof(char));

    // Create image for grabbing window
    grabImage = XCreateImage(
                             x11_dpy,
                             x11_vis,
                             24,
                             ZPixmap,
                             0,  /* offset. */
                             (char *) grabPixelData,
                             grabNx,
                             grabNy,
                             8,   /* bitmap pad. */
                             0);  /* bytes per line */
    grabInitDone = 1;
    return(0);
        
 backout_base:
    if (grabPixelData) free(grabPixelData);
    //if (grabImage) XDestroyImage(grabImage);
    grabPixelData = 0;
    grabImage = 0;
    
    return(-1);
}

/*----------------------------------------------------------------------------
 * grabUninit()
 *--------------------------------------------------------------------------*/
void grabUninit()
{
    if (grabInitDone)
    {
        if (grabImage) XDestroyImage(grabImage);
        //if (grabPixelData) free(grabPixelData);
        grabPixelData = 0;
        grabImage = 0;
        grabInitDone = 0;
    }
}

/*============================================================================
 * grabImageX()
 *      Grabs the image from the window.
 * ============================================================================*/
unsigned char *grabImageX()
{
    int ii;
    unsigned char R, B;
    
    // Make sure the window size has not changed.
    if (XGetSubImage(x11_dpy, grabWindow, 0, 0, grabNx, grabNy, 0xffffffff, ZPixmap, grabImage, 0, 0) == 0)
    {
        fprintf(stderr, "grabImage(): unable to get image");
        return(0);
    }

    // Go through and switch R<->B.
    for (ii=0; ii < 4*grabNx*grabNy; ii+=4)
    {
        R = grabPixelData[ii];
        B = grabPixelData[ii + 2];
        grabPixelData[ii] = B;
        grabPixelData[ii+2] = R;
    }
    
    return((unsigned char *)grabPixelData);
}

#define WIN_WIDTH 640
#define WIN_HEIGHT 480

/*----------------------------------------------------------------------------
 * openVrRenderUninit()
 *      Shuts down the openVR rendering module.
 *--------------------------------------------------------------------------*/
void openVrRenderUninit()
{
    if (pVr)
    {
        // Ensure that eye rendering is done.  
        VRCompositor()->WaitGetPoses(trackedDevicePose, 1, NULL, 0);
    
        VR_Shutdown();
        pVr = 0;
    }
    else
    {
        fprintf(stderr, "pVr = 0!\n");
    }

    fbFree(&fb_cap_lr);
    fbFree(&fb_l);
    fbFree(&fb_r);

    if (x11_grab_wid)
    {
        fbFree(&fb_grab);
        grabUninit();
    }
    
    if (opengl_context)
    {
        free_openGL_context_X();
        //free_openGL_context_SDL();
    }

    if (cal_mode)
    {
        fclose(f_cal);
    }
}

void updateOrigin(HmdMatrix34_t &matrix);

/*----------------------------------------------------------------------------
 * openVrRenderInit
 *      Initializes output rendering using openvr.
 * Inputs:
 *      output_width, output_height     The size of the output frame that
 *              has been allocated.  If it doesn't match the recommended size
 *              returned by openVR, it will cause an error, requiring
 *              recompiling.  
 *      frameLeft, frameRight
 *              Pointers to the RGBA8 buffers that will be used for raw data
 *              output to the device.
 *----------------------------------------------------------------------------*/
int openVrRenderInit()
{
    HmdError peError;
    int i, j, idx;
    GLenum glError;
    int ccode;

    uint32_t render_w, render_h;
    
    unsigned char *texture_data_cap_lr=0;
    unsigned char *texture_data_l=0, *texture_data_r=0;

    if (get_openGL_context_X() != 0)
    //if (get_openGL_context_SDL() != 0)
    {
        ccode = -1;
        goto backout_base;
    }
    opengl_context = 1;

    // Initialize buffer/image for grabbing scenery from X11 window
    if (x11_grab_wid)
    {
        if (grabInit(x11_grab_wid) != 0)
        {
            fprintf(stderr, "X11 grab initialization failed.\n");
            goto backout_base;
        }

        // Create a texture used to copy grabbed images to openGl
        if (fbNew(&fb_grab, grabNx, grabNy, 0, 0, 1) != 0)
        {
            fprintf(stderr, "Unable to allocate capture framebuffer.\n");
            ccode = -1;
            goto backout_base;
        }
    }
    
    // Initialize VR
    if ((pVr = VR_Init(&peError, VRApplication_Scene)) == 0)
    {
        fprintf(stderr, "VR_Init() failed: %s.\n",
                VR_GetVRInitErrorAsEnglishDescription(peError));
        ccode = -1;
        goto backout_base;

    }

    // Get recommended size of output textures
    pVr->GetRecommendedRenderTargetSize(&render_w, &render_h);
    
    // Get a compsitor instance
    if (VRCompositor() == 0)
    {
        fprintf(stderr, "No compositor available.\n");
        ccode = -3;
        goto backout_base;
    }

    // Allocate textures
    // Capture texture
    if (fbNew(&fb_cap_lr, s_cap_w, s_cap_h, 0, 0, 0) != 0)
    {
        fprintf(stderr, "Unable to allocate capture framebuffer.\n");
        ccode = -1;
        goto backout_base;
    }

    // Output render texture for left eye
    if (fbNew(&fb_l, render_w, render_h, 0x00000000, 1, 0) != 0)
    {
        fprintf(stderr, "Unable to allocate left output framebuffer.\n");
        ccode = -1;
        goto backout_base;
    }
    
    // Output render texture for right eye
    if (fbNew(&fb_r, render_w, render_h, 0x00000000, 1, 0) != 0)
    {
        fprintf(stderr, "Unable to allocate right output framebuffer.\n");
        ccode = -1;
        goto backout_base;
    }

    // Scenery render texture for left eye
    if (fbNew(&fb_s_l, render_w, render_h, 0x00000000, 1, 0) != 0)
    {
        fprintf(stderr, "Unable to allocate left scenery framebuffer.\n");
        ccode = -1;
        goto backout_base;
    }
    
    // Scenery render texture for right eye
    if (fbNew(&fb_s_r, render_w, render_h, 0x00000000, 1, 0) != 0)
    {
        fprintf(stderr, "Unable to allocate right scenery framebuffer.\n");
        ccode = -1;
        goto backout_base;
    }

    // Open calibration output file if in calibration mode
    if (cal_mode)
    {
        if ((f_cal = fopen("cal.out", "wb")) == 0)
        {
            fprintf(stderr, "Unable to open calibration file.\n");
            ccode = -1;
            goto backout_base;
        }
    }

    // Get eye pose and projection matrices
    M_eye_pose_l = getEyePoseMatrix(Eye_Left);
    M_eye_pose_r = getEyePoseMatrix(Eye_Right);
    M_proj_l = getEyeProjMatrix(Eye_Left);
    M_proj_r = getEyeProjMatrix(Eye_Right);
    
    // Send a couple of dummy frames to make sure things are really ready.
    openVrRenderFrame(0, 0);
    //VRCompositor()->WaitGetPoses(trackedDevicePose, 1, NULL, 0);
    memset(trackedDevicePose, 0, sizeof(trackedDevicePose));
    //memset(trackedDevicePose_prev, 0, sizeof(trackedDevicePose_prev));
    updateOrigin(trackedDevicePose[0].mDeviceToAbsoluteTracking);
    
    PRINTF("OpenVR Ready.\n");
    
    // Success
    return(0);
    
 backout_base:
    openVrRenderUninit();
    return(ccode);
}

/*----------------------------------------------------------------------------
 * fbCopy2
 *   Renders one framebuffer onto another one.  Function supports
 *   cropping and scaling depending on the source and target sizes.
 *   Coordinates are logical coordinates (0,0) lower left to (1,1) upper right.
 * Inputs:
 *      fb_src, fb_dest         Source and destination framebuffers
 *      src_x, src_y            Center of source frame to copy
 *      src_w, src_h            Width and height of source frame
 *      dst_x, dst_y            Center of target frame
 *      dst_w, dst_h            Width and height of target frame
 *--------------------------------------------------------------------------*/
int fbCopy2(fbDef *fb_src, fbDef *fb_dest,
            float src_x, float src_y, float src_w, float src_h,
            float dst_x, float dst_y, float dst_w, float dst_h,
            int do_clear, int do_flip)
{
    float src_x1, src_x2, src_y1, src_y2;
    float dst_x1, dst_x2, dst_y1, dst_y2;
    float tmp;
    
    src_x1 = src_x - 0.5*src_w;
    src_x2 = src_x + 0.5*src_w;
    if (do_flip)
    {
        src_y1 = src_y + 0.5*src_h;
        src_y2 = src_y - 0.5*src_h;
    }
    else
    {
        src_y1 = src_y - 0.5*src_h;
        src_y2 = src_y + 0.5*src_h;
    }
    dst_x1 = dst_x - 0.5*dst_w;
    dst_x2 = dst_x + 0.5*dst_w;
    dst_y1 = dst_y - 0.5*dst_h;
    dst_y2 = dst_y + 0.5*dst_h;
    
    glBindFramebuffer(GL_FRAMEBUFFER, fb_dest->fbo_id);
    glViewport(0, 0, fb_dest->width, fb_dest->height);

    glPushMatrix();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0,   // left
            1.0,   // right
            0.0,   // bottom
            1.0,   // top
            1.0,  // near
            -1.0);  // far

    if (do_clear)
    {
        glClearColor(0,0,0,1);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    }
    
    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glBindTexture(GL_TEXTURE_2D, fb_src->tex_id);

    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glBegin(GL_QUADS);
    // Lower left
    glTexCoord2f(src_x1, src_y1);
    glVertex2d(dst_x1, dst_y1);
    // Lower Right
    glTexCoord2f(src_x2, src_y1);
    glVertex2d(dst_x2, dst_y1);
    // Upper Right
    glTexCoord2f(src_x2, src_y2);
    glVertex2d(dst_x2, dst_y2);
    // Upper Left
    glTexCoord2f(src_x1, src_y2);
    glVertex2d(dst_x1, dst_y2);
    glEnd();

    glPopMatrix();

    glDisable(GL_TEXTURE_2D);
}

/*----------------------------------------------------------------------------
 * fbCopy3
 *   Renders one framebuffer onto another one.  Function supports
 *   cropping and scaling depending on the source and target sizes.
 *   Coordinates are logical coordinates (0,0) lower left to (1,1) upper right.
 * Inputs:
 *      fb_src, fb_dest         Source and destination framebuffers
 *      src_x, src_y            Center of source frame to copy
 *      src_w, src_h            Width and height of source frame
 *      dst_x1, dst_y1          Lower left target
 *      dst_x2, dst_y2          Lower right
 *      dst_x3, dst_y3          Upper right
 *      dst_x4, dst_y4          Upper left
 *--------------------------------------------------------------------------*/
int fbCopy3(fbDef *fb_src, fbDef *fb_dest,
            float src_x, float src_y, float src_w, float src_h,
            float dst_x1, float dst_y1,
            float dst_x2, float dst_y2,
            float dst_x3, float dst_y3,
            float dst_x4, float dst_y4,
            int do_clear, int do_flip)
{
    float src_x1, src_x2, src_y1, src_y2;
    float tmp;
    
    src_x1 = src_x - 0.5*src_w;
    src_x2 = src_x + 0.5*src_w;
    if (do_flip)
    {
        src_y1 = src_y + 0.5*src_h;
        src_y2 = src_y - 0.5*src_h;
    }
    else
    {
        src_y1 = src_y - 0.5*src_h;
        src_y2 = src_y + 0.5*src_h;
    }
    
    glBindFramebuffer(GL_FRAMEBUFFER, fb_dest->fbo_id);
    glViewport(0, 0, fb_dest->width, fb_dest->height);

    glPushMatrix();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-1.0,   // left
            1.0,   // right
            -1.0,   // bottom
            1.0,   // top
            1.0,  // near
            -1.0);  // far

    if (do_clear)
    {
        glClearColor(0,0,0,1);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    }

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   
    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glBindTexture(GL_TEXTURE_2D, fb_src->tex_id);

    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glBegin(GL_QUADS);
    // Lower left
    glTexCoord2f(src_x1, src_y1);
    glVertex2d(dst_x1, dst_y1);
    // Lower Right
    glTexCoord2f(src_x2, src_y1);
    glVertex2d(dst_x2, dst_y2);
    // Upper Right
    glTexCoord2f(src_x2, src_y2);
    glVertex2d(dst_x3, dst_y3);
    // Upper Left
    glTexCoord2f(src_x1, src_y2);
    glVertex2d(dst_x4, dst_y4);
    glEnd();

    glPopMatrix();

    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
}




void get_vectors_from_orient_matrix(HmdMatrix34_t &matrix, Vector3 &x, Vector3 &y, Vector3 &z, Vector3 &o)
{
    x.set(matrix.m[0][0], matrix.m[1][0], matrix.m[2][0]);
    y.set(matrix.m[0][1], matrix.m[1][1], matrix.m[2][1]);
    z.set(matrix.m[0][2], matrix.m[1][2], matrix.m[2][2]);
    o.set(matrix.m[0][3], matrix.m[1][3], matrix.m[2][3]);
}

// Sends HMD info to a running flightgear process.
void sendHmdInfo()
{
    Vector3 head_x, head_y, head_z, head_pos;
    Vector3 fg_hx, fg_hy, fg_hz;
    Vector3 fg_px, fg_py, fg_pz;

    Vector3 d_pos;
    float x, y, z, fg_x, fg_y, fg_z;
    Vector3 xh, yh, zh, xh1, yh1, zh1;
    Vector3 yh_, zh_;
    float Y, P, R;
    
    // Extract individual vectors and origin
    get_vectors_from_orient_matrix(trackedDevicePose[0].mDeviceToAbsoluteTracking, head_x, head_y, head_z, head_pos);
    
    d_pos = head_pos - pilot_origin;
    x = pilot_x.dot(d_pos);
    y = pilot_y.dot(d_pos);
    z = pilot_z.dot(d_pos);

    // Adjust so "zero" is right where pilot head should be
    x -= 0.25;
    y += 0.2;

    // Get local head xh, yh, zh vectors in pilot coordinate frame
    xh[0] = head_x.dot(pilot_x); xh[1] = head_x.dot(pilot_y); xh[2] = head_x.dot(pilot_z);
    yh[0] = head_y.dot(pilot_x); yh[1] = head_y.dot(pilot_y); yh[2] = head_y.dot(pilot_z);
    zh[0] = head_z.dot(pilot_x); zh[1] = head_z.dot(pilot_y); zh[2] = head_z.dot(pilot_z);

    // Convert vectors to aviation coordinates
    xh1[0] = zh[2];  xh1[1] = zh[0];  xh1[2] = -zh[1];
    yh1[0] = xh[2];  yh1[1] = xh[0];  yh1[2] = -xh[1];
    zh1[0] = -yh[2]; zh1[1] = -yh[0]; zh1[2] = yh[1];
    
    // Convert orientation vectors to YPR
    Y = atan2(xh1[1], xh1[0]);
    P = atan2(xh1[2], sqrtf(SQ(xh1[0]) + SQ(xh1[1])));

    // Get "zero roll" y and z vectors
    yh_[0] = -xh1[1];
    yh_[1] = xh1[0];
    yh_[2] = 0.0;
    zh_ = xh1.cross(yh_);
    zh_.normalize();

    // Get roll
    R = atan2(yh1.dot(zh_), yh1.dot(yh_));
    
    // Send to flightgear
    udpSendHmdData(x, y, z, Y*180.0/M_PI, P*180.0/M_PI, R*180.0/M_PI);
}

void updateOrigin(HmdMatrix34_t &matrix)
{
    Vector3 point_map;
    int i, j;
    
    // Copy the headset position
    fprintf(stdout, "Updating cockpit origin:\n");
    printMatrix(matrix);

    // Extract individual vectors and origin
    get_vectors_from_orient_matrix(matrix, pilot_x, pilot_y, pilot_z, pilot_origin);

    // Update the cockpit window location
    for (j=0; j<N_poly_win_tris; j++)
    {
        for (i=0; i<N_vert_win_tris; i++)
        {
            point_map = pilot_origin
                + pilot_x*win_tris_vertex[j][i][0]
                + pilot_y*win_tris_vertex[j][i][1]
                + pilot_z*win_tris_vertex[j][i][2];
            win_tris_vertex_map[j][i][0] = point_map[0];
            win_tris_vertex_map[j][i][1] = point_map[1];
            win_tris_vertex_map[j][i][2] = point_map[2];
        }
    }
}

GLfloat light_diffuse[] = {1.0, 1.0, 0.5, 1.0};  /* Red diffuse light. */
GLfloat light_position[] = {0.0, 0.0, 20.0, 0.0};  /* Infinite light location. */

/*============================================================================
 * setGlView2
 *      Sets the OpenGL model and projection view according to parameters
 *      that we typically will have from VR: user head position and orientation.
 * Inputs:
 *      cam_{x,y,z}     Unit vectors for camera coordinate system
 *      cam_origin      Origin of the camera
 *      z_near,z_far    Range of z coordinates that should be displayed
 *==========================================================================*/
void setGlView2(Vector3 &cam_x,
                Vector3 &cam_y,
                Vector3 &cam_z,
                Vector3 &cam_origin,
                GLfloat z_near,
                GLfloat z_far,
                int left_eye)
{
    static GLfloat m[16];
    GLfloat trans_x, trans_y, trans_z;
    GLfloat r, t;
    GLfloat adj_x, adj_y;
    Hmd_Eye eye;
    
    trans_x = -cam_x.dot(cam_origin);
    trans_y = -cam_y.dot(cam_origin);
    trans_z = -cam_z.dot(cam_origin);

    glMatrixMode(GL_MODELVIEW);

    // Get rotation and translation matrix for model according to position
    // and direction of camera (eye).
    m[0] = cam_x[0]; m[4] = cam_x[1]; m[8] = cam_x[2];   m[12] = trans_x;
    m[1] = cam_y[0]; m[5] = cam_y[1]; m[9] = cam_y[2];   m[13] = trans_y;
    m[2] = cam_z[0]; m[6] = cam_z[1]; m[10] = cam_z[2];  m[14] = trans_z;
    m[3] = 0.0;      m[7] = 0.0;      m[11] = 0.0;       m[15] = 1.0;
    glLoadMatrixf(m);
    
    if (left_eye)
    {
        eye = Eye_Left;
    }
    else
    {
        eye = Eye_Right;
    }
    
    HmdMatrix44_t mat = pVr->GetProjectionMatrix(eye, z_near, z_far);
    glMatrixMode(GL_PROJECTION);
    m[0] = mat.m[0][0]; m[4] = mat.m[0][1]; m[8] = mat.m[0][2]; m[12] = mat.m[0][3];
    m[1] = mat.m[1][0]; m[5] = mat.m[1][1]; m[9] = mat.m[1][2]; m[13] = mat.m[1][3];
    m[2] = mat.m[2][0]; m[6] = mat.m[2][1]; m[10] = mat.m[2][2]; m[14] = mat.m[2][3];
    m[3] = mat.m[3][0]; m[7] = mat.m[3][1]; m[11] = mat.m[3][2]; m[15] = mat.m[3][3];
    glLoadMatrixf(m);
}


/*----------------------------------------------------------------------------
 * renderScenery
 *      Renders scenery into the L and R scenery framebuffers.
 *--------------------------------------------------------------------------*/
void renderScenery(fbDef *fbl, fbDef *fbr)
{
    Vector3 eye_pos_left, eye_pos_right;
    Vector3 head_x, head_y, head_z, head_pos;
    Vector3 head_x2, head_y2, head_z2, head_pos2, tmp;
    //fbDef *fb;

    glPushMatrix();
    
    // Break the current headset position into components
    get_vectors_from_orient_matrix(trackedDevicePose[0].mDeviceToAbsoluteTracking, head_x, head_y, head_z, head_pos);

    // Get head position in pilot frame of reference
    tmp = head_pos - pilot_origin;
    head_pos2.set(tmp.dot(pilot_x), tmp.dot(pilot_y), tmp.dot(pilot_z));

    // Transform into the pilot position frame of reference
    head_x2.set(head_x.dot(pilot_x), head_x.dot(pilot_y), head_x.dot(pilot_z));
    head_y2.set(head_y.dot(pilot_x), head_y.dot(pilot_y), head_y.dot(pilot_z));
    head_z2.set(head_z.dot(pilot_x), head_z.dot(pilot_y), head_z.dot(pilot_z));
    
    light_position[0] = head_pos2[0];
    light_position[1] = head_pos2[1];
    light_position[2] = head_pos2[2];

    /* setup lighting. */
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    glLightfv(GL_LIGHT0, GL_POSITION, light_position);
    glEnable(GL_LIGHT0);
    glEnable(GL_LIGHTING);
    // DEBUG: Solid color.
    //glColor3f(1, 0.1, 0.1);
    
    /* Use depth buffering for hidden surface elimination. */
    glEnable(GL_DEPTH_TEST);
    
    // We assume the scenery coordinates are relative to the pilot's head origin, rather than absolute.
    eye_pos_left = eye_pos_right = head_pos2;
    
    // Render left eye
    
    glBindFramebuffer(GL_FRAMEBUFFER, fbl->fbo_id);
    glViewport(0, 0, fbl->width, fbl->height);

    setGlView2(head_x2, head_y2, head_z2, eye_pos_left, pilot_fov_z_near, pilot_fov_z_far, 1);
    
    glClearColor(0,0,0,1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    render_cubes2();

    // Render right eye
    
    glBindFramebuffer(GL_FRAMEBUFFER, fbr->fbo_id);
    glViewport(0, 0, fbr->width, fbr->height);

    setGlView2(head_x2, head_y2, head_z2, eye_pos_right, pilot_fov_z_near, pilot_fov_z_far, 0);
    
    glClearColor(0,0,0,1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    render_cubes2();

    glBindFramebuffer(GL_FRAMEBUFFER, 0);

    glPopMatrix();

    glDisable(GL_LIGHT0);
    glDisable(GL_LIGHTING);
}

/*----------------------------------------------------------------------------
 * renderSceneryGrab
 *      Grabs video from an X11 window and renders that to the scenery
 *      textures.
 *--------------------------------------------------------------------------*/
void renderSceneryGrab(fbDef *fbl, fbDef *fbr)
{
    grabImageX();
    fbWrite(&fb_grab, grabPixelData);

    // Copy grabbed image to the two scenery framebuffers
    fbCopy2(&fb_grab, fbl,
            0.5, 0.5, 1.0, 1.0,
            scen_lx, scen_ly, scen_lw, scen_lh,
            1,
            1);
    
    fbCopy2(&fb_grab, fbr,
            0.5, 0.5, 1.0, 1.0,
            scen_rx, scen_ry, scen_rw, scen_rh,
            1,
            1);
}

/*----------------------------------------------------------------------------
 * drawSceneryWindow2
 *      Draws scenery onto a virtual window in the cockpit.  Expects that
 *      renderScenery has been called already.  Uses HMD projection matrices
 *      instead of my own approximate method.
 * Inputs:
 *      fb_dest                 Destination framebuffer.
 *      fb_src                  Texture to use as source for window
 *      cent_x, cent_y          Center of the framebuffer to write to.
 *      eye_offset_x            Distance to move (L/R) for eye.
 *--------------------------------------------------------------------------*/
void drawSceneryWindow2(fbDef *fb_dest, fbDef *fb_src, int left_eye)
{
    float window_mapped_corners[3][2];
    float *g;
    int i, j;
    Vector4 gp, pt;
    int is_visible;

    Matrix4 M;

    glBindFramebuffer(GL_FRAMEBUFFER, fb_dest->fbo_id);
    glViewport(0, 0, fb_dest->width, fb_dest->height);
    
    // Draw on the render texture
    glPushMatrix();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0,   // left
            1.0,   // right
            0.0,   // bottom
            1.0,   // top
            1.0,  // near
            -1.0);  // far

    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glBindTexture(GL_TEXTURE_2D, fb_src->tex_id);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
                
    // Try everything in global coordinates
    if (left_eye)
    {
        M = M_proj_l*M_eye_pose_l*M_pose;
    }
    else
    {
        M = M_proj_r*M_eye_pose_r*M_pose;
    }
    
    // Map points
    glBegin(GL_TRIANGLES);
    for (j=0; j<N_poly_win_tris; j++)
    {
        is_visible = 1;
        for (i=0; i<3; i++)
        {
            // Get depth to point
            g = win_tris_vertex_map[j][i];
            // Make into a 4-element vector
            gp.set(g[0], g[1], g[2], 1.0);
            // Do projections
            pt = M*gp;
            window_mapped_corners[i][0] = (pt[0]/pt[3] + 1.0)/2.0;
            window_mapped_corners[i][1] = (pt[1]/pt[3] + 1.0)/2.0;

            // Clipping should be handled automatically if corners are out of range, but if the
            // z coordinate goes behind viewer, there are problems.
            if (pt[2]/pt[3] > 1.0)
            {
                is_visible = 0;
            }
        }

        // Draw the image
        if (is_visible)
        {
            if (fb_src)
            {
                for (i=0; i<3; i++)
                {
                    glTexCoord2f(window_mapped_corners[i][0], window_mapped_corners[i][1]);
                    glVertex2d(window_mapped_corners[i][0], window_mapped_corners[i][1]);
                }
            }
        }
    }
    glEnd();
    
    // Draw an triangle outlines
    glDisable(GL_TEXTURE_2D);
    glColor3f(1, 0.1, 0.1);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

    // Map points
    glBegin(GL_TRIANGLES);
    for (j=0; j<N_poly_win_tris; j++)
    {
        is_visible = 1;
        for (i=0; i<3; i++)
        {
            // Get depth to point
            g = win_tris_vertex_map[j][i];
            // Make into a 4-element vector
            gp.set(g[0], g[1], g[2], 1.0);
            // Do projections
            pt = M*gp;
            window_mapped_corners[i][0] = (pt[0]/pt[3] + 1.0)/2.0;
            window_mapped_corners[i][1] = (pt[1]/pt[3] + 1.0)/2.0;

            // Clipping should be handled automatically if corners are out of range, but if the
            // z coordinate goes behind viewer, there are problems.
            if (pt[2]/pt[3] > 1.0)
            {
                is_visible = 0;
            }
        }

        // Draw the image
        if (is_visible)
        {
            if (fb_src)
            {
                for (i=0; i<3; i++)
                {
                    glTexCoord2f(window_mapped_corners[i][0], window_mapped_corners[i][1]);
                    glVertex2d(window_mapped_corners[i][0], window_mapped_corners[i][1]);
                }
            }
        }
    }
    glEnd();
    
    glPopMatrix();

    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    if (fb_src)
    {
        glDisable(GL_TEXTURE_2D);
    }
}

void handleKey(int key)
{
    static int on_left = 0;
    static int on_width = 0;
    
    switch(key)
    {
        case 'a':
            updateOrigin(trackedDevicePose[0].mDeviceToAbsoluteTracking);
            break;

        case 'l':
            fprintf(stdout, "Left\n");
            on_left = 1;
            break;

        case 'r':
            fprintf(stdout, "Right\n");
            on_left = 0;
            break;

        case 'w':
            fprintf(stdout, "Width\n");
            on_width = 1;
            break;

        case 'h':
            fprintf(stdout, "Height\n");
            on_width = 0;
            break;

        case '4':
            if (on_left)
            {
                scen_lx -= 0.01;
            }
            else
            {
                scen_rx -= 0.01;
            }
            break;

        case '6':
            if (on_left)
            {
                scen_lx += 0.01;
            }
            else
            {
                scen_rx += 0.01;
            }
            break;
            
        case '2':
            if (on_left)
            {
                scen_ly -= 0.01;
            }
            else
            {
                scen_ry -= 0.01;
            }
            break;

        case '8':
            if (on_left)
            {
                scen_ly += 0.01;
            }
            else
            {
                scen_ry += 0.01;
            }
            break;
            
        case '+':
            if (on_width)
            {
                scen_lw += 0.01;
                scen_rw += 0.01;
            }
            else
            {
                scen_lh += 0.01;
                scen_rh += 0.01;
            }
            break;
            
        case '-':
            if (on_width)
            {
                scen_lw -= 0.01;
                scen_rw -= 0.01;                
            }
            else
            {
                scen_lh -= 0.01;
                scen_rh -= 0.01;
            }
            break;
            
        case 's':
            // Toggle scenery only mode
            render_scenery_only = !render_scenery_only;
            break;
            
        case 'c':
            if (cal_mode == 0)
            {
                // Turn calibration mode on.
                if ((f_cal = fopen("cal.out", "wb")) != 0)
                {
                    cal_mode = 1;
                }
                fprintf(stderr, "Calibration mode on.\n");
            }
            else
            {
                fclose(f_cal);
                cal_mode = 0;
                fprintf(stderr, "Calibration mode off.\n");
            }
            break;

        case 'p':
            fprintf(stdout, "scen_lx=%f\r\n", scen_lx);
            fprintf(stdout, "scen_ly=%f\r\n", scen_ly);
            fprintf(stdout, "scen_lw=%f\r\n", scen_lw);
            fprintf(stdout, "scen_lh=%f\r\n", scen_lh);
            fprintf(stdout, "scen_rx=%f\r\n", scen_rx);
            fprintf(stdout, "scen_ry=%f\r\n", scen_ry);
            fprintf(stdout, "scen_rw=%f\r\n", scen_rw);
            fprintf(stdout, "scen_rh=%f\r\n", scen_rh);
            break;

        case 'q':
            quitting = 1;
            break;
    }
}

/*----------------------------------------------------------------------------
 * findDot
 *      finds the brightest green dot in the camera frame.
 *--------------------------------------------------------------------------*/
int findDot(arOutputPixelType *frame_data, int *x, int *y, float *amp, float req_amp, int width, int height)
{
    int i, j;
    float x2, avg, val1;
    float std;

    int max_x, max_y;
    float max_amp;
    
    x2 = 0.0;
    avg = 0.0;
    max_x = -1;
    max_y = -1;
    max_amp = 0.0;
    
    for (j=0; j<height; j++)
    {
        for (i=0; i<width; i++)
        {
            val1 = (float)((frame_data[i + width*j] & 0x0000ff00) >> 8)/255.0;
            x2 += val1*val1;
            avg += val1;

            if (val1 > max_amp)
            {
                max_amp = val1;
                max_x = i;
                max_y = j;
            }
        }
    }
    x2 /= (float)(height*width);
    avg /= (float)(height*width);

    std = sqrtf(x2 - avg*avg);

    if ((max_amp - avg)/std > req_amp)
    {
        *x = max_x;
        *y = max_y;
        *amp = (max_amp - avg)/std;
        return(1);
    }
    else
    {
        *x = 0;
        *y = 0;
        *amp = 0;
        return(0);
    }
}

/*----------------------------------------------------------------------------
 * cameraCal
 *      Outputs camera calibration data to a file.  Detects a green      
 *      LED on a controller and sends both its detected position
 *      and VR position to a file.
 *--------------------------------------------------------------------------*/
void cameraCal(arOutputPixelType *cap_data)
{
    Vector3 head_x, head_y, head_z, head_pos, eye_pos;
    Vector3 cont_x, cont_y, cont_z, cont_pos;
    Vector3 cont_xp, cont_yp, cont_zp, cont_posp;
    Vector3 led_pos;
    
    Vector3 g;
    Vector3 dv;
    float z, p1, p2;
    int i;
    Vector4 gp, pt;
    int is_visible;

    /*
    float m_mod[16];
    Matrix4 M_mod;
    float m_proj[16];
    Matrix4 M_proj;
    */
    Matrix4 M;
    
    int left_dot_x, left_dot_y;
    int right_dot_x, right_dot_y;
    float left_dot_amp, right_dot_amp;
    
    int found_controller;

    int ii;
    
    // Detect the LED in the left and right frames.
    // The brightest point needs to be a certain number of
    // standard deviations above the average.
    if (findDot(cap_data, &left_dot_x, &left_dot_y, &left_dot_amp, 1.5, s_cap_w, s_cap_h/2) == 0)
    {
        return;
    }

    if (findDot(cap_data + s_cap_w*s_cap_h/2, &right_dot_x, &right_dot_y, &right_dot_amp, 1.5, s_cap_w, s_cap_h/2) == 0)
    {
        return;
    }

    // Get the position of the LED from the controller
    found_controller = 0;
    for (ii=0; ii<k_unMaxTrackedDeviceCount; ii++)
    {
        if ( trackedDevicePose[ii].bPoseIsValid )
        {
            if (pVr->GetTrackedDeviceClass(ii) == TrackedDeviceClass_Controller)
            {
                found_controller=1;
                break;
            }
        }
    }
    if (!found_controller) return;

    // Get components of controller position
    get_vectors_from_orient_matrix(trackedDevicePose[ii].mDeviceToAbsoluteTracking, cont_x, cont_y, cont_z, cont_pos);

    // Get LED position.  These hard-coded values were obtained by finding the offset in an HTC
    // provided .OBJ of the controller.  
    led_pos = cont_pos +0.0963*cont_z + 0.0029*cont_y;

    // Make into a 4-element vector
    gp.set(led_pos[0], led_pos[1], led_pos[2], 1.0);
    
    // Left Eye
    M = M_proj_l*M_eye_pose_l*M_pose;
    pt = M*gp;

    fprintf(f_cal, "LV %.6f %.6f\n", pt[0]/pt[3], pt[1]/pt[3]);
    fprintf(f_cal, "LC %d %d %.2f\n", left_dot_x, left_dot_y, left_dot_amp);

    // Right Eye
    M = M_proj_r*M_eye_pose_r*M_pose;
    pt = M*gp;

    fprintf(f_cal, "RV %.6f %.6f\n", pt[0]/pt[3], pt[1]/pt[3]);
    fprintf(f_cal, "RC %d %d %.2f\n", right_dot_x, right_dot_y, left_dot_amp);
}

/*----------------------------------------------------------------------------
 * openVrRenderFrame
 *      Renders a frame to the device.  Assumes the buffers that were passed
 *      in at initialization have been updated with the frames.
 *--------------------------------------------------------------------------*/ 
void openVrRenderFrame(arOutputProfDef *p, arOutputPixelType *cap_data)
{
    static unsigned char frame_idx = 0;
    int i, j, idx;
    EVRCompositorError err1;
    int ch;
    
    if (p)
    {
        profCpuStart(&(p->render[0]));
        profWallStart(&(p->wrender[0]));
    }

    VRCompositor()->WaitGetPoses(trackedDevicePose, k_unMaxTrackedDeviceCount, NULL, 0);

    // Update the pose matrix used for rendering
    if (trackedDevicePose[k_unTrackedDeviceIndex_Hmd].bPoseIsValid)
    {
        M_pose = ConvertSteamVRMatrixToMatrix4(trackedDevicePose[k_unTrackedDeviceIndex_Hmd].mDeviceToAbsoluteTracking);
        M_pose.invert();
    }
    
    if (cal_mode)
    {
        cameraCal(cap_data);
    }
   
    if (p) 
    {
        profCpuEnd(&(p->render[0]), 0);
        profWallEnd(&(p->wrender[0]));
    }

    if (p)
    {
        profCpuStart(&(p->render[1]));
        profWallStart(&(p->wrender[1]));
    }
        
    // New frame?
    if (cap_data)
    {
        // Copy capture data down to GPU
        fbWrite(&fb_cap_lr, cap_data);
    }

    if(p)
    {
        profCpuEnd(&(p->render[1]), 0);
        profWallEnd(&(p->wrender[1]));
    }

    if (p)
    {
        profWallStart(&(p->wrender[2]));
        profCpuStart(&(p->render[2]));
    }

    if (!render_scenery_only)
    {
        // Green screen detection?
        if (do_green)
        {
            // Make scenery the background
            if (x11_grab_wid)
            {
                // Render scenery from an X11 window.
                renderSceneryGrab(&fb_l, &fb_r);
            }
            else
            {
                // Render debug scenery
                renderScenery(&fb_l, &fb_r);
            }

            // Now put camera image on-top, where green
            // areas are see-through.
            fbCopy3(&fb_cap_lr, &fb_l,
                    s_cap_crop_lx, s_cap_crop_ly, s_cap_crop_w, s_cap_crop_h,
                    s_rend_vid_lx1, s_rend_vid_ly1,
                    s_rend_vid_lx2, s_rend_vid_ly2,
                    s_rend_vid_lx3, s_rend_vid_ly3,
                    s_rend_vid_lx4, s_rend_vid_ly4,
                    0, 0);
            
            fbCopy3(&fb_cap_lr, &fb_r,
                    s_cap_crop_rx, s_cap_crop_ry, s_cap_crop_w, s_cap_crop_h,
                    s_rend_vid_rx1, s_rend_vid_ry1,
                    s_rend_vid_rx2, s_rend_vid_ry2,
                    s_rend_vid_rx3, s_rend_vid_ry3,
                    s_rend_vid_rx4, s_rend_vid_ry4,
                    0, 0);
        }
        else
        {
            // No green screen.  Use "known" cockpit window position to
            // show scenery.
            
            // Extract left and right subimages from stereo camera.  Draw on the output
            // L and R frames.
            fbCopy3(&fb_cap_lr, &fb_l,
                    s_cap_crop_lx, s_cap_crop_ly, s_cap_crop_w, s_cap_crop_h,
                    s_rend_vid_lx1, s_rend_vid_ly1,
                    s_rend_vid_lx2, s_rend_vid_ly2,
                    s_rend_vid_lx3, s_rend_vid_ly3,
                    s_rend_vid_lx4, s_rend_vid_ly4,
                    1, 0);
            
            fbCopy3(&fb_cap_lr, &fb_r,
                    s_cap_crop_rx, s_cap_crop_ry, s_cap_crop_w, s_cap_crop_h,
                    s_rend_vid_rx1, s_rend_vid_ry1,
                    s_rend_vid_rx2, s_rend_vid_ry2,
                    s_rend_vid_rx3, s_rend_vid_ry3,
                    s_rend_vid_rx4, s_rend_vid_ry4,
                    1, 0);


            if (x11_grab_wid)
            {
                // Render scenery from an X11 window.
                renderSceneryGrab(&fb_s_l, &fb_s_r);
            }
            else
            {
                // Render a debug image
                renderScenery(&fb_s_l, &fb_s_r);
            }

            // Copy "winow" portion of scene to output frames
            drawSceneryWindow2(&fb_l, &fb_s_l, 1);
            drawSceneryWindow2(&fb_r, &fb_s_r, 0);
        }
        
    }
    else // Scenery only
    {
        if (x11_grab_wid)
        {
            // Render scenery from an X11 window.
            renderSceneryGrab(&fb_l, &fb_r);
        }
        else
        {
            // Only rendering scenery.  Render directly to output frames
            renderScenery(&fb_l, &fb_r);
        }
        // DEBUG: Draw window on top to see mapping same.
        //drawSceneryWindow2(&fb_l, 0, 1);
        //drawSceneryWindow2(&fb_r, 0, 0);
    }

    if (p)
    {
        profCpuEnd(&(p->render[2]), 0);
        profWallEnd(&(p->wrender[2]));
    }
    
    if (p)
    {
        profWallStart(&(p->wrender[3]));
        profCpuStart(&(p->render[3]));
    }
    
    if ((err1 = VRCompositor()->Submit(Eye_Left, &(fb_l.tex))) != 0)
    {
        fprintf(stderr, "OpenVR Error on Submit(Left): %d\n", err1);
    }

    if (p)
    {
        profCpuEnd(&(p->render[3]), 0);
        profWallEnd(&(p->wrender[3]));
    }

    if (p)
    {
        profWallStart(&(p->wrender[4]));
        profCpuStart(&(p->render[4]));
    }
    
    if ((err1 = VRCompositor()->Submit(Eye_Right, &(fb_r.tex))) != 0)
    {
        fprintf(stderr, "OpenVR Error on Submit(Right): %d\n", err1);
    }

    if (p)
    {
        profCpuEnd(&(p->render[4]), 0);
        profWallEnd(&(p->wrender[4]));
    }
    
    // Check for events (such as key presses)
    if (p)
    {
        profWallStart(&(p->wrender[5]));
        profCpuStart(&(p->render[5]));
    }

    if (fg_hmd_port) sendHmdInfo();

    if ((ch = kbdGet()) > 0)
    {
        //printf("%d\n", ch);
        handleKey(ch);
    }
    
    if (p)
    {
        profCpuEnd(&(p->render[5]), 0);
        profWallEnd(&(p->wrender[5]));
    }

}

