////////////////////////////////////////////////////////////////////////////
// process.cc
//      Code to receive camera and X11 capture data from the FIFO, do
//      any CPU processing, and send on to the render module.
//
// Author: Jon W. Wallace, Lafayette College
// Date:   May 2020
////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <stdio.h>
#include <inttypes.h>
#include <opencv2/opencv.hpp>
#include <opencv2/core/cuda.hpp>

#include "log.h"
#include "main.h"
#include "process.h"
#include "capture.h"
#include "fifo.h"
#include "render.h"
#include "grab_x11.h"
#include "yuv.h"
#include "cudaRender.h"
#include "aruco.h"

using namespace cv;

typedef struct
{
    cpuProfDef fifoWait;
    cpuProfDef flipRemapCam;
    cpuProfDef flipRemapX11;
    cpuProfDef profAll;
    wallProfDef profWAll;
    uint64_t numLiveFrames;
    uint64_t numAllFrames;
} profProcessDef;
static profProcessDef p;

// Camera calibration variables

float camera_matrix_L_data[] = {
    6.8610710905838869e+02, 0., 4.7950000000000000e+02, 0.,
    6.8610710905838869e+02, 4.7450000000000000e+02, 0., 0., 1.};
float camera_dist_L_data[] = {
    -3.3907258811025903e-01, 1.7107625656963976e-01, 0., 0.,
    -4.9717228090663701e-02};
float camera_matrix_R_data[] = {
    5.3958736362590241e+02, 0., 4.7950000000000000e+02, 0.,
    5.3958736362590241e+02, 4.7450000000000000e+02, 0., 0., 1.};
float camera_dist_R_data[] = {
    -1.9931858070701500e-01, 5.5016436919929058e-02, 0., 0.,
     -9.5637707193731662e-03};

Mat camera_matrix_R(3, 3, CV_32F, camera_matrix_L_data);
Mat camera_dist_R(5, 1, CV_32F, camera_dist_L_data);
Mat camera_matrix_L(3, 3, CV_32F, camera_matrix_L_data);
Mat camera_dist_L(5, 1, CV_32F, camera_dist_L_data);

Mat undistort_map1_L;
Mat undistort_map2_L;
Mat undistort_map1_R;
Mat undistort_map2_R;

cuda::GpuMat gpu_undistort_map1_L;
cuda::GpuMat gpu_undistort_map2_L;
cuda::GpuMat gpu_undistort_map1_R;
cuda::GpuMat gpu_undistort_map2_R;

uint32_t procWidth;
uint32_t procHeight;
uint32_t procSize;
uint32_t procSizeBytes;

// gpu_mat
//      This structure holds all of the GPU buffers and OpenCV matrices I need
//      for processing.  I allocate these ahead of time to avoid alloc/dealloc in
//      the processing loop.
struct
{
    void *bgr_frame_l_;
    void *bgr_frame_r_;
    void *gray_frame_lr_;
    void *rgba_frame_l_;
    void *rgba_frame_r_;
    void *rgba_frame_l_out_;
    void *rgba_frame_r_out_;
    void *rgba_frame_lr_out_;
    void *undist_frame_l_;
    void *undist_frame_r_;

    cv::cuda::GpuMat cap_frame_l;
    cv::cuda::GpuMat cap_frame_r;
    cv::cuda::GpuMat gray_frame_lr;
    cv::cuda::GpuMat bgr_frame_l;
    cv::cuda::GpuMat bgr_frame_r;
    cv::cuda::GpuMat rgba_frame_l;
    cv::cuda::GpuMat rgba_frame_r;
    cv::cuda::GpuMat rgba_frame_l_out;
    cv::cuda::GpuMat rgba_frame_r_out;
    cv::cuda::GpuMat rgba_frame_lr_out;
    cv::cuda::GpuMat undist_frame_l;
    cv::cuda::GpuMat undist_frame_r;

    Rect r_sub;
    Rect l_sub;
} gpu_mat;

////////////////////////////////////////////////////////////////////////////
// colorMapFlip
//      Convert BGRA (X11 and camera capture) to RGBA and flip images.
//      CPU-based.  Works on data in place.  Assumes height is even.
////////////////////////////////////////////////////////////////////////////
void colorMapFlip(void *data, uint32_t width, uint32_t height)
{
    uint32_t i, j;
    uint32_t p1, p2, p3, p4;

    uint32_t *bgra_data = (uint32_t *) data;
        
    // Flip and fix color mapping
    for (j=0; j<height/2; j++)
    {
        for (i=0; i<width; i++)
        {
            // Pixel from top
            p1 = bgra_data[i + j*width];
            p2 =
                (p1 & 0xff00ff00) |
                ((p1 & 0x00ff0000) >> 16) |
                ((p1 & 0x000000ff) << 16);
            // Pixel from bottom
            p3 = bgra_data[i + (height-1-j)*width];
            p4 =
                (p3 & 0xff00ff00) |
                ((p3 & 0x00ff0000) >> 16) |
                ((p3 & 0x000000ff) << 16);
            // Swap
            bgra_data[i + j*width] = p4;
            bgra_data[i + (height-1-j)*width] = p2;
        }
    }
}


// !!! Allocate dynamically later...
static unsigned char buffer_l[2048*2048*4];
static unsigned char buffer_r[2048*2048*4];
#define ROW_SIZE        1024

////////////////////////////////////////////////////////////////////////////
// deinterlaceFrame
//      Converts the interlaced OVR format to L-top, R-bottom frames.
//      Input is Bayer format, so one byte per pixel.
////////////////////////////////////////////////////////////////////////////
void deinterlaceFrame(void *data_in, uint32_t width, uint32_t height)
{
    uint32_t i, j;
    uint32_t pl, pr;
    uint8_t *data = (uint8_t *) data_in;

    for (j=0; j<height; j++)
    {
        for (i=0; i<width; i++)
        {
            buffer_l[i + j*width] = data[2*(i + j*width)];
            buffer_r[i + j*width] = data[2*(i + j*width)+1];
        }
    }
}

void processProfilePrint()
{
    log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
    log_printf(log_quiet, "Process Thread Time                                        | Output Rate\n");
    log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
    profPrintOne("FIFO Wait", &(p.fifoWait), &(p.profAll), p.numLiveFrames);
    profPrintOne("Flip Cam Data", &(p.flipRemapCam), &(p.profAll), p.numLiveFrames);
    profPrintOne("Flip X11 Data", &(p.flipRemapX11), &(p.profAll), p.numLiveFrames);
    renderProfilePrint(&p.profAll);
    log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
    profPrintOne("Total (Live)", &(p.profAll), &(p.profAll), p.numLiveFrames);
    profPrintOne("Total (All)", &(p.profAll), &(p.profAll), p.numAllFrames);
    log_printf(log_quiet, "%-20s %8.2f\n", "CPU FPS", (double)p.numLiveFrames/profTimeSec(&p.profAll));
    log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
    log_printf(log_quiet, "%-20s %8.2f\n", "FPS (Wall, Live)", (double)p.numLiveFrames/profTimeSec(&p.profWAll));
    log_printf(log_quiet, "%-20s %8.2f\n", "FPS (Wall, All)", (double)p.numAllFrames/profTimeSec(&p.profWAll));
    log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
    renderProfilePrintJitter();
}

int processInit()
{
    int ccode;
    uint32_t data_size;
    
    // Zero out GPU data pointers
    memset(&gpu_mat, 0, sizeof(gpu_mat));
    
    // Store frame size for convenience
    procWidth = mainOptions.cap_width;
    procHeight = mainOptions.cap_height;
    procSize = procWidth*procHeight;
    procSizeBytes = procSize*4;

    if (renderInit() != 0)
    {
        ccode = -1;
        goto backout_base;
    }

    initUndistortRectifyMap(camera_matrix_L, camera_dist_L, Mat(),
                            getOptimalNewCameraMatrix(camera_matrix_L, camera_dist_L, Size(procWidth, procHeight),
                                                      0.60, Size(procWidth, procHeight), 0),
                            Size(procWidth, procHeight), CV_32FC1, undistort_map1_L, undistort_map2_L);
    
    initUndistortRectifyMap(camera_matrix_R, camera_dist_R, Mat(),
                            getOptimalNewCameraMatrix(camera_matrix_R, camera_dist_R, Size(procWidth, procHeight),
                                                      0.60, Size(procWidth, procHeight), 0),
                            Size(procWidth, procHeight), CV_32FC1, undistort_map1_R, undistort_map2_R);

    // Initialize camera undistortion variables
    if (mainOptions.gpu)
    {
        gpu_undistort_map1_L.upload(undistort_map1_L);
        gpu_undistort_map2_L.upload(undistort_map2_L);
        gpu_undistort_map1_R.upload(undistort_map1_R); 
        gpu_undistort_map2_R.upload(undistort_map2_R);

        if ((crMalloc(&gpu_mat.bgr_frame_l_, 3*procWidth*procHeight) != 0) ||
            (crMalloc(&gpu_mat.bgr_frame_r_, 3*procWidth*procHeight) != 0) ||
            (crMalloc(&gpu_mat.gray_frame_lr_, 2*procWidth*procHeight) != 0) ||
            (crMalloc(&gpu_mat.rgba_frame_l_, 4*procWidth*procHeight) != 0) ||
            (crMalloc(&gpu_mat.rgba_frame_r_, 4*procWidth*procHeight) != 0) ||
            (crMalloc(&gpu_mat.rgba_frame_l_out_, 4*procWidth*procHeight) != 0) ||
            (crMalloc(&gpu_mat.rgba_frame_r_out_, 4*procWidth*procHeight) != 0) ||
            (crMalloc(&gpu_mat.rgba_frame_lr_out_, 8*procWidth*procHeight) != 0) ||
            (crMalloc(&gpu_mat.undist_frame_l_, 4*procWidth*procHeight) != 0) ||
            (crMalloc(&gpu_mat.undist_frame_r_, 4*procWidth*procHeight) != 0))
        {
            ccode = -2;
            goto backout_mem;
        }

        gpu_mat.cap_frame_l = cuda::GpuMat(procHeight, procWidth, CV_8UC1, gpuCamRawFrameL);               
        gpu_mat.cap_frame_r = cuda::GpuMat(procHeight, procWidth, CV_8UC1, gpuCamRawFrameR);               
        gpu_mat.gray_frame_lr = cuda::GpuMat(2*procHeight, procWidth, CV_8UC1, gpu_mat.gray_frame_lr_);
        gpu_mat.bgr_frame_l = cuda::GpuMat(procHeight, procWidth, CV_8UC3, gpu_mat.bgr_frame_l_);           
        gpu_mat.bgr_frame_r = cuda::GpuMat(procHeight, procWidth, CV_8UC3, gpu_mat.bgr_frame_r_);
        gpu_mat.rgba_frame_l = cuda::GpuMat(procHeight, procWidth, CV_8UC4, gpu_mat.rgba_frame_l_);
        gpu_mat.rgba_frame_r = cuda::GpuMat(procHeight, procWidth, CV_8UC4, gpu_mat.rgba_frame_r_);
        gpu_mat.rgba_frame_l_out = cuda::GpuMat(procHeight, procWidth, CV_8UC4, gpu_mat.rgba_frame_l_out_);
        gpu_mat.rgba_frame_r_out = cuda::GpuMat(procHeight, procWidth, CV_8UC4, gpu_mat.rgba_frame_r_out_);
        gpu_mat.rgba_frame_lr_out = cuda::GpuMat(2*procHeight, procWidth, CV_8UC4, gpu_mat.rgba_frame_lr_out_);
        gpu_mat.undist_frame_l = cuda::GpuMat(procHeight, procWidth, CV_8UC4, gpu_mat.undist_frame_l_);
        gpu_mat.undist_frame_r = cuda::GpuMat(procHeight, procWidth, CV_8UC4, gpu_mat.undist_frame_r_);

        // Rectangles to simplify gluing frames back together
        gpu_mat.l_sub = Rect(0, 0, procWidth, procHeight);
        gpu_mat.r_sub = Rect(0, procHeight, procWidth, procHeight);
    }
    
    // Clear profiling information
    memset(&p, 0, sizeof(p));
    
    return(0);

 backout_mem:
    if (gpu_mat.bgr_frame_l_) crFree(gpu_mat.bgr_frame_l_);
    if (gpu_mat.bgr_frame_r_) crFree(gpu_mat.bgr_frame_r_);
    if (gpu_mat.gray_frame_lr_) crFree(gpu_mat.gray_frame_lr_);
    if (gpu_mat.rgba_frame_l_) crFree(gpu_mat.rgba_frame_l_);
    if (gpu_mat.rgba_frame_r_) crFree(gpu_mat.rgba_frame_r_);
    if (gpu_mat.rgba_frame_l_out_) crFree(gpu_mat.rgba_frame_l_out_);
    if (gpu_mat.rgba_frame_r_out_) crFree(gpu_mat.rgba_frame_r_out_);
    if (gpu_mat.rgba_frame_lr_out_) crFree(gpu_mat.rgba_frame_lr_out_);
    if (gpu_mat.undist_frame_l_) crFree(gpu_mat.undist_frame_l_);
    if (gpu_mat.undist_frame_r_) crFree(gpu_mat.undist_frame_r_);

 backout_render:
    renderUninit();
    
 backout_base:
    return(ccode);
}

void processUninit()
{
    renderUninit();

    if (gpu_mat.bgr_frame_l_) crFree(gpu_mat.bgr_frame_l_);
    if (gpu_mat.bgr_frame_r_) crFree(gpu_mat.bgr_frame_r_);
    if (gpu_mat.gray_frame_lr_) crFree(gpu_mat.gray_frame_lr_);
    if (gpu_mat.rgba_frame_l_) crFree(gpu_mat.rgba_frame_l_);
    if (gpu_mat.rgba_frame_r_) crFree(gpu_mat.rgba_frame_r_);
    if (gpu_mat.rgba_frame_l_out_) crFree(gpu_mat.rgba_frame_l_out_);
    if (gpu_mat.rgba_frame_r_out_) crFree(gpu_mat.rgba_frame_r_out_);
    if (gpu_mat.rgba_frame_lr_out_) crFree(gpu_mat.rgba_frame_lr_out_);
    if (gpu_mat.undist_frame_l_) crFree(gpu_mat.undist_frame_l_);
    if (gpu_mat.undist_frame_r_) crFree(gpu_mat.undist_frame_r_);
}

////////////////////////////////////////////////////////////////////////////
// cpuRemapFlip
//      Converts the raw Bayer images to RGBA, and flips images
//      vertically.  The purpose of this code is just as a reference of
//      correct operation for the GPU code, so it has not been
//      optimized much.
////////////////////////////////////////////////////////////////////////////
void cpuRemapFlip(void *cam_data, void *x11_data)
{
    Mat cap_frame_l(procHeight, procWidth, CV_8UC1, buffer_l, procWidth);
    Mat cap_frame_r(procHeight, procWidth, CV_8UC1, buffer_r, procWidth);
    Mat rgba_frame_l; 
    Mat rgba_frame_r; 
    Mat rgba_frame_l_out; 
    Mat rgba_frame_r_out; 
    Mat undist_frame_l;
    Mat undist_frame_r;
    unsigned char *cam_data_l, *cam_data_r;

    profCpuStart(&p.flipRemapCam);
    //colorMapFlip(cam_data, procWidth, procHeight*2);
    deinterlaceFrame(cam_data, procWidth, procHeight);

    cvtColor(cap_frame_l, rgba_frame_l, COLOR_BayerGB2RGBA);
    cvtColor(cap_frame_r, rgba_frame_r, COLOR_BayerGB2RGBA);

    // Flip and optionally undistort.
    if (mainOptions.no_undistort)
    {
        // Only flip frames.  
        flip(rgba_frame_l, rgba_frame_l_out, 0);
        flip(rgba_frame_r, rgba_frame_r_out, 0);
    }
    else
    {
        // Remap
        remap(rgba_frame_l, undist_frame_l, undistort_map1_L, undistort_map2_L, INTER_LINEAR, BORDER_CONSTANT);
        remap(rgba_frame_r, undist_frame_r, undistort_map1_R, undistort_map2_R, INTER_LINEAR, BORDER_CONSTANT);

        // Flip frames.  Makes openGL use of data simpler.
        flip(undist_frame_l, rgba_frame_l_out, 0);
        flip(undist_frame_r, rgba_frame_r_out, 0);
    }
                
    // Copy back into the camera data
    cam_data_l = rgba_frame_l_out.ptr<uchar>(0);
    cam_data_r = rgba_frame_r_out.ptr<uchar>(0);
    memcpy(cam_data, cam_data_l, procSizeBytes);
    memcpy((uint8_t *) cam_data + procSizeBytes, cam_data_r, procSizeBytes);

    profCpuEnd(&p.flipRemapCam, 0);
                
    if (mainOptions.x_wid)
    {
        profCpuStart(&p.flipRemapX11);
        colorMapFlip(x11_data, grabX11Width(), grabX11Height());
        profCpuEnd(&p.flipRemapX11, 0);
    }

}

////////////////////////////////////////////////////////////////////////////
// gpuRemapFlip
//      Converts the raw stereo Bayer image from the camera to RGBA
//      data.  Additionally flips the L/R images vertically for more logical
//      orientation in the eventual OpenGL texture.
////////////////////////////////////////////////////////////////////////////
void gpuRemapFlip(void *cam_data, void *x11_data)
{
    uint8_t *cam_data_l, *cam_data_r;
    struct timespec t;
#ifdef PROCESS_SEND_ARUCO    
    arucoFrameDef *aruco_rec_l, *aruco_rec_r;
#endif
    profCpuStart(&p.flipRemapCam);

    // Get timestamp for frame
    clock_gettime(CLOCK_REALTIME, &t);
    
    // Copy the Bayer raw data frame down to the GPU.
    crLoadCamRaw(cam_data);

    crDeinterlace();

    cuda::cvtColor(gpu_mat.cap_frame_l, gpu_mat.bgr_frame_l, COLOR_BayerGB2BGR);
    cuda::cvtColor(gpu_mat.cap_frame_r, gpu_mat.bgr_frame_r, COLOR_BayerGB2BGR);

    cuda::cvtColor(gpu_mat.bgr_frame_l, gpu_mat.rgba_frame_l, COLOR_BGR2RGBA);
    cuda::cvtColor(gpu_mat.bgr_frame_r, gpu_mat.rgba_frame_r, COLOR_BGR2RGBA);

    // Flip and optionally undistort.
    if (mainOptions.no_undistort)
    {
        // Only flip frames.  
        cuda::flip(gpu_mat.rgba_frame_l, gpu_mat.rgba_frame_lr_out(gpu_mat.l_sub), 0);
        cuda::flip(gpu_mat.rgba_frame_r, gpu_mat.rgba_frame_lr_out(gpu_mat.r_sub), 0);

#ifdef PROCESS_SEND_ARUCO        
        if (mainOptions.inst_panel)
        {
            // Generate a gray version of LR frame to be used in Aruco detection.
            cuda::cvtColor(gpu_mat.rgba_frame_l, gpu_mat.gray_frame_lr(gpu_mat.l_sub), COLOR_BGRA2GRAY);
            cuda::cvtColor(gpu_mat.rgba_frame_r, gpu_mat.gray_frame_lr(gpu_mat.r_sub), COLOR_BGRA2GRAY);
            crMemcpyD2H(crHostBuffer, gpu_mat.gray_frame_lr_, procWidth*procHeight*2);
            
            if ((!fifoIsFull(arucoCapFifoL)) && (!fifoIsFull(arucoCapFifoR)))
            {
                aruco_rec_l = (arucoFrameDef *) fifoWritePtr(arucoCapFifoL);
                aruco_rec_r = (arucoFrameDef *) fifoWritePtr(arucoCapFifoR);
                memcpy(&aruco_rec_l->t, &t, sizeof(t));
                memcpy(&aruco_rec_r->t, &t, sizeof(t));
                memcpy(&aruco_rec_l->data, crHostBuffer, procWidth*procHeight);
                memcpy(&aruco_rec_r->data, (uint8_t *) crHostBuffer + procWidth*procHeight, procWidth*procHeight);
                fifoWriteDone(arucoCapFifoL);
                fifoWriteDone(arucoCapFifoR);
            }
        }
#endif        
    }
    else
    {
        // Remap
        cuda::remap(gpu_mat.rgba_frame_l, gpu_mat.undist_frame_l, gpu_undistort_map1_L, gpu_undistort_map2_L, INTER_LINEAR, BORDER_CONSTANT);
        cuda::remap(gpu_mat.rgba_frame_r, gpu_mat.undist_frame_r, gpu_undistort_map1_R, gpu_undistort_map2_R, INTER_LINEAR, BORDER_CONSTANT);

        // Flip frames.  Makes openGL use of data simpler.
        cuda::flip(gpu_mat.undist_frame_l, gpu_mat.rgba_frame_lr_out(gpu_mat.l_sub), 0);
        cuda::flip(gpu_mat.undist_frame_r, gpu_mat.rgba_frame_lr_out(gpu_mat.r_sub), 0);

        if (mainOptions.inst_panel)
        {
            // Generate a gray version of LR frame to be used in Aruco detection.
            cuda::cvtColor(gpu_mat.undist_frame_l, gpu_mat.gray_frame_lr(gpu_mat.l_sub), COLOR_BGRA2GRAY);
            cuda::cvtColor(gpu_mat.undist_frame_r, gpu_mat.gray_frame_lr(gpu_mat.r_sub), COLOR_BGRA2GRAY);
            crMemcpyD2H(crHostBuffer, gpu_mat.gray_frame_lr_, procWidth*procHeight*2);

            if ((!fifoIsFull(arucoCapFifoL)) && (!fifoIsFull(arucoCapFifoR)))
            {
                aruco_rec_l = (arucoFrameDef *) fifoWritePtr(arucoCapFifoL);
                aruco_rec_r = (arucoFrameDef *) fifoWritePtr(arucoCapFifoR);
                memcpy(&aruco_rec_l->t, &t, sizeof(t));
                memcpy(&aruco_rec_r->t, &t, sizeof(t));
                memcpy(&aruco_rec_l->data, crHostBuffer, procWidth*procHeight);
                memcpy(&aruco_rec_r->data, (uint8_t *) crHostBuffer + procWidth*procHeight, procWidth*procHeight);
                fifoWriteDone(arucoCapFifoL);
                fifoWriteDone(arucoCapFifoR);
            }
        }
    }

    
    profCpuEnd(&p.flipRemapCam, 0);

    if (mainOptions.x_wid)
    {
        profCpuStart(&p.flipRemapX11);
        crLoadX11Frame(x11_data);
        crRemapX11Frame();
        profCpuEnd(&p.flipRemapX11, 0);
    }
}


void processRun()
{
    int framesProcessed;
    void *cam_data, *x11_data;

    profCpuStart(&p.profAll);
    profWallStart(&p.profWAll);
    p.numLiveFrames = 0;
    p.numAllFrames = 0;
    framesProcessed = 0;
    x11_data = 0;
    
    for ( ; ; )
    {
        // Check for data available
        if (fifoIsEmpty(captureCamFifo) || (mainOptions.x_wid && fifoIsEmpty(captureX11Fifo)))
        {
            // No new data is available
            if (mainOptions.render_vr)
            {
                // For VR: Render old data to allow the headset to
                // keep tracking at 90 Hz.
                renderFrameBlank();
            }
            else
            {
                // If not VR, wait for live data...
                WAIT_SHORT;
                continue;
            }
        }
        else
        {
            cam_data = fifoReadPtr(captureCamFifo);
            if (mainOptions.x_wid)
            {
                x11_data = fifoReadPtr(captureX11Fifo);
            }

            // Flip images and remap colors with CPU
            if (!mainOptions.gpu)
            {
                cpuRemapFlip(cam_data, x11_data);

                // Render frame to chosen output devices
                // Note: No profiling here, since it is done inside renderFrame()
                renderFrame((unsigned char *)cam_data, (unsigned char *)x11_data);
            }
            else // GPU processing
            {
                // Flip frame
                gpuRemapFlip(cam_data, x11_data);

                // Render.  Left and right frames are passed in as data.
                // renderFrame knows to interpret these as GPU pointers.
                renderFrame(gpu_mat.rgba_frame_lr_out_, crHostBuffer);
            }
            
            fifoReadDone(captureCamFifo);
            fifoReadDone(captureX11Fifo);
        
            log_printf(log_info, "[%d] ", framesProcessed+1);
            fflush(stdout);

            p.numLiveFrames++;
            framesProcessed++;
        }
        p.numAllFrames++;

        if (stopping) break;
    }
    
    profCpuEnd(&p.profAll, 0);
    profWallEnd(&p.profWAll);
}

// Helper functions to allow the aruco threads to undistort the camera image before doing the
// detection.  This is duplicated work, since the same thing is done on the GPU, but I'm not
// sure how else to do it.  Doing it in the capture thread in CPU space would increase rendering latency.

void processUndistortL(Mat& rgba_frame_l, Mat& undist_frame_l)
{
    remap(rgba_frame_l, undist_frame_l, undistort_map1_L, undistort_map2_L, INTER_LINEAR, BORDER_CONSTANT);
}

void processUndistortR(Mat& rgba_frame_r, Mat& undist_frame_r)
{
    remap(rgba_frame_r, undist_frame_r, undistort_map1_R, undistort_map2_R, INTER_LINEAR, BORDER_CONSTANT);
}




