////////////////////////////////////////////////////////////////////////////
// render.cc
//      Code that supports combining and rendering of camera video
//      and scenery to the VR headset.
//
// Author: Jon W. Wallace, Lafayette College
// Date:   May 2020
////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <GL/glew.h>
#include <GLFW/glfw3.h>
#include <math.h>

#include "render.h"

#include "log.h"
#include "main.h"
#include "glh.h"
#include "vr_helper.h"
#include "grab_x11.h"
#include "myProfile.h"
#include "ar_parm.h"
#include "renderState.h"
#include "cudaRender.h"
#include "udp_send_hmd_pos.h"
#include "gauges.h"
#include "xplaneXfer.h"
#include "render_file.h"

// This define causes gauges to be detected in render loop rather
// than in the capture loop.
// #define RENDER_DETECT_GAUGES

// Uncomment this to generate a file of input sample times to
// analyze jitter.
//#define DUMP_OUTPUT_JITTER
static const char *fname_jitter = "out_jitter.txt";
static FILE *f_jitter;
static double t_start;

// Fraction of L/R frames to clip off when rendering to X
//#define X_RENDER_CLIP 0.2
//#define Y_RENDER_CLIP 0.2
//#define X_RENDER_CLIP 0
//#define Y_RENDER_CLIP 0

// Handle to main window
GLFWwindow *gWindow = 0;

// Texture holding captured video frame
glhFbDef fb_cam_frame;

// Texture holding captured X11 frame
glhFbDef fb_x11_frame;
// Cropped version of fb_x11_frame for L/R eyes.
glhFbDef fb_x11_frame_l, fb_x11_frame_r;

// Output framebuffers for L/R eyes
glhFbDef fb_out_l, fb_out_r;
glhFbDef fb_out_flip;
glhFbDef fb_out_lr;
uint32_t x_render_width, x_render_height;

// OpenVR rendering variables
static IVRSystem *pVr=0;
uint32_t vr_render_width, vr_render_height;
TrackedDevicePose_t trackedDevicePose[k_unMaxTrackedDeviceCount];

// Render size to use on output textues.
// If using VR HMD, this overrides any X settings.
uint32_t out_render_width, out_render_height;

// Put variables here that control the state of
// what is being rendered.  The state can be changed
// by pressing keys.
renderStateDef renderState;

// Profiling information
typedef struct
{
    cpuProfDef gpuCopy;
    cpuProfDef gpuRemap;
    cpuProfDef copyCamData;
    cpuProfDef copyX11Data;
    cpuProfDef detectTags;
    cpuProfDef drawCam;
    cpuProfDef renderGlfw;
    cpuProfDef renderVr;
    cpuProfDef renderFile;
    cpuProfDef marker;
    uint32_t numFrames;

    // Output jitter measurement
    double dt_mean, dt_mom2;
    uint32_t N_jitter;

} profRenderDef;
static profRenderDef p;

// Positioning of X11 grabbed scenery on output texture
// (find a better place for these)
float scen_lx = 0.52;
float scen_ly = 0.46;
float scen_lw = 0.63;
float scen_lh = 0.56;

float scen_rx = 0.48;
float scen_ry = 0.46;
float scen_rw = 0.63;
float scen_rh = 0.56;

////////////////////////////////////////////////////////////////////////////
// renderProfilePrint
//      Print profiling information for render portion of processing.
//      Should be called from the process profiling printer.
// Inputs:
//      profAll         CPU time profiler for all time in the process
//                      thread.
////////////////////////////////////////////////////////////////////////////
void renderProfilePrint(cpuProfDef *profAll)
{
    profPrintOne("GPU Copy", &(p.gpuCopy), profAll, p.numFrames);
    profPrintOne("GPU Remap", &(p.gpuRemap), profAll, p.numFrames);
    profPrintOne("Copy Cam Data", &(p.copyCamData), profAll, p.numFrames);
    profPrintOne("Copy X11 Data", &(p.copyX11Data), profAll, p.numFrames);
    profPrintOne("Detect Tags", &(p.detectTags), profAll, p.numFrames);
    profPrintOne("Draw Camera", &(p.drawCam), profAll, p.numFrames);
    profPrintOne("Render GLFW", &(p.renderGlfw), profAll, p.numFrames);
    profPrintOne("Render VR", &(p.renderVr), profAll, p.numFrames);
    profPrintOne("Render File", &(p.renderFile), profAll, p.numFrames);
    profPrintOne("Cal/Marker", &(p.marker), profAll, p.numFrames);
}

void renderProfilePrintJitter()
{
    double dt_mean, dt_std;
    
    dt_mean = p.dt_mean/(double)p.N_jitter;
    dt_std = sqrt(p.dt_mom2/(double)p.N_jitter - dt_mean*dt_mean);

    log_printf(log_quiet, "Render Jitter (mean / std): %.3f ms / %.3f ms\n", dt_mean*1000.0, dt_std*1000.0);
    log_printf(log_quiet, "--------------------------------------------------------------------------------------\n");
}


void glfw_error_callback(int error, const char* description)
{
    log_printf(log_error, "GLFW Error: %s\n", description);
}

void glfw_key_callback(GLFWwindow* window, int key, int scancode, int action, int mods)
{
    static char mod_parm = ' ';
    static char mod_parm2 = ' ';
    
    if (action == GLFW_PRESS)
    {
        switch(key)
        {
            case GLFW_KEY_ESCAPE:
                stopping = 1;
                break;

            case GLFW_KEY_A:
                vrhSetPilotOrigin(trackedDevicePose[0].mDeviceToAbsoluteTracking);
                break;

            case GLFW_KEY_C:
                if (renderState.cal_mode)
                {
                    renderState.cal_mode = 0;
                    vrhCameraCalEnd();
                    log_printf(log_quiet, "Calibration mode OFF.\n");
                }
                else
                {
                    if (vrhCameraCalStart() == 0)
                    {
                        log_printf(log_quiet, "Calibration mode ON.\n");
                        renderState.cal_mode = 1;
                    }
                }
                break;

            case GLFW_KEY_G:
                // Allow toggling between the green screen and geometric
                // cockpit modes.
                mainOptions.green_screen = !mainOptions.green_screen;
                break;
                
            case GLFW_KEY_1:
                mod_parm = 'h';
                break;
            case GLFW_KEY_2:
                mod_parm = 'H';
                break;
            case GLFW_KEY_3:
                mod_parm = 's';
                break;
            case GLFW_KEY_4:
                mod_parm = 'r';
                break;
            case GLFW_KEY_R:
                // "Rotate"
                mod_parm = 'R';
                break;
            case GLFW_KEY_T:
                // "Translate"
                mod_parm = 'T';
                break;
            case GLFW_KEY_X:
                mod_parm2 = 'x';
                break;
            case GLFW_KEY_Y:
                mod_parm2 = 'y';
                break;
            case GLFW_KEY_Z:
                mod_parm2 = 'z';
                break;
            case GLFW_KEY_M:
                renderState.marker_mode = !renderState.marker_mode;
                break;
            case GLFW_KEY_S:
                renderState.scenery_off = !renderState.scenery_off;
                break;
            case GLFW_KEY_P:
                fprintf(stderr, "\n");
                fprintf(stderr, "green_h_min = %f\n", renderState.green_h_min);
                fprintf(stderr, "green_h_max = %f\n", renderState.green_h_max);
                fprintf(stderr, "green_s_min = %f\n", renderState.green_s_min);
                fprintf(stderr, "green_avg_rad = %d\n", renderState.green_avg_rad);
                fprintf(stderr, "x = %f\n", vrhPilotOffset[0]);
                fprintf(stderr, "y = %f\n", vrhPilotOffset[1]);
                fprintf(stderr, "z = %f\n", vrhPilotOffset[2]);
                break;
            case GLFW_KEY_KP_ADD:
                switch(mod_parm)
                {
                    case 'h':
                        renderState.green_h_min += 1.0;
                        break;

                    case 'H':
                        renderState.green_h_max += 1.0;
                        break;

                    case 's':
                        renderState.green_s_min += 0.01;
                        break;
                        
                    case 'r':
                        renderState.green_avg_rad++;
                        break;

                    case 'T':  // Translate pilot pos
                        switch(mod_parm2)
                        {
                            case 'x':
                                pilotTrans(-0.01, 0, 0);
                                break;
                            case 'y':
                                pilotTrans(0, -0.01, 0);
                                break;
                            case 'z':
                                pilotTrans(0, 0, -0.01);
                                break;
                        }
                        break;
                    case 'R': // Rotate Pilot pos
                        switch(mod_parm2)
                        {
                            case 'x':
                                pilotRotX(1.0);
                                break;
                            case 'y':
                                pilotRotY(1.0);
                                break;
                            case 'z':
                                pilotRotZ(1.0);
                                break;
                        }
                        break;
                }
                break;
                
            case GLFW_KEY_KP_SUBTRACT:
                switch(mod_parm)
                {
                    case 'h':
                        renderState.green_h_min -= 1.0;
                        break;
                        
                    case 'H':
                        renderState.green_h_max -= 1.0;
                        break;
                        
                    case 's':
                        renderState.green_s_min -= 0.01;
                        break;
                        
                    case 'r':
                        renderState.green_avg_rad--;
                        break;
                        
                    case 'T':  // Translate pilot pos
                        switch(mod_parm2)
                        {
                            case 'x':
                                pilotTrans(0.01, 0, 0);
                                break;
                            case 'y':
                                pilotTrans(0, 0.01, 0);
                                break;
                            case 'z':
                                pilotTrans(0, 0, 0.01);
                                break;
                        }
                        break;
                    case 'R': // Rotate Pilot pos
                        switch(mod_parm2)
                        {
                            case 'x':
                                pilotRotX(-1.0);
                                break;
                            case 'y':
                                pilotRotY(-1.0);
                                break;
                            case 'z':
                                pilotRotZ(-1.0);
                                break;
                        }
                        break;
                }
                break;
        }
    }
}

int renderInit()
{
    int ccode;
    HmdError peError;
    GLuint x11_tex_id;
    
    out_render_width = 0;
    out_render_height = 0;
    vr_render_width = 0;
    vr_render_height = 0;

    memset(&renderState, 0, sizeof(renderState));
    renderState.green_h_min = 97.0;
    renderState.green_h_max = 192.0;
    renderState.green_s_min = 0.13;
    renderState.green_avg_rad = 7;

#ifdef DUMP_OUTPUT_JITTER
    f_jitter = fopen(fname_jitter, "wb");
    t_start = profWallTime();
#endif

    // OpenVR Initialization ///////////////////////////////////////////////////

    if (mainOptions.render_vr)
    {
        // Initialize VR
        if ((pVr = VR_Init(&peError, VRApplication_Scene)) == 0)
        {
            log_printf(log_error, "VR_Init() failed: %s.\n",
                       VR_GetVRInitErrorAsEnglishDescription(peError));
            ccode = -1;
            goto backout_base;
            
        }

        // Get recommended size of output textures
        pVr->GetRecommendedRenderTargetSize(&vr_render_width, &vr_render_height);

        log_printf(log_info, "OpenVR Output Texture Size: %d x %d\n", vr_render_width, vr_render_height);
        
        // Get a compsitor instance
        if (VRCompositor() == 0)
        {
            fprintf(stderr, "No compositor available.\n");
            ccode = -1;
            goto backout_vr;
        }

        // Initialize helper functions
        if (vrhInit(pVr) != 0)
        {
            goto backout_vr;
        }

        // Get initial pose
        VRCompositor()->WaitGetPoses(trackedDevicePose, k_unMaxTrackedDeviceCount, NULL, 0);
        // Set pilot origin to be this position
        vrhSetPilotOrigin(trackedDevicePose[0].mDeviceToAbsoluteTracking);
    }
    
    // X11 Init (GLFW) /////////////////////////////////////////////////////////

    x_render_width = mainOptions.render_width;
    x_render_height = mainOptions.render_height;
    
    // Initialize GLFW, which is used for OpenGL context.
    if (!glfwInit())
    {
        log_printf(log_error, "Unable to start GLFW.\n");
        ccode = -1;
        goto backout_vr;
    }

    if ((gWindow = glfwCreateWindow(x_render_width*2, x_render_height, "AR Rendering", NULL, NULL)) == 0)
    {
        log_printf(log_error, "Unable to create rendering window.\n");
        ccode = -2;
        goto backout_glfw;
    }
    
    glfwMakeContextCurrent(gWindow);

    log_printf(log_info, "OpenGL Vendor: %s\n", (const char*)glGetString(GL_VENDOR));

    glfwSetErrorCallback(glfw_error_callback);
    glfwSetKeyCallback(gWindow, glfw_key_callback);
    
    glewInit();

    glEnable(GL_MULTISAMPLE);
    
    // OpenGL Allocation ///////////////////////////////////////////////////////

    if (mainOptions.render_vr)
    {
        out_render_width = vr_render_width;
        out_render_height = vr_render_height;
    }
    else
    {
        out_render_width = x_render_width;
        out_render_height = x_render_height;
    }
    
    // Allocate required textures
    
    // Output render texture for left eye
    if (glhFbNew(&fb_out_l, out_render_width, out_render_height, 1, 0) != 0)
    {
        log_printf(log_error, "Unable to allocate left output framebuffer.\n");
        ccode = -3;
        goto backout_glfw;
    }
    
    // Output render texture for right eye
    if (glhFbNew(&fb_out_r, out_render_width, out_render_height, 1, 0) != 0)
    {
        log_printf(log_error, "Unable to allocate right output framebuffer.\n");
        ccode = -3;
        goto backout_glfw;
    }

    // Allocate additional textures used to pre-process rendered frames before
    // writing them.
    if (mainOptions.render_file) 
    {
        if (mainOptions.render_file_type == 'b')
        {
            if (glhFbNew(&fb_out_lr, 2*out_render_width, out_render_height, 1, 0) != 0)
            {
                log_printf(log_error, "Unable to allocate right output framebuffer.\n");
                ccode = -3;
                goto backout_glfw;
            }
        }
        else
        {
            if (glhFbNew(&fb_out_flip, out_render_width, out_render_height, 1, 0) != 0)
            {
                log_printf(log_error, "Unable to allocate right output framebuffer.\n");
                ccode = -3;
                goto backout_glfw;
            }
        }
    }

    // Texture to hold captured video.  Stored in top/bottom format for R/L eyes.
    if (glhFbNew(&fb_cam_frame, mainOptions.cap_width, mainOptions.cap_height*2, 0, 0) != 0)
    {
        log_printf(log_error, "Unable to allocate camera OpenGL framebuffer.\n");
        ccode = -3;
        goto backout_glfw;
    }
    
    // Texture to hold captured X11 data (flightsim scenery)
    x11_tex_id = 0;
    if (mainOptions.x_wid)
    {
        if (glhFbNew(&fb_x11_frame, grabX11Width(), grabX11Height(), 0, 0) != 0)
        {
            log_printf(log_error, "Unable to allocate X11 capture framebuffer.\n");
            ccode = -3;
            goto backout_glfw;
        }
        x11_tex_id = fb_x11_frame.tex_id;

        // Cropped/scaled version for use with left eye
        if (glhFbNew(&fb_x11_frame_l, out_render_width, out_render_height, 1, 0) != 0)
        {
            log_printf(log_error, "Unable to allocate left X11 capture framebuffer.\n");
            ccode = -3;
            goto backout_glfw;
        }
        
        // Cropped/scaled version for use with right eye
        if (glhFbNew(&fb_x11_frame_r, out_render_width, out_render_height, 1, 0) != 0)
        {
            log_printf(log_error, "Unable to allocate right X11 capture framebuffer.\n");
            ccode = -3;
            goto backout_glfw;
        }
    }

    // Gauges ///////////////////////////////////////////////////////////////////

#ifdef RENDER_DETECT_GAUGES    
    if (gaugesDetectInit(mainOptions.inst_panel_aruco_dict, mainOptions.cap_width, mainOptions.cap_height) != 0)
    {
        log_printf(log_error, "gaugesDetectInit() failed.\n");
        ccode = -5;
        goto backout_glfw;
    }
#endif
    
    if (gaugesDrawInit() != 0)
    {
        log_printf(log_error, "gaugesDrawInit() failed.\n");
        ccode = -1;
        goto backout_gauges1;
    }
    
    // CUDA ////////////////////////////////////////////////////////////////////
    if (mainOptions.gpu)
    {
        if (crInit(fb_cam_frame.tex_id, x11_tex_id) != 0)
        {
            log_printf(log_error, "Unable to initialize CUDA processing.\n");
            ccode = -4;
            goto backout_gauges2;
        }
        if (crFindGreenDotInit() != 0)
        {
            log_printf(log_error, "Unable to initialize green dot CUDA algorithm.\n");
            ccode = -4;
            goto backout_cuda1;
        }
    }
    
    // File Rendering //////////////////////////////////////////////////////////
    if (mainOptions.render_file)
    {
        int width_use;
        
        if (mainOptions.render_file_type == 'b')
        {
            width_use = 2*out_render_width;
        }
        else
        {
            width_use = out_render_width;
        }
        
        if (renderFileInit(mainOptions.render_file_fname, width_use, out_render_height, mainOptions.render_file_options) != 0)
        {
            log_printf(log_error, "File renderer initialization failed.\n");
            goto backout_cuda2;
        }
    }
    
    // Clear profiling info.
    memset(&p, 0, sizeof(p));

    return(0);

 backout_file:
    renderFileUninit();

 backout_cuda2:
    if (mainOptions.gpu) crFindGreenDotUninit();

 backout_cuda1:
    if (mainOptions.gpu) crUninit();
    
 backout_gauges2:
    gaugesDrawUninit();
    
 backout_gauges1:
#ifdef RENDER_DETECT_GAUGES    
    gaugesDetectUninit();
#endif    
    
 backout_glfw:
    glfwTerminate();

 backout_vr:
    if (mainOptions.render_vr) VR_Shutdown();

 backout_base:
    return(ccode);
}

////////////////////////////////////////////////////////////////////////////
// drawVideoGreen
//      Draws video on the L/R output textures assuming green-screen
//      detection mode.  Here, the background (flightsim) image is drawn
//      first, followed by camera video.  Green regions in the camera video
//      are made transparent, allowing the background image to show through.
//
//      Source: Combined R(Top)/L(Bottom) frame buffer.
//      Dest:   Separate L/R VR render frames.
////////////////////////////////////////////////////////////////////////////
void drawVideoGreen()
{
    int do_blank = 1;
    
    if (!renderState.scenery_off)
    {
        if (mainOptions.x_wid)
        {
            // Copy grabbed image to the L/R output framebuffers
            glhFbCopyRect(&fb_x11_frame, &fb_out_l,
                          0.5, 0.5, 1.0, 1.0,
                          scen_lx, scen_ly, scen_lw, scen_lh,
                          do_blank,
                          0);
            
            glhFbCopyRect(&fb_x11_frame, &fb_out_r,
                          0.5, 0.5, 1.0, 1.0,
                          scen_rx, scen_ry, scen_rw, scen_rh,
                          do_blank,
                          0);
        }
        do_blank = 0;
    }
    
    // Now copy camera data.  If green screen is enabled, there will be
    // transparent regions allowing the X11 video to show through.

    glhFbCopyQuad(&fb_cam_frame, &fb_out_l,
                  s_cap_crop_lx, s_cap_crop_ly, s_cap_crop_w, s_cap_crop_h,
                  s_rend_vid_lx1, s_rend_vid_ly1,
                  s_rend_vid_lx2, s_rend_vid_ly2,
                  s_rend_vid_lx3, s_rend_vid_ly3,
                  s_rend_vid_lx4, s_rend_vid_ly4,
                  do_blank, 0);
    
    glhFbCopyQuad(&fb_cam_frame, &fb_out_r,
                  s_cap_crop_rx, s_cap_crop_ry, s_cap_crop_w, s_cap_crop_h,
                  s_rend_vid_rx1, s_rend_vid_ry1,
                  s_rend_vid_rx2, s_rend_vid_ry2,
                  s_rend_vid_rx3, s_rend_vid_ry3,
                  s_rend_vid_rx4, s_rend_vid_ry4,
                  do_blank, 0);
}

////////////////////////////////////////////////////////////////////////////
// drawVideoGeom
//      Uses a geometric model to combine flightsim scenery and camera
//      video.  In this case, the camera video is drawn first as the
//      background, and then the flightsim video is drawn where the
//      airplane windows are.
////////////////////////////////////////////////////////////////////////////
void drawVideoGeom()
{
    // Camera data first.
    glhFbCopyQuad(&fb_cam_frame, &fb_out_l,
                  s_cap_crop_lx, s_cap_crop_ly, s_cap_crop_w, s_cap_crop_h,
                  s_rend_vid_lx1, s_rend_vid_ly1,
                  s_rend_vid_lx2, s_rend_vid_ly2,
                  s_rend_vid_lx3, s_rend_vid_ly3,
                  s_rend_vid_lx4, s_rend_vid_ly4,
                  1, 0);
    
    glhFbCopyQuad(&fb_cam_frame, &fb_out_r,
                  s_cap_crop_rx, s_cap_crop_ry, s_cap_crop_w, s_cap_crop_h,
                  s_rend_vid_rx1, s_rend_vid_ry1,
                  s_rend_vid_rx2, s_rend_vid_ry2,
                  s_rend_vid_rx3, s_rend_vid_ry3,
                  s_rend_vid_rx4, s_rend_vid_ry4,
                  1, 0);

    // Draw the grabbed X11 (scenery data) onto temporary L/R scenery frames
    if (!renderState.scenery_off)
    {
        glhFbCopyRect(&fb_x11_frame, &fb_x11_frame_l,
                      0.5, 0.5, 1.0, 1.0,
                      scen_lx, scen_ly, scen_lw, scen_lh,
                      0, 0);
        
        glhFbCopyRect(&fb_x11_frame, &fb_x11_frame_r,
                      0.5, 0.5, 1.0, 1.0,
                      scen_rx, scen_ry, scen_rw, scen_rh,
                      0, 0);

        // Now draw cockpit windows, using the textures above to provide
        // the scenery.

        // Left
        vrhDrawSceneryCockpit(&fb_out_l, &fb_x11_frame_l, 1);
        // Right
        vrhDrawSceneryCockpit(&fb_out_r, &fb_x11_frame_r, 0);
    }
}

////////////////////////////////////////////////////////////////////////////
// vrRenderGlfw()
//      Renders the VR scenery using a GLFW window.  Can be used in
//      conjunction with (or instead of) sending video to the VR
//      headset.
////////////////////////////////////////////////////////////////////////////
void vrRenderGlfw(GLfloat trim_x, GLfloat trim_y)
{
    GLfloat x1, y1, x2, y2;

    // Get coordinates of texture to use based on trim
    x1 = trim_x;
    x2 = 1.0-trim_x;
    y1 = trim_y;
    y2 = 1.0-trim_y;

    glPushMatrix();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0,   // left
            1.0,   // right
            0.0,   // bottom
            1.0,   // top
            1.0,  // near
            -1.0);  // far

    // Render the L frame on left side
    
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    glViewport(0, 0, 2*x_render_width, x_render_height);

    glClearColor(1,1,1,1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glBindTexture(GL_TEXTURE_2D, fb_out_l.tex_id);

    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glBegin(GL_QUADS);
    // Lower left
    glTexCoord2f(x1, y1);
    glVertex2d(0, 0);
    // Lower Right
    glTexCoord2f(x2, y1);
    glVertex2d(0.5, 0);
    // Upper Right
    glTexCoord2f(x2, y2);
    glVertex2d(0.5, 1.0);
    // Upper Left
    glTexCoord2f(x1, y2);
    glVertex2d(0.0, 1.0);
    glEnd();

    // Render the R frame on the right side
    
    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glBindTexture(GL_TEXTURE_2D, fb_out_r.tex_id);

    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glBegin(GL_QUADS);
    // Lower left
    glTexCoord2f(x1, y1);
    glVertex2d(0.5, 0);
    // Lower Right
    glTexCoord2f(x2, y1);
    glVertex2d(1.0, 0);
    // Upper Right
    glTexCoord2f(x2, y2);
    glVertex2d(1.0, 1.0);
    // Upper Left
    glTexCoord2f(x1, y2);
    glVertex2d(0.5, 1.0);
    glEnd();

    glPopMatrix();

    // Don't let the graphics card throttle us.
    // We will handle it ourselves.
    glfwSwapInterval(0);                
    glfwSwapBuffers(gWindow);
}

void vrRenderHmd()
{
    EVRCompositorError err1;
    static double t0=0.0, t1=0.0, dt;
    static vrhYprPoseDef pose;
    
    // Get the headset pose
    VRCompositor()->WaitGetPoses(trackedDevicePose, k_unMaxTrackedDeviceCount, NULL, 0);

    if (mainOptions.fg_hmd_port)
    {
        // Get the pose in YPR format for flight simulator.
        vrhComputeHmdYprPose(&pose);
        // Send pose via UDP to flight simulator
        udpSendPose(&pose);
    }

    if (mainOptions.xplane)
    {
        // Get the pose in YPR format for flight simulator.
        vrhComputeHmdYprPose(&pose);

        // Copy data for x-plane
        xplane_shm_local.w.pilots_head_x = pose.x;
        xplane_shm_local.w.pilots_head_y = pose.y;
        xplane_shm_local.w.pilots_head_z = pose.z;
        xplane_shm_local.w.pilots_head_psi = -pose.Y*180.0/M_PI;
        xplane_shm_local.w.pilots_head_the = pose.P*180.0/M_PI;
        xplane_shm_local.w.field_of_view_roll_deg = pose.R*180.0/M_PI;
        xplaneXferPut();
    }
        
    // Update view matrix for vr-helper functions
    vrhUpdatePose();
    
    if ((err1 = VRCompositor()->Submit(Eye_Left, &(fb_out_l.tex))) != 0)
    {
        log_printf(log_error, "OpenVR Error on Submit(Left): %d\n", err1);
    }

    if ((err1 = VRCompositor()->Submit(Eye_Right, &(fb_out_r.tex))) != 0)
    {
        log_printf(log_error, "OpenVR Error on Submit(Right): %d\n", err1);
    }

    t0 = t1;
    t1 = profWallTime();
#ifdef DUMP_OUTPUT_JITTER
    fprintf(f_jitter, "%.9e\n", t1-t_start);
#endif

    // Measure jitter in the output stream
    if (p.numFrames > 10)
    {
        dt = t1-t0;
        p.dt_mean += dt;
        p.dt_mom2 += dt*dt;
        p.N_jitter++;
    }
}

void renderFrameBlank()
{
    if (mainOptions.render_vr)
    {
        profCpuStart(&p.renderVr);
        vrRenderHmd();
        profCpuEnd(&p.renderVr, 0);
    }
}

void renderFrame(void *ptr1, void *ptr2) 
{
    uint8_t *cam_data, *x11_data;
    uint8_t *gpuFrameLR;
    int bytes_written;
    float cs;
    
    if (mainOptions.gpu)
    {
        gpuFrameLR = (uint8_t *) ptr1;
        cam_data = (uint8_t *) ptr2;
        
        // Do possible green screen detection and copy
        // the pre-processed LR camera frame into the OpenGL texture.
        profCpuStart(&p.gpuRemap);
        crRemapFrames(gpuFrameLR);
        profCpuEnd(&p.gpuRemap, 0);
    }
    else
    {
        // CPU only processing:
        // Copy Preprocessed data to OpenGL
        cam_data = (uint8_t *) ptr1;
        x11_data = (uint8_t *) ptr2;
        
        // Copy camera data into the combined LR camera texture.
        profCpuStart(&p.copyCamData);
        glhFbCpuWrite(&fb_cam_frame, cam_data);
        profCpuEnd(&p.copyCamData, fb_cam_frame.width*fb_cam_frame.height*4);
    
        // If we are grabbing X11 scenery from the flight simulator,
        // copy this down to its respective texture as well.
        if (mainOptions.x_wid)
        {
            profCpuStart(&p.copyX11Data);
            glhFbCpuWrite(&fb_x11_frame, x11_data);
            profCpuEnd(&p.copyX11Data, fb_x11_frame.width*fb_x11_frame.height*4);
        }
    }

    // Find location of gauge tags
    profCpuStart(&p.detectTags);
#ifdef RENDER_DETECT_GAUGES
    if (mainOptions.inst_panel)
    {
        if (mainOptions.gpu)
        {
            // cam_data is on host, but is just gray data.
            gaugesDetectTags(cam_data, cam_data+fb_cam_frame.width*fb_cam_frame.height/2);
        }
        else
        {
            gaugesDetectTags(cam_data, cam_data+fb_cam_frame.width*fb_cam_frame.height/2*4);
        }
    }
#endif    
    profCpuEnd(&p.detectTags, 0);

    profCpuStart(&p.drawCam);
    if (mainOptions.green_screen)
    {
        // Map/Draw camera video onto output frames in
        // Green-screen mode
        drawVideoGreen();
    }
    else
    {
        // Draw camera video assuming geometric mode, where
        // a model of airplane is used to know where to put
        // scenery.
        drawVideoGeom();
    }

    if (mainOptions.inst_panel)
    {
        gaugesDraw(&fb_out_l, &fb_out_r);
    }
    
    profCpuEnd(&p.drawCam, 0);

    // Do calibration and/or marker mode
    if (mainOptions.render_vr)
    {
        if (renderState.marker_mode)
        {
            profCpuStart(&p.marker);
            vrhDrawMarker(&fb_out_l, &fb_out_r, (uint32_t *) cam_data, (uint32_t *) gpuFrameLR,
                          mainOptions.cap_width, mainOptions.cap_height);
            profCpuEnd(&p.marker, 0);
        }

        if (renderState.cal_mode)
        {
            profCpuStart(&p.marker);
            vrhCameraCalSample((uint32_t *) cam_data, (uint32_t *) gpuFrameLR,
                               mainOptions.cap_width, mainOptions.cap_height, &fb_out_l, &fb_out_r);
            profCpuEnd(&p.marker, 0);
        }
    }

    // Draw L/R frames onto an GLFW window
    profCpuStart(&p.renderGlfw);
    if (mainOptions.render_x11)
    {
        vrRenderGlfw(mainOptions.render_clip/2.0, mainOptions.render_clip/2.0);
    }
    glfwPollEvents();    
    profCpuEnd(&p.renderGlfw, 0);
    
    if (mainOptions.render_vr)
    {
        profCpuStart(&p.renderVr);
        vrRenderHmd();
        profCpuEnd(&p.renderVr, 0);
        
    }

    if (mainOptions.render_file)
    {
        cs = 1.0 - mainOptions.render_clip;
        
        profCpuStart(&p.renderFile);
        switch(mainOptions.render_file_type)
        {
            case 'l':
                glhFbCopyRect(&fb_out_l, &fb_out_flip,
                              0.5, 0.5, cs, -cs,
                              0.5, 0.5, 1.0, 1.0,
                              0, 0);
                bytes_written = renderFileTexture(&fb_out_flip);
                break;

            case 'r':
                glhFbCopyRect(&fb_out_r, &fb_out_flip,
                              0.5, 0.5, cs, -cs,
                              0.5, 0.5, 1.0, 1.0,
                              0, 0);
                bytes_written = renderFileTexture(&fb_out_flip);
                break;
                
            case 'b':
                glhFbCopyRect(&fb_out_l, &fb_out_lr,
                              0.5, 0.5, cs, -cs,
                              0.25, 0.5, 0.5, 1.0,
                              0, 0);
                glhFbCopyRect(&fb_out_r, &fb_out_lr,
                              0.5, 0.5, cs, -cs,
                              0.75, 0.5, 0.5, 1.0,
                              0, 0);
                bytes_written = renderFileTexture(&fb_out_lr);
                break;
        }
        profCpuEnd(&p.renderFile, bytes_written);
    }
    
    p.numFrames++;
}

void renderUninit()
{
    if (mainOptions.render_file)
    {
        renderFileUninit();
    }

#ifdef DUMP_OUTPUT_JITTER
    fclose(f_jitter);
#endif

#ifdef RENDER_DETECT_GAUGES
    gaugesDetectUninit();
#endif    
    gaugesDrawUninit();
    
    if (mainOptions.render_vr)
    {
        // Uninit helper functions
        vrhUninit();

        // Ensure that eye rendering is done.  
        VRCompositor()->WaitGetPoses(trackedDevicePose, 1, NULL, 0);
        VR_Shutdown();
    }
    
    glfwDestroyWindow(gWindow);
    glfwTerminate();
}

    
