////////////////////////////////////////////////////////////////////////////
// udp_recv_stats.cc
//     Code to support receiving updates on aircraft information from
//     the flight simulator for updating gauges.  This is only used when
//     operating with the FlightGear flight simulator.
//
// Author: Jon W. Wallace, Lafayette College
// Date:   May 2020
////////////////////////////////////////////////////////////////////////////

#ifdef __WINDOWS__
#include <Windows.h>
#else
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netdb.h>
#include <unistd.h>
#include <pthread.h>
#endif
#include <signal.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "udp_recv_stats.h"

#include "main.h"
#include "log.h"

#define TIMEOUT_100MS     100000
#ifdef __WINDOWS__
#define SOCK_OPT_CAST   (const char *)
#define SEND_BUFF_CAST  (const char *)
#define RECV_BUFF_CAST  (char *)
#else
#define SOCK_OPT_CAST
#define SEND_BUFF_CAST  (void *)
#define RECV_BUFF_CAST
#endif

#define UDP_MAX_PACKET_SIZE             1024
static char udp_packet[UDP_MAX_PACKET_SIZE];

static int udp_recv_socket = 0;
static pthread_t udpThread;
static volatile int stop_thread;
static volatile int thread_stopped;

volatile udpRecvStatDef udp_stat_data;

void udpRecvStatThreadFunc();

////////////////////////////////////////////////////////////////////////////
// udpRecvStatInit()
//      Initializes structures needed.
////////////////////////////////////////////////////////////////////////////
int udpRecvStatInit(int port)
{
    int ccode;
    int sock_opt;
    struct sockaddr_in addr, addr2;

    memset((void *) &udp_stat_data, 0, sizeof(udp_stat_data));
    
    /* Try to create the socket. */
    if ((udp_recv_socket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
    {
        log_printf(log_error, "Unable to open UDP socket.\n");
        ccode = -1;
        goto backout_base;
    }

    sock_opt = UDP_MAX_PACKET_SIZE;
    if ((setsockopt(udp_recv_socket, SOL_SOCKET, SO_RCVBUF, SOCK_OPT_CAST &sock_opt, sizeof(sock_opt))) < 0)
    {
        log_printf(log_error, "Unable to set input socket receive size.\n");
        ccode = -2;
        goto backout_socket;
    }
    
    /* Set up the address. */
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = INADDR_ANY;
    if (bind(udp_recv_socket, (struct sockaddr *) &addr, sizeof(addr)) < 0)
    {
        log_printf(log_error, "Unable to bind input socket.\n");
        ccode = -3;
        goto backout_socket;
    }

    /* Initialize thread to start listening. */
    stop_thread = 0;
    thread_stopped = 0;
    if (pthread_create(&(udpThread),
                       NULL,
                       (void * (*)(void *)) udpRecvStatThreadFunc,
                       (void *) 0) != 0)
    {
        log_printf(log_error, "Unable to start receive UDP thread.");
        ccode = -4;
        goto backout_socket;
    }
    
    // Success
    return(0);
    
 backout_socket:
    close(udp_recv_socket);

 backout_base:
    return(ccode);
}
 
////////////////////////////////////////////////////////////////////////////
// udpRecvStatData()
//      Extracts data from a received packet.
////////////////////////////////////////////////////////////////////////////
void udpRecvStatData(char *data)
{
    float tach, alt, speed;
    
    if (sscanf(data, "%f,%f,%f\n",
               &tach, &alt, &speed) == 3)
    {
        udp_stat_data.tach = tach;
        udp_stat_data.alt = alt;
        udp_stat_data.speed = speed;
    }

    //log_printf(log_quiet, "UDP Recv: tach=%f alt=%f speed=%f\n",
    //           tach, alt, speed);
}

////////////////////////////////////////////////////////////////////////////
// udpRecvStatThreadFunc()
//      Thread that receives aircraft updates from flight simulator.
////////////////////////////////////////////////////////////////////////////
void udpRecvStatThreadFunc()
{
    struct timeval timeout;
    fd_set fd;
    int totalCount = 0;
    int i, start;
    int ccode;
    int count;
    
    while ((!stopping) && (!stop_thread))
    {
        /* Check if socket is ready. */ 
        timeout.tv_sec = 0; 
        timeout.tv_usec = TIMEOUT_100MS; 
        FD_ZERO(&fd); 
        FD_SET(udp_recv_socket, &fd); 
        ccode = select(udp_recv_socket+1, &fd, 0, 0, &timeout); 
        
        if (ccode < 0)
        {
            // Socket died
            break;
        }
       
        if (ccode == 0)
        {
            /* Timeout */
            continue;
        }
        
        // IO is ready.  Read in a packet.
        if ((count = recvfrom(udp_recv_socket, RECV_BUFF_CAST (udp_packet + totalCount), UDP_MAX_PACKET_SIZE - totalCount, 0, 0, 0)) >= 0) 
        { 
            totalCount += count; 
        } 

        // Process any full lines of data we find in the buffer
        start = 0;
        for (i=0; i<totalCount; i++)
        {
            if (udp_packet[i] == '\n')
            {
                udp_packet[i] = 0;
                udpRecvStatData(&udp_packet[start]);
                start = i+1;
            }
        }

        // Remove processed data from buffer
        if (start != 0)
        {
            for (i=0; i<(totalCount-start); i++)
            {
                udp_packet[i] = udp_packet[start+i];
            }
        }
        totalCount = (totalCount-start);
    }
    thread_stopped = 1;
}

void udpRecvStatUninit()
{
    stop_thread = 1;
    while (!thread_stopped)
    {
        WAIT_SHORT;
    }
    if (udp_recv_socket) close(udp_recv_socket);
    udp_recv_socket = 0;
}

