/*============================================================================
 * udp_send_hmd_pos.cc
 *      Code to support updating flightgear flight simulator with position
 *      and orientation information from headset.  Only used for operation
 *      with the FlightGear flight simulator.  For X-Plane, we use a
 *      plugin and shared memory.
 *
 * Author: Jon W. Wallace, Lafayette College
 * Date:   May 2020
 *==========================================================================*/

#ifdef __WINDOWS__
#include <Windows.h>
#else
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netdb.h>
#include <unistd.h>
#endif
#include <signal.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "udp_send_hmd_pos.h"

#include "main.h"
#include "vr_helper.h"
#include "fifo.h"
#include "log.h"

static int udp_send_socket = 0;
static struct sockaddr_in udp_send_addr;
static char data_buffer[64];

static fifoDef *poseFifo;
static pthread_t udpThread;
static volatile int stop_thread;
static volatile int thread_stopped;

void udpSendHmdThreadFunc();

#define NUM_POSE_FRAMES 4

/*============================================================================
 * udpSendHmdInit()
 *      Initializes structures needed.
 *==========================================================================*/
int udpSendHmdInit(int port)
{
    int ccode;
    
    /* Try to create the socket. */
    if ((udp_send_socket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
    {
        log_printf(log_error, "Unable to open UDP socket.\n");
        ccode = -1;
        goto backout_base;
    }

    /* Set up the address of socket. */
    udp_send_addr.sin_family = AF_INET;
    udp_send_addr.sin_port = htons(port);
    udp_send_addr.sin_addr.s_addr = INADDR_ANY;

    /* Create FIFO to receive pose data in a separate thread. */
    if (fifoCreate(sizeof(vrhYprPoseDef), NUM_POSE_FRAMES, &poseFifo) != 0)
    {
        log_printf(log_error, "Unable to create pose FIFO.\n");
        ccode = -2;
        goto backout_socket;
    }

    /* Initialize thread to start streaming. */
    stop_thread = 0;
    thread_stopped = 0;
    if (pthread_create(&(udpThread),
                       NULL,
                       (void * (*)(void *)) udpSendHmdThreadFunc,
                       (void *) 0) != 0)
    {
        log_printf(log_error, "Unable to start UDP thread.");
        ccode = -3;
        goto backout_fifo;
    }

    return(0);

 backout_fifo:
    fifoDestroy(poseFifo);
    
 backout_socket:
    close(udp_send_socket);
    
 backout_base:
    return(ccode);
}

/*============================================================================
 * udpSendHmdData()
 *      Sends the HMD data
 * Inputs:
 *      pose.
 *      x, y, z         Coordinates of camera (in m)
 *      Y, P, R         Yaw, pitch, roll angles (in rad)
 *==========================================================================*/
void udpSendHmdData(vrhYprPoseDef *pose)
{
    sprintf(data_buffer,"%2.3f,%2.3f,%2.3f,%3.2f,%3.2f,%3.2f\n",
            pose->x, pose->y, pose->z,
            pose->Y*180.0/M_PI, pose->P*180.0/M_PI, pose->R*180.0/M_PI);
    sendto(udp_send_socket, data_buffer, strlen(data_buffer),
           0, (const struct sockaddr *) &udp_send_addr,
           sizeof(udp_send_addr));
}

/*============================================================================
 * udpSendHmdThreadFunc()
 *      Thread that sends pose updates via UDP to flight simulator.
 *==========================================================================*/
void udpSendHmdThreadFunc()
{
    vrhYprPoseDef *pose;
    
    while ((!stopping) && (!stop_thread))
    {
        // Try to get a pose update
        if ((pose = (vrhYprPoseDef *) fifoReadPtr(poseFifo)) == 0)
        {
            WAIT_SHORT;
            continue;
        }
        
        // Send via UDP
        udpSendHmdData(pose);
        fifoReadDone(poseFifo);
    }
    thread_stopped = 1;
}

/*============================================================================
 * udpSendPose
 *      Puts the pose in the FIFO to eventually be sent.
 *==========================================================================*/
void udpSendPose(vrhYprPoseDef *pose)
{
    fifoWrite(poseFifo, pose);
}

int udpSendHmdUninit()
{
    stop_thread = 1;
    while (!thread_stopped)
    {
        WAIT_SHORT;
    }
    fifoDestroy(poseFifo);
    if (udp_send_socket) close(udp_send_socket);
    udp_send_socket = 0;
    return(0);
}

