/*============================================================================
 * v4l_cap.cc
 *      Code to support capture through video4linux2.  Maybe can add other
 *      interfaces later.
 *
 * Author: Jon W. Wallace, Lafayette College
 * Date:   May 2020
 *==========================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <fcntl.h>              /* low-level i/o */
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <linux/videodev2.h>

#include "v4l_cap.h"
#include "log.h"

#define CLEAR(x) memset(&(x), 0, sizeof(x))

#define V4L_FLAGS_JPEG_DECODING         0x1
#define V4L_CAP_NUM_BUFFERS             4

// Buffers used for processing
struct v4l_cap_buffer {
    void   *start;
    size_t  length;
};

typedef struct
{
    // Input parameters
    int cap_w;
    int cap_h;
    int cap_rate;
    const char *dev_name;
    int n_v4l2_buffers;

    // Internal state variables
    int fd;     // File descriptor for device

    // V4L2 variables
    struct v4l2_capability cap;
    struct v4l2_format fmt;
    struct v4l2_requestbuffers req;

    // Buffers used for capturing
    struct v4l_cap_buffer *buffers;
} v4lDef;

v4lDef v4l_data;

/*----------------------------------------------------------------------------
 * xioctl
 *      Helper function that continually does an ioctl as long as the specified
 *      error is received.
 *--------------------------------------------------------------------------*/
static int xioctl(int fh, int request, void *arg)
{
    int r;

    do {
        r = ioctl(fh, request, arg);
    } while (-1 == r && EINTR == errno);

    return r;
}

/*----------------------------------------------------------------------------
 * v4l_cap_reinit()
 *      Sometimes the device times out.  Try to restart it.
 *--------------------------------------------------------------------------*/
void v4l_cap_reinit()
{
    enum v4l2_buf_type btype;
    log_printf(log_error, "Restarting V4L device...\n");
        
    btype = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (xioctl(v4l_data.fd, VIDIOC_STREAMOFF, &btype) < 0)
    {
        log_printf(log_error, "Error calling VIDIOC_STREAMOFF.\n");
    }

    if (xioctl(v4l_data.fd, VIDIOC_STREAMON, &btype) < 0)
    {
        log_printf(log_error, "Error calling VIDIOC_STREAMON.\n");
    }
}

/*----------------------------------------------------------------------------
 * cap_get_frame
 *      Reads a single video frame into the provided buffer.
 *--------------------------------------------------------------------------*/
int v4l_cap_get_frame(void *output_buffer, int buffer_size, int *bytes_used_ret)
{
    fd_set fds;
    struct timeval tv;
    int r;
    int ccode;
    struct v4l2_buffer buf;

    /* Repeat until we see data, error, or timeout. */
    for ( ; ; )
    {
        FD_ZERO(&fds);
        FD_SET(v4l_data.fd, &fds);
    
        /* Timeout. */
        tv.tv_sec = 0;
        tv.tv_usec = 100*1000;
        
        r = select(v4l_data.fd + 1, &fds, NULL, NULL, &tv);

        if (r < 0)
        {
            if (errno == EINTR) continue;
            log_printf(log_error, "v4l_cap: error on select().\n");
            ccode = -1;
            goto backout_base;
        }

        if (r==0)
        {
            log_printf(log_error, "v4l_cap: select timeout.\n");
            v4l_cap_reinit();
            ccode = -2;
            goto backout_base;
        }

        /* Read in the frame. */
        CLEAR(buf);
        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;

        /* Try to dequeue a frame. */
        if (xioctl(v4l_data.fd, VIDIOC_DQBUF, &buf) < 0)
        {
            switch (errno)
            {
                case EAGAIN:
                    /* In this case do select() again. */
                    break;
                    
                case EIO:
                    /* Could ignore EIO, see spec. */
                    /* fall through */
                default:
                    log_printf(log_error, "v4l_cap: Error with VIDIOC_DQBUF.\n");
                    ccode = -3;
                    goto backout_base;
            }
        }

        if (buf.index >= v4l_data.n_v4l2_buffers)
        {
            log_printf(log_error, "v4l_cap: Invalid buffer index encountered.\n");
            ccode = -4;
            goto backout_base;
        }

        // Check size of buffer
        if (buf.bytesused > buffer_size)
        {
            log_printf(log_error, "v4l_cap: Frame too large for provided buffer.\n");
            ccode = -5;
            goto backout_base;
        }

        // Copy buffer to output (if desired).
        if (output_buffer) memcpy(output_buffer, v4l_data.buffers[buf.index].start, buf.bytesused);
        if (bytes_used_ret) *bytes_used_ret = buf.bytesused;

        // Queue the buffer again.
        if (xioctl(v4l_data.fd, VIDIOC_QBUF, &buf) < 0)
        {
            log_printf(log_error, "Error with VIDIOC_QBUF.\n");
            ccode = -6;
            goto backout_base;
        }

        break;
    }

    ccode = 0;
    return(ccode);

 backout_base:
    return(ccode);
}

/*----------------------------------------------------------------------------
 * v4l_cap_init
 *      Initializes the v4l2 device using global parameters that are set.
 *--------------------------------------------------------------------------*/
int v4l_cap_init(const char *dev_name, int w, int h, int rate, uint32_t flags)
{
    int ccode;
    struct v4l2_buffer buf;
    enum v4l2_buf_type btype;

    // Initialize the capture data structure
    CLEAR(v4l_data);
    v4l_data.dev_name = dev_name;
    v4l_data.cap_w = w;
    v4l_data.cap_h = h;
    v4l_data.cap_rate = rate;
    v4l_data.n_v4l2_buffers = V4L_CAP_NUM_BUFFERS;
    
    /* Open the video device */
    if ((v4l_data.fd = open(dev_name, O_RDWR | O_NONBLOCK, 0)) < 0)
    {
        log_printf(log_error, "Unable to open device %s.\n", v4l_data.dev_name);
        ccode = -1;
        goto backout_base;
    }

    // Check that it is a V4L device
    if (xioctl(v4l_data.fd, VIDIOC_QUERYCAP, &v4l_data.cap) < 0) {
        log_printf(log_error, "Unable to query V4L2 device for capabilities.\n");
        ccode = -2;
        goto backout_device;
    }

    if (!(v4l_data.cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
        log_printf(log_error, "%s is not a V4L2 capture_device.\n", v4l_data.dev_name);
        ccode = -3;
        goto backout_device;
    }

    if (!(v4l_data.cap.capabilities & V4L2_CAP_STREAMING)) {
        log_printf(log_error, "%s does not support streaming i/o.\n", v4l_data.dev_name);
        ccode = -4;
        goto backout_device;
    }

    CLEAR(v4l_data.fmt);
    v4l_data.fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    v4l_data.fmt.fmt.pix.width       = v4l_data.cap_w;
    v4l_data.fmt.fmt.pix.height      = v4l_data.cap_h;
    if (flags & V4L_FLAGS_JPEG_DECODING)
    {
        v4l_data.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_MJPEG;
    }
    else
    {
        v4l_data.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
    }
    v4l_data.fmt.fmt.pix.field       = V4L2_FIELD_ANY;
    if (xioctl(v4l_data.fd, VIDIOC_S_FMT, &v4l_data.fmt) < 0)
    {
        log_printf(log_error, "Error using VIDIOC_S_FMT.\n");
        ccode = -5;
        goto backout_device;
    }

    log_printf(log_quiet, "V4L2 Data format out: %dx%d\n",
               v4l_data.fmt.fmt.pix.width,
               v4l_data.fmt.fmt.pix.height);
    
    /* Setup capture buffers. */
    CLEAR(v4l_data.req);
    v4l_data.req.count = v4l_data.n_v4l2_buffers;
    v4l_data.req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    v4l_data.req.memory = V4L2_MEMORY_MMAP;
    if (xioctl(v4l_data.fd, VIDIOC_REQBUFS, &v4l_data.req) < 0)
    {
        log_printf(log_error, "%s does not support mmap.\n", v4l_data.dev_name);
        ccode = -6;
        goto backout_device;
    }

    if (v4l_data.n_v4l2_buffers != v4l_data.req.count)
    {
        log_printf(log_error, "Warning: requested %d capture buffers, got only %d.\n", v4l_data.n_v4l2_buffers, v4l_data.req.count);
        v4l_data.n_v4l2_buffers = v4l_data.req.count;
    }
    
    if (v4l_data.req.count < 2)
    {
        log_printf(log_error, "Insufficient buffer memory on %s\n", v4l_data.dev_name);
        ccode = -7;
        goto backout_device;
    }

    if ((v4l_data.buffers = (v4l_cap_buffer*) calloc(v4l_data.req.count, sizeof(*v4l_data.buffers))) == 0)
    {
        log_printf(log_error, "Unable to allocate capture buffers.\n");
        ccode = -8;
        goto backout_device;
    }

    for (int n_buffers=0; n_buffers < v4l_data.n_v4l2_buffers; ++n_buffers) {

        CLEAR(buf);

        buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory      = V4L2_MEMORY_MMAP;
        buf.index       = n_buffers;

        if (xioctl(v4l_data.fd, VIDIOC_QUERYBUF, &buf) < 0)
        {
            log_printf(log_error, "Error calling VIDIOC_QUERYBUF.\n");
            ccode = -9;
            goto backout_buffers;
        }

        v4l_data.buffers[n_buffers].length = buf.length;
        v4l_data.buffers[n_buffers].start =
            mmap(NULL /* start anywhere */,
                 buf.length,
                 PROT_READ | PROT_WRITE /* required */,
                 MAP_SHARED /* recommended */,
                 v4l_data.fd, buf.m.offset);

        if (v4l_data.buffers[n_buffers].start == MAP_FAILED)
        {
            log_printf(log_error, "Error calling mmap.\n");
            ccode = -10;
            goto backout_buffers;
        }
    }

    /* Queue the buffers. */
    for (int i = 0; i < v4l_data.n_v4l2_buffers; ++i)
    {
        CLEAR(buf);
        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        buf.index = i;
        if (xioctl(v4l_data.fd, VIDIOC_QBUF, &buf) < 0)
        {
            log_printf(log_error, "Error calling VIDIOC_QBUF.\n");
            ccode = -11;
            goto backout_buffers;
        }
    }

    /* Start capturing */
    btype = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (xioctl(v4l_data.fd, VIDIOC_STREAMON, &btype) < 0)
    {
        log_printf(log_error, "Error calling VIDIOC_STREAMON.\n");
        ccode = -12;
        goto backout_buffers;
    }

    /* Success */
    return(0);

 backout_buffers:
    free(v4l_data.buffers);
    
 backout_device:
    close(v4l_data.fd);

 backout_base:
    CLEAR(v4l_data);
    return(ccode);
}

/*----------------------------------------------------------------------------
 * v4l_cap_uninit()
 *      Uninitializes capture.
 *--------------------------------------------------------------------------*/
void v4l_cap_uninit()
{
    enum v4l2_buf_type btype;
    struct v4l2_buffer buf;

    btype = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (xioctl(v4l_data.fd, VIDIOC_STREAMOFF, &btype) < 0)
    {
        log_printf(log_error, "Error calling VIDIOC_STREAMON.\n");
    }

    /* Dequeue the buffers. */
    /* Causes errors here.  Need to clean up something else first? */
    /*
    for (int i = 0; i < n_v4l2_buffers; ++i)
    {
        CLEAR(buf);
        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;
        buf.index = i;
        if (xioctl(fd, VIDIOC_DQBUF, &buf) < 0)
        {
            log_printf(log_error, "Error calling VIDIOC_DQBUF.\n");
        }
    }
    */
    
    close(v4l_data.fd);
    free(v4l_data.buffers);
}

