////////////////////////////////////////////////////////////////////////////
// vr_helper.cc
//      Some useful functions for simplifying work with VR headset.
////////////////////////////////////////////////////////////////////////////

#include <openvr/openvr.h>
#include <string.h>

#include <GL/glew.h>

using namespace vr;

#include "log.h"

#include "Vectors.h"
#include "Matrices.h"
#include "vr_helper.h"
#include "c172_win1.h"
#include "glh.h"
#include "ar_parm.h"
#include "main.h"
#include "cudaRender.h"

#define SQ(x) ((x)*(x))
#define MAX(x, y) (((x) > (y)) ? (x) : (y))

static IVRSystem *pVr=0;

// Local coordinate system when pilot is sitting
// in seat and looking straight forward, in terms
// of the global coordinate system.
struct
{
    Vector3 x;
    Vector3 y;
    Vector3 z;
    Vector3 origin;
} pilot_home;

const float pilot_fov_z_near = 0.01;
const float pilot_fov_z_far = 5.0;

// Headset View/Projection matrices
Matrix4 M_eye_pose_l;
Matrix4 M_eye_pose_r;
Matrix4 M_proj_l;
Matrix4 M_proj_r;
Matrix4 M_pose;

// Offset of pilot head home position to FS airplane model
// center
float vrhPilotOffset[3];

// Constants for FG / X-Plane model origin offset to pilot-head-home
const float vrhPilotOffsetFg[3] = {-0.25, 0.2, -0.2};
const float vrhPilotOffsetXp[3] = {-0.25, 0.73, -0.4};
//const float vrhPilotOffsetXp[3] = {-0.31, 0.61, -0.17};

// Variables used for calibration mode
const char *fname_cal = "cal_out.txt";
FILE *f_cal = 0;

// Eventually, we should move all VR functions from render() to this file.
// For now, allow access to poses through global.
extern TrackedDevicePose_t trackedDevicePose[k_unMaxTrackedDeviceCount];

// Data for airplane model, used in geometric mode.
float win_tris_vertex_map[N_poly_win_tris][N_vert_win_tris][3];

void get_vectors_from_orient_matrix(HmdMatrix34_t &matrix, Vector3 &x, Vector3 &y, Vector3 &z, Vector3 &o)
{
    x.set(matrix.m[0][0], matrix.m[1][0], matrix.m[2][0]);
    y.set(matrix.m[0][1], matrix.m[1][1], matrix.m[2][1]);
    z.set(matrix.m[0][2], matrix.m[1][2], matrix.m[2][2]);
    o.set(matrix.m[0][3], matrix.m[1][3], matrix.m[2][3]);
}

void printMatrix(log_level_enum level, HmdMatrix34_t &matrix)
{
    int i, j;

    for (i=0; i<3; i++)
    {
        for (j=0; j<4; j++)
        {
            log_printf(level, "%+4.2f  ", matrix.m[i][j]);
        }
        log_printf(level, "\n");
    }
    log_printf(level, "----\n");
}

Matrix4 getEyeProjMatrix( vr::Hmd_Eye nEye )
{
    HmdMatrix44_t mat = pVr->GetProjectionMatrix( nEye, pilot_fov_z_near, pilot_fov_z_far);

    return Matrix4(
                   mat.m[0][0], mat.m[1][0], mat.m[2][0], mat.m[3][0],
                   mat.m[0][1], mat.m[1][1], mat.m[2][1], mat.m[3][1], 
                   mat.m[0][2], mat.m[1][2], mat.m[2][2], mat.m[3][2], 
                   mat.m[0][3], mat.m[1][3], mat.m[2][3], mat.m[3][3]
                   );
}

Matrix4 getEyePoseMatrix( vr::Hmd_Eye nEye )
{
	HmdMatrix34_t matEye = pVr->GetEyeToHeadTransform( nEye );
	Matrix4 matrixObj(
		matEye.m[0][0], matEye.m[1][0], matEye.m[2][0], 0.0, 
		matEye.m[0][1], matEye.m[1][1], matEye.m[2][1], 0.0,
		matEye.m[0][2], matEye.m[1][2], matEye.m[2][2], 0.0,
		matEye.m[0][3], matEye.m[1][3], matEye.m[2][3], 1.0f
		);

	return matrixObj.invert();
}

Matrix4 convertSteamVRMatrixToMatrix4( const HmdMatrix34_t &matPose )
{
	Matrix4 matrixObj(
		matPose.m[0][0], matPose.m[1][0], matPose.m[2][0], 0.0,
		matPose.m[0][1], matPose.m[1][1], matPose.m[2][1], 0.0,
		matPose.m[0][2], matPose.m[1][2], matPose.m[2][2], 0.0,
		matPose.m[0][3], matPose.m[1][3], matPose.m[2][3], 1.0f
		);
	return matrixObj;
}

////////////////////////////////////////////////////////////////////////////
// findDot
//      finds the brightest green dot in the camera frame.
////////////////////////////////////////////////////////////////////////////
int findDot(uint32_t *frame_data, int *x, int *y, float *amp, float req_amp, int width, int height)
{
    int i, j;
    float x2, avg, val1;
    float std;

    int max_x, max_y;
    float max_amp;
    float R, G, B;
    
    x2 = 0.0;
    avg = 0.0;
    max_x = -1;
    max_y = -1;
    max_amp = 0.0;
    
    for (j=0; j<height; j++)
    {
        for (i=0; i<width; i++)
        {
            R = ((frame_data[i + width*j]) & 0x000000ff)/255.0;
            G = ((frame_data[i + width*j] & 0x0000ff00) >> 8)/255.0;
            B = ((frame_data[i + width*j] & 0x00ff0000) >> 16)/255.0;
            val1 = (G - MAX(R,B));
            if (val1 < 0.0) val1 = 0.0;
            x2 += val1*val1;
            avg += val1;

            if (val1 > max_amp)
            {
                max_amp = val1;
                max_x = i;
                max_y = j;
            }
        }
    }
    x2 /= (float)(height*width);
    avg /= (float)(height*width);

    std = sqrtf(x2 - avg*avg);

    if ((max_amp - avg)/std > req_amp)
    {
        *x = max_x;
        *y = max_y;
        *amp = (max_amp - avg)/std;
        return(1);
    }
    else
    {
        *x = 0;
        *y = 0;
        *amp = 0;
        return(0);
    }
}

////////////////////////////////////////////////////////////////////////////
// draw_dot
//      Puts a dot of a specified color in the framebuffer.
////////////////////////////////////////////////////////////////////////////
void draw_dot(glhFbDef *fb, float x, float y, float R, float G, float B)
{
    glBindFramebuffer(GL_FRAMEBUFFER, fb->fbo_id);
    glViewport(0, 0, fb->width, fb->height);
    
    // Draw on the render texture
    glPushMatrix();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0,   // left
            1.0,   // right
            0.0,   // bottom
            1.0,   // top
            1.0,  // near
            -1.0);  // far

    glPointSize(3.0f);
    glColor3f(R, G, B);

    glBegin(GL_POINTS);
    glVertex2d(x, y);
    glEnd();
    
    glPopMatrix();

    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

////////////////////////////////////////////////////////////////////////////
// Public functions
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// vrhCameraCalStart
//      Outputs camera calibration data to a file.  Detects a green      
//      LED on a controller and sends both its detected position
//      and VR position to a file.
////////////////////////////////////////////////////////////////////////////
int vrhCameraCalStart()
{
    int ccode;
    
    // Open calibration output file if in calibration mode
    if ((f_cal = fopen(fname_cal, "wb")) == 0)
    {
        log_printf(log_error, "Unable to open calibration file.\n");
        ccode = -1;
        goto backout_base;
    }

    return(0);
    
 backout_base:
    return(ccode);
}

////////////////////////////////////////////////////////////////////////////
// vrhCameraCal
//      Outputs camera calibration data to a file.  Detects a green      
//      LED on a controller and sends both its detected position
//      and VR position to a file.  Shows the detected LED position from
//      camera data with a blue dot if fb_l and fb_r are provided.
////////////////////////////////////////////////////////////////////////////
void vrhCameraCalSample(uint32_t *cap_data, uint32_t *cap_data_gpu, int cap_w, int cap_h, glhFbDef *fb_l, glhFbDef *fb_r)
{
    Vector3 cont_x, cont_y, cont_z, cont_pos;
    Vector3 cont_xp, cont_yp, cont_zp, cont_posp;
    Vector3 led_pos;
    
    Vector3 g;
    Vector3 dv;
    float z, p1, p2;
    int i;
    Vector4 gp, pt;

    Matrix4 M;
    
    int left_dot_x, left_dot_y;
    int right_dot_x, right_dot_y;
    float left_dot_amp, right_dot_amp;
    float x_view, y_view;

    float rend_dot_xl, rend_dot_yl;
    float rend_dot_xr, rend_dot_yr;
    
    int found_controller;

    int ii;

    if (!f_cal) return;

    if (mainOptions.gpu)
    {
        // Use CUDA accelerated method.  Here we don't check for a threshold, like with the CPU version.
        crFindDot(&left_dot_x, &left_dot_y, &right_dot_x, &right_dot_y, cap_data_gpu);
    }
    else
    {
        // Detect the LED in the left and right frames.
        // The brightest point needs to be a certain number of
        // standard deviations above the average.
        if (findDot(cap_data + cap_w*cap_h, &left_dot_x, &left_dot_y, &left_dot_amp, 1.5, cap_w, cap_h) == 0)
        {
            return;
        }
        
        if (findDot(cap_data, &right_dot_x, &right_dot_y, &right_dot_amp, 1.5, cap_w, cap_h) == 0)
        {
            return;
        }
    }
    
    // Get the position of the LED from the controller
    found_controller = 0;
    for (ii=0; ii<k_unMaxTrackedDeviceCount; ii++)
    {
        if ( trackedDevicePose[ii].bPoseIsValid )
        {
            if (pVr->GetTrackedDeviceClass(ii) == TrackedDeviceClass_Controller)
            {
                found_controller=1;
                break;
            }
        }
    }
    if (!found_controller) return;

    // Get components of controller position
    get_vectors_from_orient_matrix(trackedDevicePose[ii].mDeviceToAbsoluteTracking, cont_x, cont_y, cont_z, cont_pos);

    // Get LED position.  These hard-coded values were obtained by finding the offset in
    // an HTC-provided .OBJ of the controller.  
    led_pos = cont_pos +0.0963*cont_z + 0.0029*cont_y;

    // Make into a 4-element vector
    gp.set(led_pos[0], led_pos[1], led_pos[2], 1.0);
    
    // Left Eye
    M = M_proj_l*M_eye_pose_l*M_pose;
    pt = M*gp;

    fprintf(f_cal, "LV %.6f %.6f\n", pt[0]/pt[3], pt[1]/pt[3]);
    fprintf(f_cal, "LC %d %d %.2f\n", left_dot_x, left_dot_y, left_dot_amp);

    // Right Eye
    M = M_proj_r*M_eye_pose_r*M_pose;
    pt = M*gp;

    fprintf(f_cal, "RV %.6f %.6f\n", pt[0]/pt[3], pt[1]/pt[3]);
    fprintf(f_cal, "RC %d %d %.2f\n", right_dot_x, right_dot_y, left_dot_amp);

    if ((fb_l) && (fb_r))
    {
        // Draw blue detected dots on L/R frames to provide feedback that this is working.
        ar_parm_map1(left_dot_x, left_dot_y, &rend_dot_xl, &rend_dot_yl, M_cap_l);
        ar_parm_map1(right_dot_x, right_dot_y, &rend_dot_xr, &rend_dot_yr, M_cap_r);

        // Draw detected position in camera frame in BLUE
        // Left
        x_view = (rend_dot_xl + 1.0)/2.0;
        y_view = (rend_dot_yl + 1.0)/2.0;
        draw_dot(fb_l, x_view, y_view, 0.1, 0.1, 1.0);
        // Right
        x_view = (rend_dot_xr + 1.0)/2.0;
        y_view = (rend_dot_yr + 1.0)/2.0;
        draw_dot(fb_r, x_view, y_view, 0.1, 0.1, 1.0);
    }
}

////////////////////////////////////////////////////////////////////////////
// vrhCameraCalEnd()
//      Ends the calibration phase.
////////////////////////////////////////////////////////////////////////////
void vrhCameraCalEnd()
{
    // Open calibration output file if in calibration mode
    fclose(f_cal);
    f_cal = 0;
}

////////////////////////////////////////////////////////////////////////////
// vrhDrawMarker
//      Draws a marker where the controller LED should be.
////////////////////////////////////////////////////////////////////////////
void vrhDrawMarker(glhFbDef *fb_l, glhFbDef *fb_r, uint32_t *cam_data, uint32_t *cam_data_gpu, uint32_t cap_w, uint32_t cap_h)
{
    Vector3 cont_x, cont_y, cont_z, cont_pos;
    Vector3 cont_xp, cont_yp, cont_zp, cont_posp;
    Vector3 led_pos;
    
    Vector3 g;
    Vector3 dv;
    float z, p1, p2;
    int i;
    Vector4 gp, pt;

    Matrix4 M;
    
    int left_dot_x, left_dot_y;
    int right_dot_x, right_dot_y;
    float left_dot_amp, right_dot_amp;
    float rend_dot_xl, rend_dot_yl;
    float rend_dot_xr, rend_dot_yr;
    
    int found_controller;

    int ii;

    float x_view, y_view;

    if (mainOptions.gpu)
    {
        crFindDot(&left_dot_x, &left_dot_y, &right_dot_x, &right_dot_y, cam_data_gpu);
    }
    else
    {
        if (cam_data)
        {
            if (findDot(cam_data, &left_dot_x, &left_dot_y, &left_dot_amp, 1.5, cap_w, cap_h) == 0)
            {
                left_dot_x = 0; left_dot_y = 0;
            }
            
            if (findDot(cam_data + cap_w*cap_h, &right_dot_x, &right_dot_y, &right_dot_amp, 1.5, cap_w, cap_h) == 0)
            {
                right_dot_x = 0; right_dot_y = 0;
            }
        }
    }

    ar_parm_map1(left_dot_x, left_dot_y, &rend_dot_xl, &rend_dot_yl, M_cap_l);
    ar_parm_map1(right_dot_x, right_dot_y, &rend_dot_xr, &rend_dot_yr, M_cap_r);
    
    found_controller = 0;
    for (ii=0; ii<k_unMaxTrackedDeviceCount; ii++)
    {
        if ( trackedDevicePose[ii].bPoseIsValid )
        {
            if (pVr->GetTrackedDeviceClass(ii) == TrackedDeviceClass_Controller)
            {
                found_controller=1;
                break;
            }
        }
    }
    if (!found_controller) return;
    
    // Get components of controller position
    get_vectors_from_orient_matrix(trackedDevicePose[ii].mDeviceToAbsoluteTracking, cont_x, cont_y, cont_z, cont_pos);

    // Get LED position.  These hard-coded values were obtained by finding the offset in
    // an HTC-provided .OBJ of the controller.  
    led_pos = cont_pos +0.0963*cont_z + 0.0029*cont_y;

    // Make into a 4-element vector
    gp.set(led_pos[0], led_pos[1], led_pos[2], 1.0);

    // Updates M_pose according to headset position.
    vrhUpdatePose();
    
    // Left Eye
    M = M_proj_l*M_eye_pose_l*M_pose;
    pt = M*gp;

    // Convert to our (0,1) range
    x_view = (pt[0]/pt[3] + 1.0)/2.0;
    y_view = (pt[1]/pt[3] + 1.0)/2.0;

    // Draw VR computed green-LED position in RED
    draw_dot(fb_l, x_view, y_view, 1.0, 0.1, 0.1);
    // Draw detected position in camera frame in BLUE
    x_view = (rend_dot_xl + 1.0)/2.0;
    y_view = (rend_dot_yl + 1.0)/2.0;
    draw_dot(fb_l, x_view, y_view, 0.1, 0.1, 1.0);
    
    // Right Eye
    M = M_proj_r*M_eye_pose_r*M_pose;
    pt = M*gp;

    // Convert to our (0,1) range
    x_view = (pt[0]/pt[3] + 1.0)/2.0;
    y_view = (pt[1]/pt[3] + 1.0)/2.0;

    // Draw VR computed green-LED position in RED
    draw_dot(fb_r, x_view, y_view, 1.0, 0.1, 0.1);
    // Draw detected position in camera frame in BLUE
    x_view = (rend_dot_xr + 1.0)/2.0;
    y_view = (rend_dot_yr + 1.0)/2.0;
    draw_dot(fb_r, x_view, y_view, 0.1, 0.1, 1.0);

}

////////////////////////////////////////////////////////////////////////////
// vrhSetPilotOrigin()
//      Make the current position and pose the pilot seat local coordinate
//      system.
////////////////////////////////////////////////////////////////////////////
void vrhSetPilotOrigin(HmdMatrix34_t &matrix)
{
    Vector3 point_map;
    int i, j;
    
    // Copy the headset position
    log_printf(log_info, "Updating cockpit origin:\n");
    printMatrix(log_info, matrix);

    // Extract individual vectors and origin
    get_vectors_from_orient_matrix(matrix, pilot_home.x, pilot_home.y, pilot_home.z, pilot_home.origin);

    // Update model for the cockpit windows.  We take the win_tris_vertex[], which are in local
    // coordinates relative to the pilot home position, and convert them to the VR global
    // coordinates, stored in win_tris_vertex_map[].
    for (j=0; j<N_poly_win_tris; j++)
    {
        for (i=0; i<N_vert_win_tris; i++)
        {
            point_map = pilot_home.origin
                + pilot_home.x*win_tris_vertex[j][i][0]
                + pilot_home.y*win_tris_vertex[j][i][1]
                + pilot_home.z*win_tris_vertex[j][i][2];
            win_tris_vertex_map[j][i][0] = point_map[0];
            win_tris_vertex_map[j][i][1] = point_map[1];
            win_tris_vertex_map[j][i][2] = point_map[2];
        }
    }
}

////////////////////////////////////////////////////////////////////////////
// vrhUpdatePose
//      Computes the pose (view) matrix used for rendering, based on the
//      head position.
////////////////////////////////////////////////////////////////////////////
int vrhUpdatePose()
{
    if (trackedDevicePose[k_unTrackedDeviceIndex_Hmd].bPoseIsValid)
    {
        M_pose = convertSteamVRMatrixToMatrix4(trackedDevicePose[k_unTrackedDeviceIndex_Hmd].mDeviceToAbsoluteTracking);
        M_pose.invert();
    }
}

////////////////////////////////////////////////////////////////////////////
// vrhComputeHmdYprPose
//      Uses the current HMD position and pilot home coordinate system
//      to compute Y, P, R angles and translation of the head.  These
//      can then be sent to the flight simulator software to update
//      view.
// Note: Eventually we need to see if there is a better way than using
//      YPR angles.  When you look straight up, there is an ambiguity.
////////////////////////////////////////////////////////////////////////////
void vrhComputeHmdYprPose(vrhYprPoseDef *pose)
{
    Vector3 head_x, head_y, head_z, head_pos;
    Vector3 fg_hx, fg_hy, fg_hz;
    Vector3 fg_px, fg_py, fg_pz;

    Vector3 d_pos;
    float x, y, z, fg_x, fg_y, fg_z;
    Vector3 xh, yh, zh, xh1, yh1, zh1;
    Vector3 yh_, zh_;
    float Y, P, R;
    
    // Extract individual vectors and origin
    get_vectors_from_orient_matrix(trackedDevicePose[0].mDeviceToAbsoluteTracking, head_x, head_y, head_z, head_pos);
    
    d_pos = head_pos - pilot_home.origin;
    x = pilot_home.x.dot(d_pos);
    y = pilot_home.y.dot(d_pos);
    z = pilot_home.z.dot(d_pos);

    // In FlightGear, offset location (0,0,0) is the center of the airplane model.
    // In our pilot home coordinate system (0,0,0) is the pilot head home position.
    // So, if we are at (0,0,0) in the pilot head coordinate system, we need to adjust
    // this to where the pilot's head home position is in the airplane model.  Figuring this
    // offset out requires inspection the FG airplane mode.
    // C172 offsets
    x += vrhPilotOffset[0];
    y += vrhPilotOffset[1];
    z += vrhPilotOffset[2];
    
    // Get local head xh, yh, zh vectors in pilot coordinate frame.
    xh[0] = head_x.dot(pilot_home.x); xh[1] = head_x.dot(pilot_home.y); xh[2] = head_x.dot(pilot_home.z);
    yh[0] = head_y.dot(pilot_home.x); yh[1] = head_y.dot(pilot_home.y); yh[2] = head_y.dot(pilot_home.z);
    zh[0] = head_z.dot(pilot_home.x); zh[1] = head_z.dot(pilot_home.y); zh[2] = head_z.dot(pilot_home.z);

    // Convert vectors from OpenGL frame convention
    // to aviation coordinates
    xh1[0] = zh[2];  xh1[1] = zh[0];  xh1[2] = -zh[1];
    yh1[0] = xh[2];  yh1[1] = xh[0];  yh1[2] = -xh[1];
    zh1[0] = -yh[2]; zh1[1] = -yh[0]; zh1[2] = yh[1];
    
    // Convert orientation vectors to YPR
    Y = atan2(xh1[1], xh1[0]);
    P = atan2(xh1[2], sqrtf(SQ(xh1[0]) + SQ(xh1[1])));

    // Get "zero roll" y and z vectors
    yh_[0] = -xh1[1];
    yh_[1] = xh1[0];
    yh_[2] = 0.0;
    zh_ = xh1.cross(yh_);
    zh_.normalize();

    // Get roll
    R = atan2(yh1.dot(zh_), yh1.dot(yh_));
    
    // Return values
    pose->x = x;
    pose->y = y;
    pose->z = z;
    pose->Y = Y;
    pose->P = P;
    pose->R = R;
}

int vrhInit(IVRSystem *pVr_in)
{
    pVr = pVr_in;

    // Set global pilot position
    pilot_home.x = Vector3(1.0, 0.0, 0.0);
    pilot_home.y = Vector3(0.0, 1.0, 0.0);
    pilot_home.z = Vector3(0.0, 0, 1.0);
    pilot_home.origin = Vector3(0.0, 0.0, 0.0);

    // Get eye pose and projection matrices from OpenVr
    M_eye_pose_l = getEyePoseMatrix(Eye_Left);
    M_eye_pose_r = getEyePoseMatrix(Eye_Right);
    M_proj_l = getEyeProjMatrix(Eye_Left);
    M_proj_r = getEyeProjMatrix(Eye_Right);

    // Copy offset from model center to pilot head home
    if (mainOptions.xplane)
    {
        memcpy(vrhPilotOffset, vrhPilotOffsetXp, sizeof(vrhPilotOffset));
    }
    else
    {
        memcpy(vrhPilotOffset, vrhPilotOffsetFg, sizeof(vrhPilotOffset));
    }
    return(0);
}

void vrhUninit()
{
    if (f_cal) vrhCameraCalEnd();
}

////////////////////////////////////////////////////////////////////////////
// vrhDrawSceneryCockpit
//      Code to draw background scenery from the flight simulator onto
//      winow regions according to a geometric model of the airplane.
// Inputs:
//      fb_dest                 Destination framebuffer.
//      fb_src                  Texture to use as source (background scenery)
//                              for window regions.
//      left_eye                Flag indicating which eye (1=left, 0=right)                        
////////////////////////////////////////////////////////////////////////////
void vrhDrawSceneryCockpit(glhFbDef *fb_dest, glhFbDef *fb_src, int left_eye)
{
    float window_mapped_corners[3][2];
    float *g;
    int i, j;
    Vector4 gp, pt;
    int is_visible;

    Matrix4 M;

    glBindFramebuffer(GL_FRAMEBUFFER, fb_dest->fbo_id);
    glViewport(0, 0, fb_dest->width, fb_dest->height);
    
    // Draw on the render texture
    glPushMatrix();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0,   // left
            1.0,   // right
            0.0,   // bottom
            1.0,   // top
            1.0,  // near
            -1.0);  // far

    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glBindTexture(GL_TEXTURE_2D, fb_src->tex_id);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
                
    // Get the view/projection matrix for the appropriate eye
    if (left_eye)
    {
        M = M_proj_l*M_eye_pose_l*M_pose;
    }
    else
    {
        M = M_proj_r*M_eye_pose_r*M_pose;
    }
    
    // Map the cockpit window geometry points.
    glBegin(GL_TRIANGLES);
    for (j=0; j<N_poly_win_tris; j++)
    {
        is_visible = 1;
        for (i=0; i<3; i++)
        {
            // Get depth to point
            g = win_tris_vertex_map[j][i];
            // Make into a 4-element vector
            gp.set(g[0], g[1], g[2], 1.0);
            // Do projections
            // pt = M_proj*(M_mod*gp);
            pt = M*gp;
            window_mapped_corners[i][0] = (pt[0]/pt[3] + 1.0)/2.0;
            window_mapped_corners[i][1] = (pt[1]/pt[3] + 1.0)/2.0;

            // Clipping should be handled automatically if corners are out of range, but if the
            // z coordinate goes behind viewer, there are problems, so avoid that.
            if (pt[2]/pt[3] > 1.0)
            {
                is_visible = 0;
            }
        }

        // Draw the image
        if (is_visible)
        {
            if (fb_src)
            {
                for (i=0; i<3; i++)
                {
                    glTexCoord2f(window_mapped_corners[i][0], window_mapped_corners[i][1]);
                    glVertex2d(window_mapped_corners[i][0], window_mapped_corners[i][1]);
                }
            }
        }
    }
    glEnd();
    
    // Draw triangle outlines (mainly for debug)
    glDisable(GL_TEXTURE_2D);
    glColor3f(1, 0.1, 0.1);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

    // Map points
    glBegin(GL_TRIANGLES);
    for (j=0; j<N_poly_win_tris; j++)
    {
        is_visible = 1;
        for (i=0; i<3; i++)
        {
            // Get depth to point
            g = win_tris_vertex_map[j][i];
            // Make into a 4-element vector
            gp.set(g[0], g[1], g[2], 1.0);
            // Do projections
            // pt = M_proj*(M_mod*gp);
            pt = M*gp;
            window_mapped_corners[i][0] = (pt[0]/pt[3] + 1.0)/2.0;
            window_mapped_corners[i][1] = (pt[1]/pt[3] + 1.0)/2.0;

            // Clipping should be handled automatically if corners are out of range, but if the
            // z coordinate goes behind viewer, there are problems.
            if (pt[2]/pt[3] > 1.0)
            {
                is_visible = 0;
            }
        }

        // Draw the image
        if (is_visible)
        {
            if (fb_src)
            {
                for (i=0; i<3; i++)
                {
                    glTexCoord2f(window_mapped_corners[i][0], window_mapped_corners[i][1]);
                    glVertex2d(window_mapped_corners[i][0], window_mapped_corners[i][1]);
                }
            }
        }
    }
    glEnd();
    
    glPopMatrix();

    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    if (fb_src)
    {
        glDisable(GL_TEXTURE_2D);
    }
}

////////////////////////////////////////////////////////////////////////////
// pilotTrans
//      Translate pilot head origin by given amounts.  Translates
//      in the local pilot coordinate system.
////////////////////////////////////////////////////////////////////////////
void pilotTrans(float x, float y, float z)
{
    pilot_home.origin = pilot_home.origin +
        pilot_home.x*x + pilot_home.y*y + pilot_home.z*z;
}

////////////////////////////////////////////////////////////////////////////
// pilotRot
//      Rotates pilot head by given angles (deg).
////////////////////////////////////////////////////////////////////////////
void pilotRotX(float th)
{
    Vector3 y_new, z_new;

    th = th*M_PI/180.0;
    
    y_new = pilot_home.y*cos(th) - pilot_home.z*sin(th);
    z_new = pilot_home.y*sin(th) + pilot_home.z*cos(th);
    pilot_home.y = y_new;
    pilot_home.z = z_new;
}
void pilotRotY(float th)
{
    Vector3 x_new, z_new;

    th = th*M_PI/180.0;

    x_new = pilot_home.x*cos(th) + pilot_home.z*sin(th);
    z_new = -pilot_home.x*sin(th) + pilot_home.z*cos(th);
    pilot_home.x = x_new;
    pilot_home.z = z_new;
}
void pilotRotZ(float th)
{
    Vector3 x_new, y_new;

    th = th*M_PI/180.0;
    
    x_new = pilot_home.x*cos(th) - pilot_home.y*sin(th);
    y_new = pilot_home.x*sin(th) + pilot_home.y*cos(th);
    pilot_home.x = x_new;
    pilot_home.y = y_new;
}
