////////////////////////////////////////////////////////////////////////////
// xplaneXfer.cc
//      Code to allow data to be exchanged with X-plane through a
//      shared memory interface.  Requires the dataref_server plugin
//      to be used.
//
// Author: Jon W. Wallace, Lafayette College
// Date:   May 2020
////////////////////////////////////////////////////////////////////////////

#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "xplaneXfer.h"
#include "log.h"

const char *shm_name = "/x_plane_dataref_server_plugin_jon_wallace";

int xplane_shm_handle = -1;
xplaneShmDef *xplane_shm_pointer = 0;
xplaneShmDef xplane_shm_local;

int xplaneXferInit()
{
    mode_t oldmask;
    
    // Set some defaults to use here
    memset(&xplane_shm_local, 0, sizeof(xplane_shm_local));
    
    // Get access to the shared memory
    oldmask = umask(0);
    xplane_shm_handle = shm_open(shm_name, O_RDWR | O_CREAT, 0666);
    umask(oldmask);
    if (xplane_shm_handle < 0)
    {
        log_printf(log_error, "Unable to open shared memory for X-plane.\n");
        goto backout_base;
    }

    if (ftruncate(xplane_shm_handle,sizeof(xplaneShmDef)) != 0)
    {
        log_printf(log_error, "ftruncate() failed on shared memory for X-plane.\n");
        goto backout_shm;
    }

    // Allocate space for the shared memory
    if ((xplane_shm_pointer = (xplaneShmDef *) mmap(0, sizeof(xplaneShmDef), PROT_READ | PROT_WRITE, MAP_SHARED, xplane_shm_handle, 0)) == 0)
    {
        log_printf(log_error, "Unable to map shared memory for X-plane.\n");
        goto backout_shm;
    }

    return(0);

 backout_shm:
   close(xplane_shm_handle);

 backout_base:
   return(-1);
}

void xplaneXferUninit()
{
    munmap(xplane_shm_pointer, sizeof(xplaneShmDef));
    close(xplane_shm_handle);
}

////////////////////////////////////////////////////////////////////////////
// xplaneXferPut/Get
//      Idea is to use xplane_shm_local to read or write data and then
//      use the Put or Get functions to copy data all at once.  Should
//      avoid most problems with inconsistent data.  For what we are
//      doing, small inconsistencies won't matter.
////////////////////////////////////////////////////////////////////////////
void xplaneXferPut()
{
    memcpy(&(xplane_shm_pointer->w), &(xplane_shm_local.w), sizeof(xplane_shm_local.w));
}
void xplaneXferGet()
{
    memcpy(&(xplane_shm_local.r), &(xplane_shm_pointer->r), sizeof(xplane_shm_local.r));
}
