/*===========================================================================
 * yuv.cc
 *      Code for decoding YUV video formats.  I used the information in
 *      V4L2 to write this.  Probably the conversions are not exactly
 *      right, but they work well enough for initial testing.
 *
 *      Since switching over to the CUDA OpenCV library, this code is
 *      no longer used.
 *
 * Author: Jon W. Wallace, Lafayette College
 * Date:   May 2020
 *==========================================================================*/

#include <inttypes.h>

#include "yuv.h"

/*----------------------------------------------------------------------------
 * yuv422_to_rgba
 * Note:
 *      Used the following matrix for RGB to YUV conversion and inverted it:
 *      M = [0.2990 0.5870 0.1140; -0.1687 -0.3313 0.5; 0.5 -0.4187 -0.0813];
 *--------------------------------------------------------------------------*/
void yuv422_to_rgba(uint32_t *data_in, uint32_t *data_out, int cap_Nx, int cap_Ny)
{
    int i, j;
    uint32_t yuyv;
    uint8_t y1;
    uint8_t y2;
    uint8_t u;
    uint8_t v;
    float yf, uf, vf;
    float r, g, b;
    uint32_t rgba;
    
    for (j=0; j<cap_Ny; j++)
    {
        for (i=0; i<cap_Nx/2; i++)
        {
            yuyv = data_in[i + j*(cap_Nx/2)];
            y1 = yuyv & 0xff;
            u = (yuyv >> 8) & 0xff;
            y2 = (yuyv >> 16) & 0xff;
            v = (yuyv >> 24) & 0xff;
            uf = (float)u - 128.0;
            vf = (float)v - 128.0;
            yf = (float)y1;
            r = yf + 1.402*vf;
            g = yf - 0.3411*uf - 0.7141*vf;
            b = yf + 1.7720*uf;
            if (r < 0) r = 0; if (r > 255) r = 255;
            if (g < 0) g = 0; if (g > 255) g = 255;
            if (b < 0) b = 0; if (b > 255) b = 255;
            rgba = ((uint8_t)r) |
                ((uint8_t)g << 8) |
                ((uint8_t)b << 16);
            data_out[i*2 + j*cap_Nx + 0] = rgba;
            yf = (float)y2;
            r = yf + 1.402*vf;
            g = yf - 0.3411*uf - 0.7141*vf;
            b = yf + 1.7720*uf;
            if (r < 0) r = 0; if (r > 255) r = 255;
            if (g < 0) g = 0; if (g > 255) g = 255;
            if (b < 0) b = 0; if (b > 255) b = 255;
            rgba = ((uint8_t)r) |
                ((uint8_t)g << 8) |
                ((uint8_t)b << 16);
            data_out[i*2 + j*cap_Nx + 1] = rgba;
        }
    }
}

float M1[9] = 
    {1.0, 0.0, 1.402,
     1.0, -0.3411, -0.7141,
     1.0, 1.7720, 0.0};

/*
float M1[9] =
    {1.0, 0.0, 1.371,
     1.0, -0.698, -0.336,
     1.0, 1.732, 0.0};
*/


/*
float M1[9] = 
    {1.0, 0.0, 1.13983,
     1.0, -0.39465, -0.58060,
     1.0, 2.03211, 0};
*/

/*
float M1[9] =
    {1.164, 0.0, 1.596,
     1.164, -0.813, -0.392,
     1.164, 2.017, 0.0};
*/

inline void yuv_to_rgb(uint8_t y, uint8_t u, uint8_t v,
                       uint8_t &r, uint8_t &g, uint8_t &b)
{
    float uf, vf, yf;
    int ri, gi, bi;
    
    uf = (float)u - 32.0;
    vf = (float)v - 32.0;
    yf = (float)y;
    ri = (int)(M1[0]*yf + M1[1]*uf + M1[2]*vf);
    gi = (int)(M1[3]*yf + M1[4]*uf + M1[5]*vf);
    bi = (int)(M1[6]*yf + M1[7]*uf + M1[8]*vf);
    if (ri < 0) ri = 0; if (ri > 255) ri = 255;
    if (gi < 0) gi = 0; if (gi > 255) gi = 255;
    if (bi < 0) bi = 0; if (bi > 255) bi = 255;
    r = (uint8_t)ri;
    g = (uint8_t)gi;
    b = (uint8_t)bi;
}

// Bayer input pixels (left/right frames)
#define PIX_IN_L(xi, yi) (2*((xi) + (yi)*cap_Nx)+0)
#define PIX_IN_R(xi, yi) (2*((xi) + (yi)*cap_Nx)+1)

// RGBA output pixels
#define PIX_OUT(xi, yi) ((((xi) + (yi)*cap_Nx)*4))
#define AVG2(a, b) (((int)(a) + (int)(b))/2)
#define AVG4(a, b, c, d) (((int)(a) + (int)(b) + (int)(c) + (int)(d))/4)

/*----------------------------------------------------------------------------
 * bayer_s_to_rgba
 *      Stereo interleaved Bayer raw to RGBA.  We do averaging at pixel
 *      centers.  Current code skips the border pixels.
 * OVR camera uses following format:
 *      G R G R
 *      B G B G 
 * Note:
 *--------------------------------------------------------------------------*/
// Combined loops
void bayer_s_to_rgba(uint8_t *data_in, uint8_t *data_out_l, uint8_t *data_out_r, int cap_Nx, int cap_Ny)
{
    int i, j, idx;
    uint8_t g1_l;
    uint8_t g1_r;
    uint8_t g2_l;
    uint8_t g2_r;
    uint8_t r_l;
    uint8_t r_r;
    uint8_t b_l;
    uint8_t b_r;

    // Left Frame

    for (j=1; j<cap_Ny-1; j+=2)
    {
        for (i=1; i<cap_Nx-1; i+=2)
        {
            // Do green_1 centered pixels (i+0, j+0)
            idx = PIX_OUT(i, j);
            data_out_l[idx + 0] = AVG2(data_in[PIX_IN_L(i, j-1)],
                                       data_in[PIX_IN_L(i, j+1)]);
            data_out_l[idx + 1] = data_in[PIX_IN_L(i, j)];
            data_out_l[idx + 2] = AVG2(data_in[PIX_IN_L(i-1, j)],
                                       data_in[PIX_IN_L(i+1, j)]);
            data_out_l[idx + 3] = 255;
           
            // Do blue centered pixels (i+1)
            idx = PIX_OUT(i+1, j);
            data_out_l[idx + 0] = AVG4(data_in[PIX_IN_L(i-1+1,j-1)],
                                       data_in[PIX_IN_L(i-1+1,j+1)],
                                       data_in[PIX_IN_L(i+1+1,j-1)],
                                       data_in[PIX_IN_L(i+1+1,j+1)]);
            data_out_l[idx + 1] = AVG4(data_in[PIX_IN_L(i+1,j-1)], 
                                       data_in[PIX_IN_L(i+1,j+1)], 
                                       data_in[PIX_IN_L(i+1+1,j)], 
                                       data_in[PIX_IN_L(i+1+1,j)]);
            data_out_l[idx + 2] = data_in[PIX_IN_L(i+1, j)];
            data_out_l[idx + 3] = 255;

            // Do Red pixel centered pixels (j+1)
            idx = PIX_OUT(i, (j+1));
            data_out_l[idx + 0] = data_in[PIX_IN_L(i, j+1)];
            data_out_l[idx + 1] = AVG4(data_in[PIX_IN_L(i,j-1+1)], 
                                       data_in[PIX_IN_L(i,j+1+1)], 
                                       data_in[PIX_IN_L(i+1,j+1)], 
                                       data_in[PIX_IN_L(i+1,j+1)]);
            data_out_l[idx + 2] = AVG4(data_in[PIX_IN_L(i-1,j-1+1)],
                                       data_in[PIX_IN_L(i-1,j+1+1)],
                                       data_in[PIX_IN_L(i+1,j-1+1)],
                                       data_in[PIX_IN_L(i+1,j+1+1)]);
            data_out_l[idx + 3] = 255;

            // Do green_2 centered pixels (i+1,j+1)
            idx = PIX_OUT(i+1, (j+1));
            data_out_l[idx + 0] = AVG2(data_in[PIX_IN_L(i+1, j+1)],
                                       data_in[PIX_IN_L(i-1+1, j+1)]);
            data_out_l[idx + 1] = data_in[PIX_IN_L(i, j)];
            data_out_l[idx + 2] = AVG2(data_in[PIX_IN_L(i+1, j+1+1)],
                                       data_in[PIX_IN_L(i+1, j-1+1)]);
            data_out_l[idx + 3] = 255;
            
            //// Right Frame
                
            // Do green_1 centered pixels
            idx = PIX_OUT(i, j);
            data_out_r[idx + 0] = AVG2(data_in[PIX_IN_R(i, j-1)],
                                       data_in[PIX_IN_R(i, j+1)]);
            data_out_r[idx + 1] = data_in[PIX_IN_R(i, j)];
            data_out_r[idx + 2] = AVG2(data_in[PIX_IN_R(i-1, j)],
                                       data_in[PIX_IN_R(i+1, j)]);
            data_out_r[idx + 3] = 255;
           
            // Do blue centered pixels (i+1)
            idx = PIX_OUT(i+1, j);
            data_out_r[idx + 0] = AVG4(data_in[PIX_IN_R(i-1+1,j-1)],
                                       data_in[PIX_IN_R(i-1+1,j+1)],
                                       data_in[PIX_IN_R(i+1+1,j-1)],
                                       data_in[PIX_IN_R(i+1+1,j+1)]);
            data_out_r[idx + 1] = AVG4(data_in[PIX_IN_R(i+1,j-1)], 
                                       data_in[PIX_IN_R(i+1,j+1)], 
                                       data_in[PIX_IN_R(i+1+1,j)], 
                                       data_in[PIX_IN_R(i+1+1,j)]);
            data_out_r[idx + 2] = data_in[PIX_IN_R(i+1, j)];
            data_out_r[idx + 3] = 255;

            // Do Red pixel centered pixels (j+1)
            idx = PIX_OUT(i, (j+1));
            data_out_r[idx + 0] = data_in[PIX_IN_R(i, j+1)];
            data_out_r[idx + 1] = AVG4(data_in[PIX_IN_R(i,j-1+1)], 
                                       data_in[PIX_IN_R(i,j+1+1)], 
                                       data_in[PIX_IN_R(i+1,j+1)], 
                                       data_in[PIX_IN_R(i+1,j+1)]);
            data_out_r[idx + 2] = AVG4(data_in[PIX_IN_R(i-1,j-1+1)],
                                       data_in[PIX_IN_R(i-1,j+1+1)],
                                       data_in[PIX_IN_R(i+1,j-1+1)],
                                       data_in[PIX_IN_R(i+1,j+1+1)]);
            data_out_r[idx + 3] = 255;

            // Do green_2 centered pixels (i+1, j+1)
            idx = PIX_OUT(i+1, (j+1));
            data_out_r[idx + 0] = AVG2(data_in[PIX_IN_R(i+1, j+1)],
                                       data_in[PIX_IN_R(i-1+1, j+1)]);
            data_out_r[idx + 1] = data_in[PIX_IN_R(i+1, j+1)];
            data_out_r[idx + 2] = AVG2(data_in[PIX_IN_R(i+1, j+1+1)],
                                       data_in[PIX_IN_R(i+1, j-1+1)]);
            data_out_r[idx + 3] = 255;
        }
    }
}

